
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DOCBChDepc.h: DOCB combined dependency and regional protocol(classes)
///////////////////////////////////////////////////////////////////////////


class HybridMovingWindDep {

 private:

  // when using traversedMoveMap make sure that a traversal starting from the
  // same root object, always touches the same set of objects, due to the 
  // way the current implementation is done.

  hash_map<int, MovingWindowChange *, hash<int>, eqoid> _bySRefMoveMap;
  hash_map<int, MovingWindowChange *, hash<int>, eqoid> _byDRefMoveMap;
  hash_map<int, MovingWindowChange *, hash<int>, eqoid> _traversedMoveMap;
  hash_map<int, MovingWindowChange *, hash<int>, eqoid> _sameClassMoveMap;

  MovingWindowChange * _randomMove;

  float _depRandomProb;
  float _depSRefProb;
  float _depDRefProb;
  float _depTraversedProb;
  float _depClassProb;
  
  ByRefDep * _bySRefDep;
  ByRefDep * _byDRefDep;
  TraversedDep * _traversedDep;
  SameClassDep * _sameClassDep;

  float _regionSize;
  float _highestProbW;
  float _lowestProbW;
  int _H;
  objAssMeth_t _assignMethod;
  int _R;
  int _numRoots;
  
  RandomDep * _randomDep;

 public:

   HybridMovingWindDep( float depRandomProb, float depSRefProb, 
		        float depDRefProb, float depTraversedProb, 
		        float depClassProb, float U, float C, float regionSize,
		        float highestProbW, float lowestProbW, int H,
		        objAssMeth_t assignMethod, int R, float hotRegionSize,
			float hotRegionProb);
  
   void InsertObjectRandom(int classId, int objId);
   void InsertSRef(int rootId, int refId);
   void InsertDRef(int rootId, int refId);
   void InsertTraversedObject(int rootId, int traversedId);
   void InsertSameClassObject(int classId, int objId);
   void Assign();
   int NextRoot(int currRoot);
   
};

class HybridGMovingWindDep {

 private:

  // when using traversedMoveMap make sure that a traversal starting from the
  // same root object, always touches the same set of objects, due to the 
  // way the current implementation is done.
  
  hash_map<int, GMovingWindowChange *, hash<int>, eqoid> _bySRefMoveMap;
  hash_map<int, GMovingWindowChange *, hash<int>, eqoid> _byDRefMoveMap;
  hash_map<int, GMovingWindowChange *, hash<int>, eqoid> _traversedMoveMap;
  hash_map<int, GMovingWindowChange *, hash<int>, eqoid> _sameClassMoveMap;

  GMovingWindowChange * _randomMove;

  float _depRandomProb;
  float _depSRefProb;
  float _depDRefProb;
  float _depTraversedProb;
  float _depClassProb;
  
  ByRefDep * _bySRefDep;
  ByRefDep * _byDRefDep;
  TraversedDep * _traversedDep;
  SameClassDep * _sameClassDep;

  float _regionSize;
  float _highestProbW;
  float _lowestProbW;
  int _H;
  objAssMeth_t _assignMethod;
  float _probWIncrSize;

  int _R;
  int _numRoots;

  RandomDep * _randomDep;

 public:

  // WARNING: if depTraversedProb is not zero, then the type of traversal
  //          used must always traverse the same set of objects from the
  //          same root, else, the program will only remember the objects,
  //          accessed in the first ever traversal (from this root), all
  //          subsequent traversals from the same root will not count.
  
  HybridGMovingWindDep(float depRandomProb, float depSRefProb, 
		       float depDRefProb, float depTraversedProb, 
		       float depClassProb, float U, float C, float regionSize,
		       float highestProbW, float lowestProbW, int H,
		       objAssMeth_t assignMethod, float probWIncrSize, int R,
		       float hotRegionSize, float hotRegionProb);
  
  void InsertObjectRandom(int classId, int objId);
  void InsertSRef(int rootId, int refId);
  void InsertDRef(int rootId, int refId);
  void InsertTraversedObject(int rootId, int traversedId);
  void InsertSameClassObject(int classId, int objId);
  void Assign();
  int NextRoot(int currRoot);
  
};

class HybridHeatGrowthDep {
  
 private:

  // when using traversedMoveMap make sure that a traversal starting from the
  // same root object, always touches the same set of objects, due to the 
  // way the current implementation is done.

  hash_map<int, HeatGrowthChange *, hash<int>, eqoid> _bySRefMoveMap;
  hash_map<int, HeatGrowthChange *, hash<int>, eqoid> _byDRefMoveMap;
  hash_map<int, HeatGrowthChange *, hash<int>, eqoid> _traversedMoveMap;
  hash_map<int, HeatGrowthChange *, hash<int>, eqoid> _sameClassMoveMap;

  HeatGrowthChange * _randomMove;
  
  float _depRandomProb;
  float _depSRefProb;
  float _depDRefProb;
  float _depTraversedProb;
  float _depClassProb;
  
  ByRefDep * _bySRefDep;
  ByRefDep * _byDRefDep;
  TraversedDep * _traversedDep;
  SameClassDep * _sameClassDep;
  
  float _smallRegionSize;
  float _largeRegionSize;
  float _highestProbW;
  float _lowestProbW;
  
  int _numPeriods;
  int _numTransactions;
  objAssMeth_t _assignMethod;
  
  int _R;
  int _numRoots;
    
  RandomDep * _randomDep;
  
 public:
  
  HybridHeatGrowthDep(float depRandomProb, float depSRefProb, 
		      float depDRefProb, float depTraversedProb, 
		      float depClassProb, float U, float C, 
		      float smallRegionSize, float largeRegionSize, 
		      float highestProbW, float lowestProbW,
		      int numPeriods, objAssMeth_t assignMethod,
		      int numTransactions, int R, float hotRegionSize,
		      float hotRegionProb);
		       
  void InsertObjectRandom(int classId, int objId);
  void InsertSRef(int rootId, int refId);
  void InsertDRef(int rootId, int refId);
  void InsertTraversedObject(int rootId, int traversedId);
  void InsertSameClassObject(int classId, int objId);
  void Assign();  
  int NextRoot(int currRoot);

};

class HybridCycleDep {

  
 private:

  // when using traversedMoveMap make sure that a traversal starting from the
  // same root object, always touches the same set of objects, due to the 
  // way the current implementation is done.

  hash_map<int, CycleOfChange *, hash<int>, eqoid> _bySRefMoveMap;
  hash_map<int, CycleOfChange *, hash<int>, eqoid> _byDRefMoveMap;
  hash_map<int, CycleOfChange *, hash<int>, eqoid> _traversedMoveMap;
  hash_map<int, CycleOfChange *, hash<int>, eqoid> _sameClassMoveMap;

  CycleOfChange * _randomMove;
  
  float _depRandomProb;
  float _depSRefProb;
  float _depDRefProb;
  float _depTraversedProb;
  float _depClassProb;
  
  ByRefDep * _bySRefDep;
  ByRefDep * _byDRefDep;
  TraversedDep * _traversedDep;
  SameClassDep * _sameClassDep;
  
  float _regionSize;
  float _highestProbW;
  float _lowestProbW;

  objAssMeth_t _assignMethod;
  
  int _H;
  int _R;
  int _numRoots;
    
  RandomDep * _randomDep;
  
 public:
  
  HybridCycleDep(float depRandomProb, float depSRefProb, 
		 float depDRefProb, float depTraversedProb, 
		 float depClassProb, float U, float C, float regionSize,
		 float highestProbW, float lowestProbW, int H,
		 objAssMeth_t assignMethod, int R, float hotRegionSize,
		 float hotRegionProb);  
         
  void InsertObjectRandom(int classId, int objId);
  void InsertSRef(int rootId, int refId);
  void InsertDRef(int rootId, int refId);
  void InsertTraversedObject(int rootId, int traversedId);
  void InsertSameClassObject(int classId, int objId);
  void Assign();  
  int NextRoot(int currRoot);
};
