
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DOCBChDepm.h: DOCB combined dependency and regional protocol(methods)
///////////////////////////////////////////////////////////////////////////

HybridMovingWindDep::HybridMovingWindDep( float depRandomProb, 
					  float depSRefProb, 
					  float depDRefProb, 
					  float depTraversedProb, 
					  float depClassProb, float U, float C,
					  float regionSize,float highestProbW, 
					  float lowestProbW, int H,
					  objAssMeth_t assignMethod, int R,
					  float hotRegionSize, 
					  float hotRegionProb) {
  
  _regionSize = regionSize;
  _highestProbW = highestProbW;
  _lowestProbW = lowestProbW;
  _H = H;
  _R = R;
  _assignMethod = assignMethod;
  
  _depRandomProb = depRandomProb;
  _depSRefProb = depSRefProb;
  _depDRefProb = depDRefProb;
  _depTraversedProb = depTraversedProb;
  _depClassProb = depClassProb;

  _bySRefDep = new ByRefDep();
  _byDRefDep = new ByRefDep();
  _traversedDep = new TraversedDep(C);
  _sameClassDep = new SameClassDep(U);

  _randomMove = new MovingWindowChange(regionSize, highestProbW, lowestProbW,
				       H, assignMethod);

  _randomDep = new RandomDep(hotRegionSize, hotRegionProb, assignMethod);

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

  _numRoots = 0;
};

void HybridMovingWindDep::InsertObjectRandom(int classId, int objId) {

  _randomMove->InsertObject(classId, objId);
  _randomDep->InsertObject(classId, objId);

};

void HybridMovingWindDep::InsertSRef(int rootId, int refId) {
  
  _bySRefDep->InsertRef(rootId, refId);

};

void HybridMovingWindDep::InsertDRef(int rootId, int refId) {

  _byDRefDep->InsertRef(rootId, refId);

};

void HybridMovingWindDep::InsertTraversedObject(int rootId, int traversedId) {

  _traversedDep->InsertObject(rootId, traversedId);

};

void HybridMovingWindDep::InsertSameClassObject(int classId, int objId) {

  _sameClassDep->InsertObject(classId, objId);

};

void HybridMovingWindDep::Assign() {
  
  _randomMove->AssignObjects();
  _randomDep->AssignObjects();

  // assign objects for SRefDep move window
  vector<int> refSRootVect = _bySRefDep->GetRootSet();
  for (unsigned int i = 0; i < refSRootVect.size(); i++) {

    set<int> * candidateSet = _bySRefDep->GetCandidateSet(refSRootVect[i]);
    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = round((1.0 / (float) _regionSize));

    if (candidateSet->size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet->size();
    }

    //    cout << "candidate size = " << candidateSet->size() << endl;

    float tempRegionSize = 1.0 / (float) tempNumRegions;

    // cout << "tempRegionSize = " << tempRegionSize << endl;

    MovingWindowChange * mwc = new MovingWindowChange(tempRegionSize, 
						      _highestProbW,
						      _lowestProbW, _H,
						      _assignMethod);
    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      mwc->InsertObject(refSRootVect[i], *it);
    }

    mwc->AssignObjects();    
    _bySRefMoveMap[refSRootVect[i]] = mwc;
  }

  
  // assign objects for DRefDep move window
  vector<int> refDRootVect = _byDRefDep->GetRootSet();
  for (unsigned int i = 0; i < refDRootVect.size(); i++) {

    
    set<int> * candidateSet = _byDRefDep->GetCandidateSet(refDRootVect[i]);

    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = (round)( (1.0 / (float) _regionSize) );

    if (candidateSet->size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet->size();
    }

    //    cout << "candidate size = " << candidateSet->size() << endl;

    float tempRegionSize = 1.0 / (float) tempNumRegions;

    //    cout << "tempRegionSize = " << tempRegionSize << endl;

    MovingWindowChange * mwc = new MovingWindowChange(tempRegionSize, 
						      _highestProbW,
						      _lowestProbW, _H,
						      _assignMethod);

    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      mwc->InsertObject(refDRootVect[i], *it);
    }

    mwc->AssignObjects();    
    _byDRefMoveMap[refDRootVect[i]] = mwc;

  }
   
  _sameClassDep->AssignSets();
  //  cout << "here" << endl;

  // assign objects for SameClassDep move window
  vector<int> sameClassRootVect = _sameClassDep->GetRootSet();
  for (unsigned int i = 0; i < sameClassRootVect.size(); i++) {
 
    //    cout << "candidate set = ";
    vector<int> candidateSet 
      = _sameClassDep->GetCandidateSet(sameClassRootVect[i]);
        
    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = (round) ((1.0 / (float) _regionSize));

    if (candidateSet.size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet.size();
    }

    //    cout << "candidate size = " << candidateSet.size() << endl;

    float tempRegionSize = 1.0 / (float) tempNumRegions;

    //    cout << "tempRegionSize = " << tempRegionSize << endl;

    MovingWindowChange * mwc = new MovingWindowChange(tempRegionSize, 
						      _highestProbW,
						      _lowestProbW, _H,
						      _assignMethod);
 
    for (unsigned int j = 0; j < candidateSet.size(); j++) {
      mwc->InsertObject(sameClassRootVect[i], candidateSet[j]);
      //      cout << candidateSet[j] << ", ";
    }
    //    cout << endl;
    mwc->AssignObjects();    
    _sameClassMoveMap[sameClassRootVect[i]] = mwc;

    //    cout << "root considered in same class = " << sameClassRootVect[i] << endl;

  }

}


int HybridMovingWindDep::NextRoot(int currRoot) {

  float randomValue = (float)randu(lp_tt);
  int returnValue;

  // every _R roots selected, moves goto back to random root.
  if ((_R != -1)  && ((_numRoots % (_R + 1)) == 0)) {


    returnValue = _randomDep->NextRoot();
    
  }
  else { 


    if (randomValue <= _depRandomProb) {
      
      returnValue = _randomMove->NextRoot();
    
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb)) {
      
      if (_bySRefMoveMap.find(currRoot) != _bySRefMoveMap.end()) {
	
	returnValue = _bySRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
      returnValue = NULL_OID;
      
      }	
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb)) {
      
      if (_byDRefMoveMap.find(currRoot) != _byDRefMoveMap.end()) {
	
	returnValue = _byDRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
      returnValue = NULL_OID;
      
    }
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb + 
			     _depTraversedProb)) {
      if (_traversedMoveMap.find(currRoot) != _traversedMoveMap.end()) {
	returnValue  = _traversedMoveMap[currRoot]->NextRoot();
      }
      else {      
	vector<int> candidateSet 
	  = _traversedDep->GetCandidateSet(currRoot);
	
	
	int tempNumRegions;
	int numRegions;
	
	// make the number of regions smaller or equal to the candidate set, 
	// since can not have empty regions.
	
	numRegions = round ( (1.0 / (float) _regionSize));
	
	if (candidateSet.size() >= ( unsigned int ) numRegions) {
	  tempNumRegions = numRegions;
	} 
	else {
	  tempNumRegions = candidateSet.size();
	}
	
	cout << "candidate size = " << candidateSet.size() << endl;
	
	float tempRegionSize = 1.0 / (float) tempNumRegions;
	
	cout << "tempRegionSize = " << tempRegionSize << endl;

	MovingWindowChange * mwc = new MovingWindowChange(tempRegionSize, 
							  _highestProbW,
							  _lowestProbW, _H,
							  _assignMethod);
	
	cout << "candidate set = ";
	for (unsigned int j = 0; j < candidateSet.size(); j++) {
	  mwc->InsertObject(currRoot, candidateSet[j]);
	  cout << candidateSet[j] << ", ";
	}
	
	cout << endl;
	mwc->AssignObjects();    
	_traversedMoveMap[currRoot] = mwc;
	
	returnValue = _traversedMoveMap[currRoot]->NextRoot();
      }
      
    }
    else {
      
      if (_sameClassMoveMap.find(currRoot) != _sameClassMoveMap.end()) {
	
	returnValue = _sameClassMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
	returnValue = NULL_OID;
	
      }
    }
    
  }
  _numRoots++;

  return returnValue;
  
};


HybridGMovingWindDep::HybridGMovingWindDep( float depRandomProb, 
					    float depSRefProb, 
					    float depDRefProb, 
					    float depTraversedProb, 
					    float depClassProb, float U,
					    float C, float regionSize, 
					    float highestProbW, 
					    float lowestProbW, int H,
					    objAssMeth_t assignMethod,
					    float probWIncrSize, int R,
					    float hotRegionSize,
					    float hotRegionProb) {
  
  _regionSize = regionSize;
  _highestProbW = highestProbW;
  _lowestProbW = lowestProbW;
  _H = H;
  _R = R;
  _assignMethod = assignMethod;
  _probWIncrSize = probWIncrSize;

  _depRandomProb = depRandomProb;
  _depSRefProb = depSRefProb;
  _depDRefProb = depDRefProb;
  _depTraversedProb = depTraversedProb;
  _depClassProb = depClassProb;

  _bySRefDep = new ByRefDep();
  _byDRefDep = new ByRefDep();
  _traversedDep = new TraversedDep(C);
  _sameClassDep = new SameClassDep(U);

  _randomMove = new GMovingWindowChange(regionSize, highestProbW, lowestProbW,
					H, assignMethod, probWIncrSize);

  _randomDep = new RandomDep(hotRegionSize, hotRegionProb, assignMethod);

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

  _numRoots = 0;
};

void HybridGMovingWindDep::InsertObjectRandom(int classId, int objId) {

  _randomMove->InsertObject(classId, objId);  
  _randomDep->InsertObject(classId, objId);
  
};

void HybridGMovingWindDep::InsertSRef(int rootId, int refId) {
  
  _bySRefDep->InsertRef(rootId, refId);

};

void HybridGMovingWindDep::InsertDRef(int rootId, int refId) {

  _byDRefDep->InsertRef(rootId, refId);

};

void HybridGMovingWindDep::InsertTraversedObject(int rootId, int traversedId) {

  _traversedDep->InsertObject(rootId, traversedId);

};

void HybridGMovingWindDep::InsertSameClassObject(int classId, int objId) {

  _sameClassDep->InsertObject(classId, objId);

};

void HybridGMovingWindDep::Assign() {

  
  _randomMove->AssignObjects();
  _randomDep->AssignObjects();

  // assign objects for SRefDep move window
  vector<int> refSRootVect = _bySRefDep->GetRootSet();
  for (unsigned int i = 0; i < refSRootVect.size(); i++) {

    set<int> * candidateSet = _bySRefDep->GetCandidateSet(refSRootVect[i]);
    
    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = round((1.0 / (float) _regionSize));

    if (candidateSet->size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet->size();
    }


    float tempRegionSize = 1.0 / (float) tempNumRegions;

    GMovingWindowChange * gmwc = new GMovingWindowChange(tempRegionSize, 
							 _highestProbW,
							 _lowestProbW, _H,
							 _assignMethod,
							 _probWIncrSize);
    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      gmwc->InsertObject(refSRootVect[i], *it);
    }

    gmwc->AssignObjects();    
    _bySRefMoveMap[refSRootVect[i]] = gmwc;
  }

  
  // assign objects for DRefDep move window
  vector<int> refDRootVect = _byDRefDep->GetRootSet();
  for (unsigned int i = 0; i < refDRootVect.size(); i++) {

    set<int> * candidateSet = _byDRefDep->GetCandidateSet(refDRootVect[i]);

    
    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = (round)( (1.0 / (float) _regionSize) );

    if (candidateSet->size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet->size();
    }

    float tempRegionSize = 1.0 / (float) tempNumRegions;

    GMovingWindowChange * gmwc = new GMovingWindowChange(tempRegionSize, 
							 _highestProbW,
							 _lowestProbW, _H,
							 _assignMethod,
							 _probWIncrSize);
    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      gmwc->InsertObject(refDRootVect[i], *it);
    }

    gmwc->AssignObjects();    
    _byDRefMoveMap[refDRootVect[i]] = gmwc;
    
  }
  
  _sameClassDep->AssignSets();
  
  // assign objects for SameClassDep move window
  vector<int> sameClassRootVect = _sameClassDep->GetRootSet();
  for (unsigned int i = 0; i < sameClassRootVect.size(); i++) {
    
     //    cout << "candidate set = ";
    vector<int> candidateSet 
      = _sameClassDep->GetCandidateSet(sameClassRootVect[i]);

    
    int tempNumRegions;
    int numRegions;

    // make the number of regions smaller or equal to the candidate set, 
    // since can not have empty regions.

    numRegions = (round) ((1.0 / (float) _regionSize));

    if (candidateSet.size() >= ( unsigned int ) numRegions) {
      tempNumRegions = numRegions;
    } 
    else {
      tempNumRegions = candidateSet.size();
    }

    float tempRegionSize = 1.0 / (float) tempNumRegions;


    GMovingWindowChange * gmwc = new GMovingWindowChange(tempRegionSize, 
							 _highestProbW,
							 _lowestProbW, _H,
							 _assignMethod,
							 _probWIncrSize);
   
    for (unsigned int j = 0; j < candidateSet.size(); j++) {
      gmwc->InsertObject(sameClassRootVect[i], candidateSet[j]);
    //   cout << candidateSet[j] << ", ";
    }
    // cout << endl;
    gmwc->AssignObjects();    
    _sameClassMoveMap[sameClassRootVect[i]] = gmwc;

    // cout << "root considered in same class = " << sameClassRootVect[i] << endl;

  }
  
}


int HybridGMovingWindDep::NextRoot(int currRoot) {

  float randomValue = (float)randu(lp_tt);
  int returnValue;

  if ((_R != -1)  && ((_numRoots % (_R + 1)) == 0)) {

    returnValue = _randomDep->NextRoot();

  }
  else {


    if (randomValue <= _depRandomProb) {
    
      returnValue = _randomMove->NextRoot();

    }
    else if (randomValue <= (_depRandomProb + _depSRefProb)) {
    
      if (_bySRefMoveMap.find(currRoot) != _bySRefMoveMap.end()) {
	
	returnValue = _bySRefMoveMap[currRoot]->NextRoot();
      
      }
      else {
	
	returnValue = NULL_OID;

      }	

    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb)) {
           
      if (_byDRefMoveMap.find(currRoot) != _byDRefMoveMap.end()) {
	
	returnValue = _byDRefMoveMap[currRoot]->NextRoot();
      
      }
      else {

	returnValue = NULL_OID;

      }
   
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb + 
			     _depTraversedProb)) {
      if (_traversedMoveMap.find(currRoot) != _traversedMoveMap.end()) {
	returnValue = _traversedMoveMap[currRoot]->NextRoot();
      }
      else {
	vector<int> candidateSet 
	  = _traversedDep->GetCandidateSet(currRoot);
	
	
	int tempNumRegions;
	int numRegions;
	
	// make the number of regions smaller or equal to the candidate set, 
	// since can not have empty regions.
	
	numRegions = round ( (1.0 / (float) _regionSize));
	
	if (candidateSet.size() >= ( unsigned int ) numRegions) {
	  tempNumRegions = numRegions;
	} 
	else {
	  tempNumRegions = candidateSet.size();
	}
	
	cout << "candidate size = " << candidateSet.size() << endl;
	
	float tempRegionSize = 1.0 / (float) tempNumRegions;
	
	GMovingWindowChange * gmwc = new GMovingWindowChange(tempRegionSize, 
							     _highestProbW,
							     _lowestProbW, _H,
							     _assignMethod,
							     _probWIncrSize);
	
	cout << "candidate set = ";
	for (unsigned int j = 0; j < candidateSet.size(); j++) {
	  gmwc->InsertObject(currRoot, candidateSet[j]);
	  cout << candidateSet[j] << ", ";
	}
	
	cout << endl;
	gmwc->AssignObjects();    
	_traversedMoveMap[currRoot] = gmwc;
	
	returnValue = _traversedMoveMap[currRoot]->NextRoot();
      }
      
    }
    else {
      
      if (_sameClassMoveMap.find(currRoot) != _sameClassMoveMap.end()) {
	
	returnValue = _sameClassMoveMap[currRoot]->NextRoot();
      
      }
      else {
	
	returnValue = NULL_OID;
	
      }
    
    }
  }

  _numRoots++;

  return returnValue;
  
};

HybridHeatGrowthDep::HybridHeatGrowthDep(float depRandomProb,
					 float depSRefProb, 
					 float depDRefProb, 
					 float depTraversedProb, 
					 float depClassProb, float U,
					 float C, float smallRegionSize,
					 float largeRegionSize, 
					 float highestProbW, 
					 float lowestProbW, 
					 int numPeriods,
					 objAssMeth_t assignMethod,
					 int numTransactions, int R,
					 float hotRegionSize,
					 float hotRegionProb) {

  _depRandomProb = depRandomProb;
  _depSRefProb = depSRefProb;
  _depDRefProb = depDRefProb;
  _depTraversedProb = depTraversedProb;
  _depClassProb = depClassProb;

  _bySRefDep = new ByRefDep();
  _byDRefDep = new ByRefDep();
  _traversedDep = new TraversedDep(C);
  _sameClassDep = new SameClassDep(U);

  _smallRegionSize = smallRegionSize;
  _largeRegionSize = largeRegionSize;
  _highestProbW = highestProbW;
  _lowestProbW = lowestProbW;
  _numPeriods = numPeriods;
  _numTransactions = numTransactions;
  _assignMethod = assignMethod;
  _R = R;

  int likelyNumTrans;
  if (R != -1) {
    likelyNumTrans = round(((float) numTransactions / (float)(R + 1)) 
    * ((float) R) * depRandomProb);
  }
  else {
    likelyNumTrans = round ((float) numTransactions * depRandomProb);
  }

  _randomMove = new HeatGrowthChange(smallRegionSize, largeRegionSize,
				     highestProbW, lowestProbW, numPeriods,
				     assignMethod, likelyNumTrans);

  _randomDep = new RandomDep(hotRegionSize, hotRegionProb, assignMethod);

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;
  
  _numRoots = 0;
}


void HybridHeatGrowthDep::InsertObjectRandom(int classId, int objId) {

  _randomMove->InsertObject(classId, objId);
  _randomDep->InsertObject(classId, objId);

};

void HybridHeatGrowthDep::InsertSRef(int rootId, int refId) {
  
  _bySRefDep->InsertRef(rootId, refId);

};

void HybridHeatGrowthDep::InsertDRef(int rootId, int refId) {

  _byDRefDep->InsertRef(rootId, refId);

};

void HybridHeatGrowthDep::InsertTraversedObject(int rootId,
						   int traversedId) {

  _traversedDep->InsertObject(rootId, traversedId);

};

void HybridHeatGrowthDep::InsertSameClassObject(int classId, int objId) {

  _sameClassDep->InsertObject(classId, objId);

};

void HybridHeatGrowthDep::Assign() {

  _randomMove->AssignObjects();
  _randomDep->AssignObjects();

  vector<int> refSRootVect = _bySRefDep->GetRootSet();
  for (unsigned int i = 0; i < refSRootVect.size(); i++) {

    set<int> * candidateSet = _bySRefDep->GetCandidateSet(refSRootVect[i]);
   
    int likelyNumTrans;
    if (_R != -1) {
      likelyNumTrans = round(((float) _numTransactions / (float)(_R + 1)) 
	* ((float) _R) * _depSRefProb * _highestProbW);
    }
    else {
      likelyNumTrans = round((float) _numTransactions * _depSRefProb
			     * _highestProbW);
    }
 
    HeatGrowthChange * hgc = new HeatGrowthChange(_smallRegionSize,
						  _largeRegionSize,
						  _highestProbW,
						  _lowestProbW, _numPeriods,
						  _assignMethod,
						  likelyNumTrans);
    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      hgc->InsertObject(refSRootVect[i], *it);
    }

    hgc->AssignObjects();    
    _bySRefMoveMap[refSRootVect[i]] = hgc;

  }

  
  // assign objects for DRefDep move window
  vector<int> refDRootVect = _byDRefDep->GetRootSet();
  for (unsigned int i = 0; i < refDRootVect.size(); i++) {

    set<int> * candidateSet = _byDRefDep->GetCandidateSet(refDRootVect[i]);

    
    int likelyNumTrans;
    if (_R != -1) {
      likelyNumTrans = round(((float) _numTransactions / (float)(_R + 1)) 
	* ((float) _R) * _depDRefProb * _highestProbW);
    }
    else {
      likelyNumTrans = round((float) _numTransactions * _depDRefProb 
			     * _highestProbW);
    }

    HeatGrowthChange * hgc = new HeatGrowthChange(_smallRegionSize,
						  _largeRegionSize,
						  _highestProbW,
						  _lowestProbW, _numPeriods,
						  _assignMethod,
						  likelyNumTrans);

    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      hgc->InsertObject(refDRootVect[i], *it);
    }

    hgc->AssignObjects();    
    _byDRefMoveMap[refDRootVect[i]] = hgc;
    
  }
  
  
  
  _sameClassDep->AssignSets();
  
  // assign objects for SameClassDep move window
  vector<int> sameClassRootVect = _sameClassDep->GetRootSet();
  for (unsigned int i = 0; i < sameClassRootVect.size(); i++) {
    
     //    cout << "candidate set = ";
    vector<int> candidateSet 
      = _sameClassDep->GetCandidateSet(sameClassRootVect[i]);
    
    
    int likelyNumTrans;
    if (_R != -1) {
      likelyNumTrans = round(((float) _numTransactions / (float)(_R + 1)) 
			     * ((float) _R) * _depClassProb * _highestProbW);
    }
    else {
      likelyNumTrans =
	round((float) _numTransactions * _depClassProb * _highestProbW);
    }

    HeatGrowthChange * hgc = new HeatGrowthChange(_smallRegionSize,
						  _largeRegionSize,
						  _highestProbW,
						  _lowestProbW, _numPeriods,
						  _assignMethod,
						  likelyNumTrans);
   
    for (unsigned int j = 0; j < candidateSet.size(); j++) {
      hgc->InsertObject(sameClassRootVect[i], candidateSet[j]);
    //   cout << candidateSet[j] << ", ";
    }
    // cout << endl;
    hgc->AssignObjects();    
    _sameClassMoveMap[sameClassRootVect[i]] = hgc;

    // cout << "root considered in same class = " << sameClassRootVect[i] << endl;

  }
}


int HybridHeatGrowthDep::NextRoot(int currRoot) {

  float randomValue = (float)randu(lp_tt);
  int returnValue;

  // every _R roots selected, moves goto back to random root.
  if ((_R != -1)  && ((_numRoots % (_R + 1)) == 0)) {

    returnValue = _randomDep->NextRoot();
    
  }
  else { 


    if (randomValue <= _depRandomProb) {
      
      returnValue = _randomMove->NextRoot();
    
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb)) {
      
      if (_bySRefMoveMap.find(currRoot) != _bySRefMoveMap.end()) {
	
	returnValue = _bySRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
	returnValue = NULL_OID;
      
      }	
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb)) {
      
      if (_byDRefMoveMap.find(currRoot) != _byDRefMoveMap.end()) {
	
	returnValue = _byDRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
	returnValue = NULL_OID;
      
    }
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb + 
			     _depTraversedProb)) {
      if (_traversedMoveMap.find(currRoot) != _traversedMoveMap.end()) {
	returnValue  = _traversedMoveMap[currRoot]->NextRoot();
      }
      else {      
	vector<int> candidateSet 
	  = _traversedDep->GetCandidateSet(currRoot);

	int likelyNumTrans;
	if (_R != -1) {
	  likelyNumTrans = round(((float) _numTransactions / (float)(_R + 1)) 
		       * ((float) _R) * _depTraversedProb * _highestProbW);
	}
	else {
	  likelyNumTrans =
	    round((float) _numTransactions * _depTraversedProb 
		  * _highestProbW);
	}
		
	HeatGrowthChange * hgc = new HeatGrowthChange(_smallRegionSize,
						      _largeRegionSize,
						      _highestProbW,
						      _lowestProbW, 
						      _numPeriods,
						      _assignMethod,
						      likelyNumTrans);
	cout << "candidate set = ";
	for (unsigned int j = 0; j < candidateSet.size(); j++) {
	  hgc->InsertObject(currRoot, candidateSet[j]);
	  cout << candidateSet[j] << ", ";
	}
	
	cout << endl;
	hgc->AssignObjects();    
	_traversedMoveMap[currRoot] = hgc;
	
	returnValue = _traversedMoveMap[currRoot]->NextRoot();
      }
      
    }
    else {
      
      if (_sameClassMoveMap.find(currRoot) != _sameClassMoveMap.end()) {
	
	returnValue = _sameClassMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
	returnValue = NULL_OID;
	
      }
    }
    
  }
  _numRoots++;

  return returnValue;
  
};



HybridCycleDep::HybridCycleDep(float depRandomProb, float depSRefProb, 
			       float depDRefProb, float depTraversedProb, 
			       float depClassProb, float U, float C, 
			       float regionSize,
			       float highestProbW, float lowestProbW, int H,
			       objAssMeth_t assignMethod, int R,
			       float hotRegionSize, float hotRegionProb) {

  _depRandomProb = depRandomProb;
  _depSRefProb = depSRefProb;
  _depDRefProb = depDRefProb;
  _depTraversedProb = depTraversedProb;
  _depClassProb = depClassProb;

  _bySRefDep = new ByRefDep();
  _byDRefDep = new ByRefDep();
  _traversedDep = new TraversedDep(C);
  _sameClassDep = new SameClassDep(U);

  _regionSize = regionSize;  
  _highestProbW = highestProbW;
  _lowestProbW = lowestProbW;  
  _H = H;
  _assignMethod = assignMethod;
  _R = R;
  
  _randomMove = new CycleOfChange(regionSize, highestProbW, lowestProbW,
				  H, assignMethod);

  _randomDep = new RandomDep(hotRegionSize, hotRegionProb, assignMethod);    

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;
  
  _numRoots = 0;
}


void HybridCycleDep::InsertObjectRandom(int classId, int objId) {
  
  _randomMove->InsertObject(classId, objId);
  _randomDep->InsertObject(classId, objId);

};


void HybridCycleDep::InsertSRef(int rootId, int refId) {
  
  _bySRefDep->InsertRef(rootId, refId);

};

void HybridCycleDep::InsertDRef(int rootId, int refId) {

  _byDRefDep->InsertRef(rootId, refId);

};

void HybridCycleDep::InsertTraversedObject(int rootId,
						   int traversedId) {

  _traversedDep->InsertObject(rootId, traversedId);

};

void HybridCycleDep::InsertSameClassObject(int classId, int objId) {

  _sameClassDep->InsertObject(classId, objId);

};



void HybridCycleDep::Assign() {

  _randomMove->AssignObjects();
  _randomDep->AssignObjects();

  vector<int> refSRootVect = _bySRefDep->GetRootSet();
  for (unsigned int i = 0; i < refSRootVect.size(); i++) {

    set<int> * candidateSet = _bySRefDep->GetCandidateSet(refSRootVect[i]);
    
    CycleOfChange * coc = new CycleOfChange(_regionSize,
					    _highestProbW,
					    _lowestProbW, _H,
					    _assignMethod);

    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      coc->InsertObject(refSRootVect[i], *it);
    }

    coc->AssignObjects();    
    _bySRefMoveMap[refSRootVect[i]] = coc;

  }


  
  // assign objects for DRefDep move window
  vector<int> refDRootVect = _byDRefDep->GetRootSet();
  for (unsigned int i = 0; i < refDRootVect.size(); i++) {

    set<int> * candidateSet = _byDRefDep->GetCandidateSet(refDRootVect[i]);
    
    CycleOfChange * coc = new CycleOfChange(_regionSize,
					    _highestProbW,
					    _lowestProbW, _H,
					    _assignMethod);

    for (set<int>::const_iterator it = candidateSet->begin();
	 it != candidateSet->end(); it++) {
      coc->InsertObject(refDRootVect[i], *it);
    }

    coc->AssignObjects();    
    _byDRefMoveMap[refDRootVect[i]] = coc;
    
  }
  
  _sameClassDep->AssignSets();
  
  // assign objects for SameClassDep move window
  vector<int> sameClassRootVect = _sameClassDep->GetRootSet();
  for (unsigned int i = 0; i < sameClassRootVect.size(); i++) {
    
     //    cout << "candidate set = ";
    vector<int> candidateSet 
      = _sameClassDep->GetCandidateSet(sameClassRootVect[i]);
    
    CycleOfChange * coc = new CycleOfChange(_regionSize,
					    _highestProbW,
					    _lowestProbW, _H,
					    _assignMethod);

   
    for (unsigned int j = 0; j < candidateSet.size(); j++) {
      coc->InsertObject(sameClassRootVect[i], candidateSet[j]);
    //   cout << candidateSet[j] << ", ";
    }
    // cout << endl;
    coc->AssignObjects();    
    _sameClassMoveMap[sameClassRootVect[i]] = coc;

    // cout << "root considered in same class = " << sameClassRootVect[i] << endl;

  }
}



int HybridCycleDep::NextRoot(int currRoot) {

  float randomValue = (float)randu(lp_tt);
  int returnValue;

  // every _R roots selected, moves goto back to random root.
  if ((_R != -1)  && ((_numRoots % (_R + 1)) == 0)) {

    returnValue = _randomDep->NextRoot();
    
  }
  else { 


    if (randomValue <= _depRandomProb) {
      
      returnValue = _randomMove->NextRoot();
    
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb)) {
      
      if (_bySRefMoveMap.find(currRoot) != _bySRefMoveMap.end()) {
	
	returnValue = _bySRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
      returnValue = NULL_OID;
      
      }	
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb)) {
      
      if (_byDRefMoveMap.find(currRoot) != _byDRefMoveMap.end()) {
	
	returnValue = _byDRefMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
      returnValue = NULL_OID;
      
    }
      
    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb + 
			     _depTraversedProb)) {
      if (_traversedMoveMap.find(currRoot) != _traversedMoveMap.end()) {
	returnValue  = _traversedMoveMap[currRoot]->NextRoot();
      }
      else {      
	vector<int> candidateSet 
	  = _traversedDep->GetCandidateSet(currRoot);
		
	
	CycleOfChange * coc = new CycleOfChange(_regionSize,
						_highestProbW,
						_lowestProbW, _H,
						_assignMethod);
	
	cout << "candidate set = ";
	for (unsigned int j = 0; j < candidateSet.size(); j++) {
	  coc->InsertObject(currRoot, candidateSet[j]);
	  cout << candidateSet[j] << ", ";
	}
	
	cout << endl;
	coc->AssignObjects();    
	_traversedMoveMap[currRoot] = coc;
	
	returnValue = _traversedMoveMap[currRoot]->NextRoot();
      }
      
    }
    else {
      
      if (_sameClassMoveMap.find(currRoot) != _sameClassMoveMap.end()) {
	
	returnValue = _sameClassMoveMap[currRoot]->NextRoot();
	
      }
      else {
	
	returnValue = NULL_OID;
	
      }
    }
    
  }
  _numRoots++;

  return returnValue;
  
};

