
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DOCBChangec.h:  DOCB regional protocol (classes)
///////////////////////////////////////////////////////////////////////////
// Access Pattern change for dynamic object clustering benchmark (DOCB)

enum objAssMeth_t {
  random_method, class_method
};

enum moveDir_t {
  down, up
};


const int NULL_INT_PARAM = -1;

class HRegion {
  
 private:
  float _HRegionSize;
  float _initProbW;
  float _lowestProbW;
  float _highestProbW;
  float _probWIncrSize;
  int _probWChangeFreq;
  objAssMeth_t _objectAssignmentMethod;
  moveDir_t _initDir;
  vector< int > _objects;
  float _currProbW;
  moveDir_t _currDir;

 public:
  
  // A probWChangeFreq of NULL_INT_PARAM means this parameter is not used

  HRegion(float HRegionSize, float initProbW, float lowestProbW, 
	  float highestProbW, float probWIncrSize, int probWChangeFreq,
	  objAssMeth_t objectAssignmentMethod, moveDir_t initDir);  
  void InsertObject(int objId);  
  float GetSize();    
  objAssMeth_t GetAssignMethod();
  void MoveCurrProbW();
  void ChangeDir(moveDir_t dir);
  int GetRandomObjId();
  float GetCurrProbW();
  void PrintOids();
  void CycleDir();
  void ToggleDir();
};

// Container that stores all the Hregions and knows the entire scope of
// objects involved in Hregion assigning.

class HRegionContainer {
  
 private:
  vector< HRegion * > _regions;

  // map of objects, indexed by class id
  // since multimap is an assoicative container, its entries are naturally
  // sorted by key which is class id.
  multimap<const int,  int, int_cmp> _objectMap;  

 public:
  HRegionContainer();
  void AssignObjects(long int seed = DEFAULT_SEED);
  void InsertObject(int classId, int objId);
  // returns an id for the Hregion
  int InsertHRegion(HRegion * region);
  void MoveCurrProbW();
  void ChangeDir(int regionId, moveDir_t dir);
  int GetRandomObjId();
  void PrintOids();
  void CycleDir();
  void ToggleDir(int regionId);
};


class GMovingWindowChange {
 private:
  
  HRegionContainer _HRegions;
  int _numRootsProcessed;
  int _H;
  vector< int > _HRegionIds;
  int _currWindowIndex;

 public:

  // set probWIncrSize to highestProbW - lowestProbW normal (non-gradual) 
  // moving window of change

  GMovingWindowChange(float regionSize, float highestProbW, float lowestProbW,
		      int H, objAssMeth_t assignMethod, float probWIncrSize);
  void InsertObject(int classId, int objId);
  void AssignObjects(long int seed = DEFAULT_SEED);
  int NextRoot();
};


class MovingWindowChange {
 private:
  
  HRegionContainer _HRegions;
  int _numRootsProcessed;
  int _H;
  vector< int > _HRegionIds;
  int _currHotIndex;
  
 public:

  // set probWIncrSize to highestProbW - lowestProbW normal (non-gradual) 
  // moving window of change

  MovingWindowChange(float regionSize, float highestProbW, float lowestProbW,
		     int H, objAssMeth_t assignMethod);
  void InsertObject(int classId, int objId);
  void AssignObjects(long int seed = DEFAULT_SEED);
  int NextRoot();
};


class HeatGrowthChange {

 private:

  HRegionContainer _HRegions;
  vector< int > _HRegionIds;
  int _moveFrequency;
  int _numRootsProcessed;
  int _secondRegionId;

 public:

  HeatGrowthChange(float smallRegionSize, float largeRegionSize, 
		   float highestProbW, float lowestProbW, int numPeriods,
		   objAssMeth_t assignMethod, int numTransactions);
  void InsertObject(int classId, int objId);
  void AssignObjects(long int seed = DEFAULT_SEED);
  int NextRoot();
};


class CycleOfChange {

 private:

  HRegionContainer _HRegions;
  vector< int > _HRegionIds;
  int _H;
  int _numRootsProcessed;
  int _region1Id;
  int _region2Id;

 public:

  CycleOfChange(float regionSize, float highestProbW, 
		float lowestProbW, int H, objAssMeth_t assignMethod);
  void InsertObject(int classId, int objId);
  void AssignObjects(long int seed = DEFAULT_SEED);
  int NextRoot();

};
