
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DOCBChangem.h :  DOCB regional protocol (methods)
///////////////////////////////////////////////////////////////////////////

HRegion::HRegion(float HRegionSize, float initProbW, float lowestProbW, 
		 float highestProbW, float probWIncrSize, int probWChangeFreq,
		 objAssMeth_t objectAssignmentMethod, moveDir_t initDir) {
  
  _HRegionSize = HRegionSize;
  _initProbW = initProbW;
  _lowestProbW = lowestProbW;
  _highestProbW = highestProbW;
  _probWIncrSize = probWIncrSize;
  _probWChangeFreq = probWChangeFreq;
  _objectAssignmentMethod = objectAssignmentMethod;
  _initDir = initDir; 

  _currProbW = initProbW;
  _currDir = initDir;
}


void HRegion::InsertObject(int objId) {
  _objects.push_back(objId);
}

float HRegion::GetSize() {
  return _HRegionSize;
}

objAssMeth_t HRegion::GetAssignMethod() {
  return _objectAssignmentMethod;
}


void HRegion::CycleDir() {

  if ((_currDir == up) && (_currProbW >= (_highestProbW - 0.000001))) {
    _currDir = down;
  }
  else if ((_currDir == down) && (_currProbW <= (_lowestProbW + 0.000001))) {
    _currDir = up;
  }
  
}

void HRegion::MoveCurrProbW() { 
  
  //  cout << "----------------" << endl;
  // cout << "initProb = " << _currProbW << endl;

  if (_currDir == up) {
    _currProbW += _probWIncrSize;

    if (_currProbW > _highestProbW) {
      _currProbW = _highestProbW;
    }
  }
  else {
    _currProbW -= _probWIncrSize;
    if (_currProbW < _lowestProbW) {
      _currProbW = _lowestProbW;
    }
  }

  // cout << "dir = " << (int)_currDir << endl;
  // cout << "**** prob = " << _currProbW << endl;
  // cout << "incrSize = " << _probWIncrSize << endl;
  // cout << "----------------" << endl;

}

void HRegion::ChangeDir(moveDir_t dir) {
  _currDir = dir;
}

int HRegion::GetRandomObjId() {
  return _objects[CRand(1, 0, (_objects.size()-1))];
}

float HRegion::GetCurrProbW() {
  return _currProbW;
}

void HRegion::PrintOids() {
    
  for (unsigned int i = 0; i < _objects.size(); i++) {    
    cout << _objects[i] << ", ";
  }
  
}

void HRegion::ToggleDir() {

  if (_currDir == down) {
    _currDir = up;
  }
  else {
    _currDir = down;
  }

}


HRegionContainer::HRegionContainer() {
  
}

void HRegionContainer::InsertObject(int classId, int objId) {
  
  _objectMap.insert(pair<const int, int>(classId, objId));
  
}

int HRegionContainer::InsertHRegion(HRegion * region) {

  _regions.push_back(region);  
  return ((int)_regions.size() - 1);
}


// Fix it:  When have time later, should make it so that all classes where 
//          random numbers a different instance of the random number generator
//          should be created, with the seed as a parameter.  This way
//          benchmark random numbers will be generated irrespective of the
//          rest of the system.

void HRegionContainer::AssignObjects(long int seed) {  

  // firstly check to see if the total size of all Hregions are correct
  float totalSize = 0.0;

  for (unsigned int i = 0; i < _regions.size(); i++) {
    totalSize += _regions[i]->GetSize();
  }

  //if ((totalSize < 0.99999) || (totalSize > 1.000001)) {
  //  cerr << "Hregion sizes do not added up to 1.0 when AssignObjects is called"
  //	 << endl;
  //    cerr << "total Hregion size = " << totalSize << endl;
    // exit(1);
    // }

  // second break Hregions into two sets, one where objects are assigned
  // by class and the other random.
  
  vector< HRegion * > classAssignRegions;
  vector< HRegion * > randomAssignRegions;

  for (unsigned int i = 0; i < _regions.size(); i++) {
    if (_regions[i]->GetAssignMethod() == class_method) {
      classAssignRegions.push_back(_regions[i]);
    }
    else {
      randomAssignRegions.push_back(_regions[i]);
    }
  }

  // assign objects in a class order into hregions which has object 
  // assignment method specified as class_method

  unsigned int totalNumberObj = _objectMap.size();
  unsigned int totalAssigned = 0;

  multimap<const int,  int, int_cmp>::const_iterator objIter;
  objIter = _objectMap.begin();
  HRegion * tempHRegion;
  unsigned int numberObjReq;
  for (unsigned int i = 0; i < classAssignRegions.size(); i++) {
    tempHRegion = classAssignRegions[i];
    numberObjReq = (int) (tempHRegion->GetSize() * (float) totalNumberObj);
    if (numberObjReq < 1 ) {
      numberObjReq = 1;
    }
    for (unsigned int j = 0 ; j < numberObjReq ; j++) {
      tempHRegion->InsertObject(objIter->second);
      totalAssigned++;
      objIter++;
      if (totalAssigned == totalNumberObj) {
	break;
      }     
    }

    if (totalAssigned == totalNumberObj) {
      break;
    }
  }


  // if all objects are to be assigned by class, then remainder objects are
  // assigned to the last tempHRegion, remainder objects may occur due rounding
 
  if ( classAssignRegions.size() == _regions.size() ) {
    while (objIter != _objectMap.end()) {
      tempHRegion->InsertObject(objIter->second);
      totalAssigned++;
      objIter++;
    }
  }

  // now assign remaining objects which are to be assigned using the random
  // method

  vector< int > randomObjs;
  vector< bool > objAssigned;

  while (objIter != _objectMap.end()) {
    randomObjs.push_back(objIter->second);
    objAssigned.push_back(false);
    objIter++;
  }

  // need to fix this up when we have time, need to implement independent
  // random number generators, so that a random seed provided for this 
  // function will not reset the gobal random number generator.
  if (seed > 0) {
    lp_tt = seed;
  }
  else {
    lp_tt = DEFAULT_SEED;
  }
  

  int currObjIndex;
  int numUnAssignedObjs = randomObjs.size();

  for (unsigned int i = 0 ; i < randomAssignRegions.size(); i++) {
    tempHRegion = randomAssignRegions[i];

    numberObjReq = (int)(randomAssignRegions[i]->GetSize() 
			 * (float) totalNumberObj);
    if (numberObjReq < 1) {
      numberObjReq = 1;
    }

    for (unsigned int j = 0 ; j < numberObjReq ; j++) {
      currObjIndex = CRand(1, 0, (numUnAssignedObjs -1));
      while (objAssigned[currObjIndex] == true) {
	currObjIndex = CRand(1, 0, (numUnAssignedObjs - 1));
      }
      objAssigned[currObjIndex] = true;
      tempHRegion->InsertObject(randomObjs[currObjIndex]);
      totalAssigned++;
      if (totalAssigned == totalNumberObj) {
	break;
      }      
    }

    if (totalAssigned == totalNumberObj) {
      break;
    }     
  }  
}

void HRegionContainer::ChangeDir(int regionId, moveDir_t dir) {
  _regions[regionId]->ChangeDir(dir);
}

int HRegionContainer::GetRandomObjId() {

  float randNumber = (float)randu(lp_tt);
  float currProb = 0.0;
  float sumProbW = 0.0;

  
  for (unsigned int i = 0 ; i < _regions.size() ; i++ ) {
    sumProbW += _regions[i]->GetCurrProbW();
  }


  for (unsigned int i = 0 ; i < _regions.size() ; i++ ) {
    currProb += _regions[i]->GetCurrProbW() / sumProbW;
    if (currProb > randNumber) {
      return _regions[i]->GetRandomObjId();
    }
  }
  
  cerr << "ERROR when getting random objId in HRegionContainer " << endl;
  cerr << "sumProbW = " << sumProbW << endl;
  cerr << "randNumber = " << randNumber << endl;
  exit(1);

  return 0;
}

void HRegionContainer::MoveCurrProbW() {
  for (unsigned int i = 0 ; i < _regions.size(); i++ ) {
    _regions[i]->MoveCurrProbW();
  }
}


void HRegionContainer::ToggleDir(int regionId) {
    _regions[regionId]->ToggleDir();
}



void HRegionContainer::CycleDir() {
  for (unsigned int i = 0 ; i < _regions.size(); i++ ) {
    _regions[i]->CycleDir();
  }
}

void HRegionContainer::PrintOids() {
  for (unsigned int i = 0 ; i < _regions.size(); i++ ) {
    _regions[i]->PrintOids(); 
    cout << endl;
  }
}


GMovingWindowChange::GMovingWindowChange(float regionSize, float highestProbW, 
					 float lowestProbW, int H,
					 objAssMeth_t assignMethod,
					 float probWIncrSize) {

  HRegion * newHRegion;
  int numRegions;
  int HRegionId;

  _H = H;
  _numRootsProcessed = 0;


  numRegions = round((1.0 / (float) regionSize));

  //  cout << "number of regions = " << numRegions << endl;  

  _currWindowIndex = 0;

  for (int i = 0; i < numRegions; i++) {    
    
    if (i == 0) {
      newHRegion = new HRegion(regionSize, 
			       highestProbW,
			       lowestProbW, highestProbW, probWIncrSize,
			       NULL_INT_PARAM, assignMethod, down);
    }
    else {
      newHRegion = new HRegion(regionSize, 
			       lowestProbW, 
			       lowestProbW, highestProbW, probWIncrSize,
			       NULL_INT_PARAM, assignMethod, down);
    }

    HRegionId = _HRegions.InsertHRegion(newHRegion);
    _HRegionIds.push_back(HRegionId);
  }

}

void GMovingWindowChange::InsertObject(int classId, int objId) {
  
  _HRegions.InsertObject(classId, objId);

}

void GMovingWindowChange::AssignObjects(long int seed) {
  
  _HRegions.AssignObjects(seed);
}

int GMovingWindowChange::NextRoot() {
  
  _numRootsProcessed++;
  
  if ((_numRootsProcessed % _H) == 0) {
    _currWindowIndex = (_currWindowIndex + 1) % _HRegionIds.size();     
    _HRegions.ToggleDir(_HRegionIds[_currWindowIndex]);    
    _HRegions.MoveCurrProbW();
  }
 
  return _HRegions.GetRandomObjId();
    
}



MovingWindowChange::MovingWindowChange(float regionSize, float highestProbW, 
				       float lowestProbW, int H,
				       objAssMeth_t assignMethod) {

  HRegion * newHRegion;
  int HRegionId;
  float probWIncrSize;
  int numRegions;

  _H = H;
  _numRootsProcessed = 0;
  _currHotIndex = 0;

  probWIncrSize = highestProbW - lowestProbW;

  numRegions = round((1.0 / (float) regionSize));

  //  cout << "number of regions = " << numRegions << endl;

  for (int i = 0; i < numRegions ; i++) {
    
    if (i == 0) {
      newHRegion = new HRegion(regionSize, highestProbW, lowestProbW,
			       highestProbW, probWIncrSize,
			       NULL_INT_PARAM, assignMethod, down);
    }
    else {
      newHRegion = new HRegion(regionSize, lowestProbW, lowestProbW, 
			       highestProbW, probWIncrSize,
			       NULL_INT_PARAM, assignMethod, down);
    }
    HRegionId = _HRegions.InsertHRegion(newHRegion);
    _HRegionIds.push_back(HRegionId);
  }

}

void MovingWindowChange::InsertObject(int classId, int objId) {
  
  _HRegions.InsertObject(classId, objId);

}

void MovingWindowChange::AssignObjects(long int seed) {
  
  _HRegions.AssignObjects(seed);
}

int MovingWindowChange::NextRoot() {
  
  _numRootsProcessed++;
  
  if ((_numRootsProcessed % _H) == 0) {  
      // time to change the hot region to be a different HRegion
    _HRegions.ChangeDir(_HRegionIds[_currHotIndex], down);
    _currHotIndex = (_currHotIndex + 1) % _HRegionIds.size();     
    _HRegions.ChangeDir(_HRegionIds[_currHotIndex], up);    
    _HRegions.MoveCurrProbW();
  }
 
  return _HRegions.GetRandomObjId();
    
}

HeatGrowthChange::HeatGrowthChange(float smallRegionSize,
				   float largeRegionSize, 
				   float highestProbW, float lowestProbW,
				   int numPeriods, objAssMeth_t assignMethod,
				   int numTransactions) {

  HRegion * newHRegion;
  int HRegionId;
  float region1Size =  smallRegionSize;
  float region2Size =  largeRegionSize - smallRegionSize;
  float region3Size =  1.0 - largeRegionSize;
  float probWIncrSize;

  probWIncrSize = highestProbW - lowestProbW;
  
  // make sure numPeriods divides numTransactions
  if ((numTransactions % numPeriods) != 0) {
    numPeriods = numPeriods - (numTransactions % numPeriods);
  }

  _moveFrequency = numTransactions / numPeriods;
  //  cout << "_moveFreq = " << _moveFrequency << endl;

  // first region, region which is always hot
  newHRegion = new HRegion(region1Size, highestProbW, highestProbW,
			   highestProbW, 0, NULL_INT_PARAM, assignMethod, up);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _HRegionIds.push_back(HRegionId);

  // second region, region which is sometimes hot and sometimes cold
  // The second region is the region responsible for growing and shrinking
  // of the hot region. When the second region is at lowestProbW the hot 
  // region shrinks, when it is at the hottestProbW the hot region expands.

  newHRegion = new HRegion(region2Size, lowestProbW, lowestProbW,
			   highestProbW, probWIncrSize,
			   NULL_INT_PARAM, assignMethod, up);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _secondRegionId = HRegionId;
  _HRegionIds.push_back(HRegionId);

  // third region, region which is always cold

  newHRegion = new HRegion(region3Size, lowestProbW, lowestProbW,
			   lowestProbW, 0, NULL_INT_PARAM, assignMethod, up);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _HRegionIds.push_back(HRegionId);

  _numRootsProcessed = 0;
}

void HeatGrowthChange::InsertObject(int classId, int objId) {

  _HRegions.InsertObject(classId, objId);

}

void HeatGrowthChange::AssignObjects(long int seed = DEFAULT_SEED) {

    _HRegions.AssignObjects(seed);

}

int HeatGrowthChange::NextRoot() {  
  
  _numRootsProcessed++;
  
  if ((_numRootsProcessed % (_moveFrequency + 1)) == 0) {  
    _HRegions.MoveCurrProbW();
    // toggle the direction of the second region.  The second region is the
    // region responsible for growing and shrinking of the hot region.
    // When the second region is at lowestProbW the hot region shrinks,
    // when it is at the hottestProbW the hot region expands.
    _HRegions.ToggleDir(_secondRegionId);
  }

  return _HRegions.GetRandomObjId();    
  
}

CycleOfChange::CycleOfChange(float regionSize, float highestProbW, 
			     float lowestProbW, int H, 
			     objAssMeth_t assignMethod) {
  HRegion * newHRegion;
  int HRegionId;
  float region1Size =  regionSize;
  float region2Size =  regionSize;
  float region3Size = 1.0 - (region1Size * 2.0);
  float probWIncrSize = highestProbW - lowestProbW;
  
  _H = H;

  // first small region, region which is sometimes hot and sometimes cold

  newHRegion = new HRegion(region1Size, highestProbW, lowestProbW,
			   highestProbW, probWIncrSize, NULL_INT_PARAM, 
			   assignMethod, down);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _region1Id = HRegionId;
  _HRegionIds.push_back(HRegionId);
  
  // second small region, region which is sometimes hot and sometimes cold
  
  newHRegion = new HRegion(region2Size, lowestProbW, lowestProbW,
			   highestProbW, probWIncrSize,
			   NULL_INT_PARAM, assignMethod, up);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _region2Id = HRegionId;
  _HRegionIds.push_back(HRegionId);
  
  // third large region, region which is always cold

  newHRegion = new HRegion(region3Size, lowestProbW, lowestProbW,
			   lowestProbW, 0, NULL_INT_PARAM, assignMethod, up);
  
  HRegionId = _HRegions.InsertHRegion(newHRegion);
  _HRegionIds.push_back(HRegionId);
  

  _numRootsProcessed = 0;  
}

void CycleOfChange::InsertObject(int classId, int objId) {

  _HRegions.InsertObject(classId, objId);

}

void CycleOfChange::AssignObjects(long int seed = DEFAULT_SEED) {

    _HRegions.AssignObjects(seed);

}

int CycleOfChange::NextRoot() {  
  

  if ((_numRootsProcessed % (_H)) == 0) {  
    _HRegions.ToggleDir(_region1Id);
    _HRegions.ToggleDir(_region2Id);
    _HRegions.MoveCurrProbW(); 
  }
  
  _numRootsProcessed++;

  return _HRegions.GetRandomObjId();    
  
}
