#include <hash_set.h>

RandomDep::RandomDep(float hotRegionSize, float hotRegionProb, 
		     objAssMeth_t assignMethod) {

  _hotRegionSize = hotRegionSize;
  _hotRegionProb = hotRegionProb;
  _assignMethod = assignMethod;
  _classObjMap = new multimap<const int, int, int_cmp>();
  _setObjs = new set<int>();

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;
  
};


void RandomDep::InsertObject(int classId, int objId) {
  
  _classObjMap->insert(pair<const int, int>(classId, objId));

};


void RandomDep::AssignObjects() {

  multimap<const int, int, int_cmp>::iterator it;

  for (it = _classObjMap->begin(); it != _classObjMap->end(); it++) {
    
    _setObjs->insert(it->second);

  }

  if (_assignMethod == class_method) {
    
    int hotSize = (int)(_hotRegionSize * (float) _classObjMap->size());

    it = _classObjMap->begin();

    for (int i = 0; i < hotSize; i++) {

      _hotObjects.push_back( it->second );
      it++;
      
    }

    while (it != _classObjMap->end()) {

      _coldObjects.push_back( it->second );
      it++;

    }

  }  
  else if (_assignMethod == random_method) {

    vector<bool> hotObjectBool;
    int hotSize = (int)(_hotRegionSize * (float) _classObjMap->size());

    for (unsigned int i =0 ; i < _classObjMap->size(); i++) {
      hotObjectBool.push_back(false);
    }

    int numAssigned = 0;
    int chosenIndex;
    int sizeMinus1 = _classObjMap->size() - 1;

    while (numAssigned < hotSize) {
      
      chosenIndex = CRand(1, 0, sizeMinus1 );
      
      if (hotObjectBool[chosenIndex] == false) {    
	hotObjectBool[chosenIndex] = true;
	numAssigned++;
      }

    }

    int index = 0;
    for (it = _classObjMap->begin(); it != _classObjMap->end(); it++) {

      if (hotObjectBool[index] == true) {

	_hotObjects.push_back(it->second);
	
      }
      else {

	_coldObjects.push_back(it->second);
	
      }
      index++;
    }
    
  }
  else {
    
    cerr << "This assignment method has not been implemented yet in ";
    cerr << "DOCBDepm.h, RandomDep"<< endl;
    exit(1);

  }
  
  delete _classObjMap;

};


int RandomDep::NextRoot() {

  float regionSelectRandValue = (float)randu(lp_tt);

  if ( regionSelectRandValue <= _hotRegionProb) {

    // should select and object from the hot region.
    int selectedIndex = CRand(1, 0, ((int)_hotObjects.size() - 1));
    return _hotObjects[selectedIndex]; 
 
  }
  else {

    // an object from the cold region should be selected.
    int selectedIndex = CRand(1, 0, ((int)_coldObjects.size() - 1));
    return _coldObjects[selectedIndex];

  }

};

set<int> * RandomDep::GetCandidateSet() {  

  return _setObjs;

};


RandomDep::~RandomDep() {
  
  delete _setObjs;
  delete _classObjMap;

};


ByRefDep::ByRefDep() {

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

};

void ByRefDep::InsertRef(int rootId, int refId) {
  
  if (_refs.find(rootId) == _refs.end()) {
    set<int> * newRefs = new set<int>();
    newRefs->insert(refId);
    _refs[rootId] = newRefs;
  }
  else {
    _refs[rootId]->insert(refId);
  }
  
};


int ByRefDep::NextRoot(int currRoot) { 

  int returnId;

  if (_refs.find(currRoot) == _refs.end()) {
    returnId = NULL_OID;
  }
  else {
    int numRefs;
    set<int> * tempSet;
    int chosenIndex;

    tempSet = _refs[currRoot];
    numRefs = tempSet->size();
    chosenIndex = CRand(1, 0, (numRefs - 1));

    set<int>::const_iterator it = tempSet->begin();

    for (int i = 0; i < chosenIndex; i++) {
      it++;
    }

    returnId = *it;
  }
  
  return returnId;

};

set<int> * ByRefDep::GetCandidateSet(int rootId) {
  
  if (_refs.find(rootId) == _refs.end()) {
    return NULL;
  }
  else {
    return _refs[rootId];
  }

};

vector<int> ByRefDep::GetRootSet() {

  vector<int> rootSet;
  hash_map<int, set<int> * , hash<int>, eqoid>::const_iterator it;

  for (it = _refs.begin(); it != _refs.end(); it++) {

    rootSet.push_back(it->first);

  }
  
  return rootSet;

}

ByRefDep::~ByRefDep() {
  
  hash_map<int, set<int> * , hash<int>, eqoid>::iterator it;
  
  for (it = _refs.begin(); it != _refs.end(); it++) {
    delete it->second;
  }

};


TraversedDep::TraversedDep(float C) {

  _currRoot = NULL_OID;
  _C = C;

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

};

void TraversedDep::InsertObject(int rootId, int traversedId) {
  if ((rootId == NULL_OID) || (rootId != _currRoot)) {
    _traversedObjects.clear();
    _currRoot = rootId;
  }
  _traversedObjects.insert(pair<const int, int>(traversedId, traversedId));
};

int TraversedDep::NextRoot(int currRoot) {
 
  int returnId;
  
  if ((currRoot != _currRoot) || ( _traversedObjects.size() == 0 )) {
    if (_traversedObjects.size() == 0) {
      cout << "number of traversed objects equal 0" << endl;
    }
    returnId = NULL_OID;
  }
  else {
    int numTraversed;
    int chosenIndex;
    int candidateSize;

    numTraversed = _traversedObjects.size();
    candidateSize = (int)(_C * (float) numTraversed);
    if (candidateSize <= 0) {
      candidateSize = 1;
    }
    chosenIndex = CRand(1, 0, (candidateSize - 1));

    multimap<const int, int, int_cmp>::iterator it = _traversedObjects.begin();
    for (int i = 0; i < chosenIndex; i++) {
      it++;
    }    

    returnId = it->first;   
  }
  
  _traversedObjects.clear();
  _currRoot = NULL_OID;

  return returnId;
};


// note: to be able to use this function for moving access pattern: must be 
//       using traversals that always traverse the same set of objects
//       for the same root object.
vector<int> TraversedDep::GetCandidateSet(int currRoot) {

  vector<int> returnSet;

  if (currRoot == _currRoot) {
    int numTraversed;
    int candidateSize;

    numTraversed = _traversedObjects.size();
    candidateSize = (int)(_C * (float) numTraversed);
    if (candidateSize <= 0) {
      candidateSize = 1;
    }

    if (_traversedObjects.size() > 0) {
      multimap<const int, int, int_cmp>::iterator it 
	= _traversedObjects.begin();
      returnSet.push_back(it->first);
      for (int i = 0; i < (candidateSize - 1); i++) {
	it++;
	returnSet.push_back(it->first);
      } 
    }
  }
  

  _traversedObjects.clear();
  _currRoot = NULL_OID;

  return returnSet;
};


SameClassDep::SameClassDep(float U) {

  _U = U;

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

};


void SameClassDep::InsertObject(int classId, int objId) {

  _classToObjMap.insert(pair<int, int>(classId, objId));

};

void SameClassDep::AssignSets() {
 
  multimap<int, int, int_cmp>::const_iterator it;
  int currClass;
  int classSize;
  int currObjId;
  int classId;
  int setStartIndex;

  vector<int> sameClassVect;

  it = _classToObjMap.begin();

  while (it != _classToObjMap.end()) {
        
    currClass = it->first;
    classSize = 0;
    classId = it->first;
    sameClassVect.clear();

    while ((currClass == it->first) && (it != _classToObjMap.end())) {
      classSize++;
      sameClassVect.push_back(it->second);
      it++;
    }       
    if (classSize > 0) {
      _classSizeMap[classId] = classSize;
    }
    /*    // round up to 1 any set sizes of 0.
	  if (setSize == 0) {
	  setSize = 1;
	  }
    */

    // For each object of this class assign a set of same class objects that
    // form the candidate set for that object.
    for (int i = 0; i < classSize; i++) {
      currObjId = sameClassVect[i];
      setStartIndex = CRand(1, 0, (classSize - 1));
      _setStartIndexMap[currObjId] = setStartIndex;
      _objToClassMap[currObjId] = classId;
    }
  }

};

int SameClassDep::NextRoot(int currRoot) {

  int returnId;

  if (_setStartIndexMap.find(currRoot) == _setStartIndexMap.end()) {
    returnId = NULL_OID;
  }
  else {
    int setStartIndex;
    int classSize;
    int setSize;
    int classId;
    int chosenIndex;

    setStartIndex = _setStartIndexMap[currRoot];
    classId = _objToClassMap[currRoot];
    classSize = _classSizeMap[classId];
    setSize = (int)((float) classSize * _U);
    chosenIndex = (setStartIndex + CRand(1, 0, (setSize - 1))) % classSize;

    multimap<int, int, int_cmp>::const_iterator it;
    it = _classToObjMap.lower_bound(classId);

    for (int i = 0; i < chosenIndex; i++) {
      it++;
    }

    returnId = it->second; 
  }
  
  return returnId;

};

vector<int> SameClassDep::GetCandidateSet(int currRoot) {
  
  vector<int> returnSet;

  if (_setStartIndexMap.find(currRoot) != _setStartIndexMap.end()) {
    int setStartIndex;
    int classSize;
    int setSize;
    int classId;

    setStartIndex = _setStartIndexMap[currRoot];
    classId = _objToClassMap[currRoot];
    classSize = _classSizeMap[classId];
    setSize = (int)((float) classSize * _U);

    multimap<int, int, int_cmp>::const_iterator it;
    it = _classToObjMap.lower_bound(classId);
    int currIndex = 0;
    for (int i = 0; i < setStartIndex; i++) {
      it++;
      currIndex++;
    }

    for (int i = 0; i < setSize; i++) {      
      if (currIndex >= (classSize)) {
	currIndex = 0;
	it = _classToObjMap.lower_bound(classId);
      }      
      returnSet.push_back(it->second);
      currIndex++;
      it++;
    }
  }

  return returnSet;
};

vector<int> SameClassDep::GetRootSet() {

  vector<int> rootSet;
  hash_map<int, int , hash<int>, eqoid>::const_iterator it;

  for (it = _setStartIndexMap.begin(); it != _setStartIndexMap.end(); it++) {

    rootSet.push_back(it->first);

  }
  
  return rootSet;

}


HybridDep::HybridDep(float depRandomProb, float depSRefProb, float depDRefProb,
		     float depTraversedProb, float depClassProb, float U,
		     float C, float randHotRegSize, float randHotRegionProb,
		     objAssMeth_t randAssMethod, int R) {

  _depRandomProb = depRandomProb;
  _depSRefProb = depSRefProb;
  _depDRefProb = depDRefProb;
  _depTraversedProb = depTraversedProb;
  _depClassProb = depClassProb;

  _randomDep = new RandomDep(randHotRegSize, randHotRegionProb, randAssMethod);
  _bySRefDep = new ByRefDep();
  _byDRefDep = new ByRefDep();
  _traversedDep = new TraversedDep(C);
  _sameClassDep = new SameClassDep(U);

  // reset random seed if needed
  if (lp_tt <= 0)
    lp_tt = DEFAULT_SEED;

  _R = R;
  _numRoots = 0;
};

void HybridDep::InsertSRef(int rootId, int refId) {

  _bySRefDep->InsertRef(rootId, refId);

};

void HybridDep::InsertDRef(int rootId, int refId) {
  
  _byDRefDep->InsertRef(rootId, refId);
  
};

void HybridDep::InsertTraversedObject(int rootId, int traversedId) {

  _traversedDep->InsertObject(rootId, traversedId);
  
};

void HybridDep::InsertSameClassObject(int classId, int objId) {
  
  _sameClassDep->InsertObject(classId, objId);

};

void HybridDep::InsertObjectRandom(int classId, int objId) {
  
  _randomDep->InsertObject(classId, objId);

};

void HybridDep::Assign() {

  _randomDep->AssignObjects();
  _sameClassDep->AssignSets();

}; 

int HybridDep::NextRoot(int currRoot) {

  float randomValue = (float)randu(lp_tt);
  int returnValue;
  
  if ((_R != -1) && (_numRoots % (_R + 1)) == 0) {

    returnValue = _randomDep->NextRoot();

  }
  else {

    if (randomValue <= _depRandomProb) {
      
    returnValue =  _randomDep->NextRoot();

    }
    else if (randomValue <= (_depRandomProb + _depSRefProb)) {
      
      returnValue = _bySRefDep->NextRoot(currRoot);

    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb)) {
      
      returnValue = _byDRefDep->NextRoot(currRoot);

    }
    else if (randomValue <= (_depRandomProb + _depSRefProb + _depDRefProb + 
			     _depTraversedProb)) {
      returnValue = _traversedDep->NextRoot(currRoot);
    
    }
    else {
      
      returnValue = _sameClassDep->NextRoot(currRoot);
    
    }
  }

  _numRoots++;

  return returnValue;

}


HybridDep::~HybridDep() {

  delete _randomDep;
  delete _bySRefDep;
  delete _byDRefDep;
  delete _traversedDep;
  delete _sameClassDep;
  
};
