
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// DROStatm.h:  DRO statistics (methods)
///////////////////////////////////////////////////////////////////////////

DROObjStats::DROObjStats(int objId) {
  _objId = objId;
  _accessFreq = 0;
  _usageIndicator = false;
}

void DROObjStats::setUsageIndicator(bool ui) {
  _usageIndicator = ui;
}

bool DROObjStats::getUsageIndicator() {
  return _usageIndicator;
}

void DROObjStats::resetUsageIndicator() {
  _usageIndicator = false;
}

void DROObjStats::incAccessFreq() {
  _accessFreq++;
}

int DROObjStats::getAccessFreq() {
  return _accessFreq;
}

int DROObjStats::getOid() {

  return _objId;

}

void DROObjStats::resetStats() {

  _accessFreq = 0;
  _usageIndicator = false;
  
}


DROPageStats::DROPageStats(int pageId) {

  _pageId = pageId;
  _nbLoad = 0;
}

DROPageStats::~DROPageStats() {

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {

    delete it->second;

  }

}

void DROPageStats::incNbLoad() {

  _nbLoad++;

}

int DROPageStats::getNbLoad() {

  return _nbLoad;
  
}

DROObjStats * DROPageStats::getObjStats(int objId) {

  if (_objStats.find(objId) != _objStats.end()) {
    
    return _objStats[objId];

  }
  else {

    return NULL;
    
  }

}

bool DROPageStats::deleteObjStats(int objId) {

  hash_map< int, DROObjStats *, hash<int>, eqoid >::iterator it;

  it = _objStats.find(objId);
  if (it != _objStats.end()) {

    delete _objStats[objId];
    _objStats.erase(it);

    return true;
    
  }
  else {

    return false;

  }
}

bool DROPageStats::insertObject(int objId) {

  if (_objStats.find(objId) == _objStats.end()) {
    
    _objStats[objId] = new DROObjStats(objId);
    
    return true;
    
  }
  else {

    return false;

  }

}

bool DROPageStats::objExists(int objId) {

  if (_objStats.find(objId) != _objStats.end()) {
    return  true;
  }
  else {
    return false;
  }
}

bool DROPageStats::setObjUsageIndicator(int objId, bool ui) {

  if (_objStats.find(objId) != _objStats.end()) {
    _objStats[objId]->setUsageIndicator(ui);
    return true;
  }
  else {
    return false;
  }

}

bool DROPageStats::getObjUsageIndicator(int objId) {

  return _objStats[objId]->getUsageIndicator();

}


bool DROPageStats::incObjAccessFreq(int objId) {

  if (_objStats.find(objId) != _objStats.end()) {

    _objStats[objId]->incAccessFreq();
    return true;
    
  }
  else {

    return false;
    
  }
  
}


int DROPageStats::getObjAccessFreq(int objId) {

  return _objStats[objId]->getAccessFreq();

}

float DROPageStats::computeUsageRate() {
  int objSize;
  int currOid;
  int effectiveSize = 0;
  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {

    if (it->second->getUsageIndicator() == true) {
      currOid = it->second->getOid();
      objSize = (DB->GetObj(currOid))->ClassPtr()->GetSize();
      effectiveSize += objSize;      
    }
    
  }

  return (float) effectiveSize / (float) PGSIZE;

}

void DROPageStats::resetAllObjUsageIndicator() {

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {

    it->second->setUsageIndicator(false);

  }

}

int DROPageStats::getPid() {
  return _pageId;
}

vector<int> DROPageStats::getUsedObjects() {
  vector<int> returnVect;
  DROObjStats * objS;

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {
    objS = it->second;

    if (objS->getUsageIndicator() == true) {
      returnVect.push_back(it->first);
    }

  }

  return returnVect;

}


vector<int> DROPageStats::getUnusedObjects() {
  vector<int> returnVect;
  DROObjStats * objS;

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {
    objS = it->second;

    if (objS->getUsageIndicator() == false) {
      returnVect.push_back(it->first);
    }

  }

  return returnVect;

}

void DROPageStats::resetStats() {

  _nbLoad = 0;

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {

    it->second->resetStats();

  }

}

vector<int> DROPageStats::getAllObjects() {

  vector<int> returnVect;

  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;

  for (it = _objStats.begin(); it != _objStats.end(); it++) {

    returnVect.push_back(it->first);

  }

  return returnVect;

}

void DROPageStats::resetNbLoad() {
  _nbLoad = 0;
}

vector<DROObjStats *> DROPageStats::removeAllObjStats() {

  vector<DROObjStats *> returnStats;

  
  hash_map< int, DROObjStats *, hash<int>, eqoid >::const_iterator it;
  
  for (it = _objStats.begin(); it != _objStats.end(); it++) {
    
    returnStats.push_back(it->second);

  }

  
  _objStats.clear();

  return returnStats;

}

void DROPageStats::insertObjStats(DROObjStats * stats) {

  _objStats[stats->getOid()] = stats;

}
