
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database)
// Modele d'evaluation de performances generique pour SGBDOO
// Version 2.0 C++, avril 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
// basec.h : Definition de la base de donnees (Classes)
/////////////////////////////////////////////////////////////////////

class Database;
class Class;
class Object;
class Page;
class BuffMgr;
class AnalyPage;

#include<vector.h>

/////////////////////////////////////////////////////////////////////
// CLASSE Database
/////////////////////////////////////////////////////////////////////


class Database {

  public:                          // Methodes

    Database();                    // Constructeur
    ~Database();                   // Destructeur
    void Generation();             // Generation de la base
    void Display();                // Affichage de la base
    void Initp();                  // Placement initial des objets
    void SeqPl();                  // Placement initial sequentiel
    void OSeqPl();                 // Placement initial sequentiel optimise
    void OtherPl();                // Placement initial autre
    Class *GetClass(int i);        // Retourne la classe numero I
    Object *GetObj(int i);         // Retourne l'objet numero I
    Page *NewPage();               // Creation nouvelle page
    void AddPage(Page *page);      // Ajout page a la liste
    Page *GetPage(int num);        // Retourne un ptr sur la page num
    void DelPage(int num);         // Suppression d'une page
    void PurgePages();             // Suppression de toutes les pages
    void DisplayPages();           // Affichage des pages utilisees
    int NbPages();                 // Retourne le nombre de pages utilisees
    
    // Attributs utilises par la strategie de regroupement DSTC
    // Rendus publics pour facilite d'utilisation
    
    // mo stores reference count info.
    // mo[i][0] = stores the object on the other side of the link from object i
    // mo[i][1] = stores the number of times the link is traversed
    // ChList * mo[MNO][2]; 
    vector<ChList ** >mo;                 // Matrice d'observation MO
                                          // 0 - indice "j"
                                          // 1 - valeur compteur
    
    bool LinkExists(int prev_oid, int curr_oid);
  
    // vo[i] stores the number of times object i has been traversed
    // int vo[MNO]
    vector<int> vo;                        // Vecteur d'observation VO
    // ChList * mc[MNO][4]; 
    vector<ChList **>mc;                // Matrice consolidee MC
                                        // 0 - indice "j"
                                        // 1 - valeur facteur fc(i,j)
                                        // 2 - date de mise a jour
                                        // 3 - flag groupement
    int pi;                             // Periode courante
    int ni;                             // Nombre d'entrees courantes dans MO

    // Methodes DSTC

    void NewPeriod();                   // Nouvelle periode d'observation
    void DisplayMO();                   // Affichage matrice MO
    void DisplayMC();                   // Affichage matrice MC
    
    int InsertNewObj();
    void PlaceNewObj( int new_oid );
    bool ValidOid(int i_oid);
    int GetRandomOid(int dist, int * past_non_reg = NULL);
    void DeleteObj(int del_oid);
    void InsertRef(int * obj_oid, int dist);
    void DisplayNumRef();
    void DeleteRef(int * obj_id, int dist);
    float CalcAvgObjSize();
    int CalcMinObjSize();
    int CalcMaxObjSize();
    bool PageEmpty(int pg_num);

    void MoveDepAssignObjects();
    
    vector<Page *> GetPageList();

  private:                      // Attributs

    int nbclasses;              // Nombre de classes
    //Class * classtable[MNC]; // Index des classes
    vector<Class *>  classtable;// Index des classes
    int nbobj;                  // Nombre d'objets
    //Object * objtable[MNO];   // Index des instances
    vector<Object *> objtable;  // Index des instances
    int nbpages;                // Nombre de pages
    Page *pagelist;             // Pointeur sur a liste chainee des pages
    int cumulp;                 // Nombre de pages crees (=> id uniques)
    short Inheritance[MNREFT];  // Indicateur hierarchie d'heritage
    short NoCycle[MNREFT];      // Indicateur hierarchie cyclique
    
    Page * cur_page;
    void _OSeqPl(int new_oid);
    void _SeqPl(int new_oid);
    void _OtherPl(int new_oid);        
};

/////////////////////////////////////////////////////////////////////
// CLASSE Class
/////////////////////////////////////////////////////////////////////

class Class {

  public:                       // Methodes

    Class(int i, short maxn, int bs); // Constructeur
    ~Class();                   // Destructeur
    short MaxNRef();            // Retourne maxnref
    int BaseSize();             // Retourne basesize
    void SetTRef(int i, short t); // Initialise tref[i] a t
    short GetTRef(int i);       // Retourne tref[i]
    void SetCRef(int i, Class *c); // Initialise cref[i] a c
    Class *GetCRef(int i);      // Retourne cref[i]
    ChList *Iterator();         // Retourne l'adresse de l'iterateur
    void SetSize(int s);        // Initialise instsize a s
    int GetSize();              // Retourne intsize
    int Id();                   // Retourne id
    void Display();             // Affichage ecran
    bool InitialRefExists(int ref);
    void SetInitialRef(int ref, bool ire);

  private:                      // Attributs

    int id;                     // Identificateur classe
    short maxnref;              // Nombre maxi de references
    int basesize;               // Taille de base des instances
    short tref[MMAXNREF];       // Table des types de reference
    Class *cref[MMAXNREF];      // Table des classes referencees
    ChList *iterator;           // Iterateur (liste des indices d'instances)
    int instsize;               // Taille des instances
    bool InitialRefEx[MMAXNREF]; // array that stores the initial reference
                                 // existence information, used so at object
                                 // we now when a new object should have in
                                 // terms of references to other objects
};


/////////////////////////////////////////////////////////////////////
// CLASSE Object
/////////////////////////////////////////////////////////////////////

class Object {

  public:                       // Methodes

    Object(int id, Class *c);   // Constructeur
    ~Object();                  // Destructeur
    Class *ClassPtr();          // Retourne la classe de l'objet
    void SetORef(int i, Object *o); // Initialise oref[i] a o
    Object *GetORef(int i);     // Retourne oref[i]
    ChList *BackRef();          // Retourne l'adresse des refs inverses
    ChList *PageDisk();         // Retourne la liste des pages contenant l'objet
    int OID();                  // Retourne l'OID
    void Display();             // Affichage ecran

    void DeleteAllPageDisks();
    void CreateNewPageDisk();
    void InsertPage(int pg_num);

    private:                    // Attributs

    int oid;                    // OID
    Class *classptr;            // Classe de l'objet
    Object *oref[MMAXNREF];     // Table des references d'objet
    ChList *backref;            // Liste des indices des references inverses
    ChList *pagedisk;           // Liste des pages contenant l'objet
  
};

/////////////////////////////////////////////////////////////////////
// CLASSE Page
/////////////////////////////////////////////////////////////////////

class Page {

  public:                       // Methodes

    Page(int num);              // Constructeur
    ~Page();                    // Destructeur
    void SetNext(Page *n);      // Initialisation pointeur next
    int PageNum();              // Retourne le numero de page
    int Free();                 // Retourne la place libre
    void ResetFree(int f);      // Reinitialise la place libre
    ChList *Objects();          // Retourne la liste des objets
    Page *Next();               // Retourne la page suivante
    int CalcExtTension(BuffMgr * bmgr);       // Calculate the total external tension of the 
                                // page
    int CalcHeat();             // Calculate the total heat of the page

    float CalcWaste();

    void InsertObj(int i_oid);
    void DeleteAllObjs();
    void CreateNewObjList();
    bool DeleteObj(int del_oid);
    bool PageEmpty();
    void ReCalcFree();

  private:                      // Attributs

    int pagenum;                // Numero de la page
//  int sector;                 // Numero de secteur
//  int cyl;                    // Numero de cylindre
    int free;                   // Place libre dans la page
    ChList *objects;            // Objets contenus dans la page
    Page *next;                 // Page suivante dans la liste

};


