
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database)
// Modele d'evaluation de performances generique pour SGBDOO
// Version 2.0 C++, avril 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// basem.h : Definition de la base de donnees (Methodes)
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// CLASSE Database
/////////////////////////////////////////////////////////////////////

// CLASSE Database : Constructeur

#include<hash_set.h>


Database::Database() {
  ChList ** newChListPtrPtr;

  nbclasses=0;
  nbobj=0;
  nbpages=0;
  cumulp=0;
  pagelist=NULL;

  // DSTC

  int i, j;
  
  NC = MNC;
  NO = MNO;
  pi=0;                                 // Premiere periode
  ni=0;
  // Initialisation MO + MC
  for (i=0; i<NO; i++) {
    newChListPtrPtr = new (ChList*) [2];
    mo.push_back(newChListPtrPtr);
    newChListPtrPtr = new (ChList*) [4];
    mc.push_back(newChListPtrPtr);
    vo.push_back(0);
    for (j=0; j<2; j++)
      mo[i][j]=new ChList;
    for (j=0; j<4; j++)
      mc[i][j]=new ChList;
  }

}

// CLASSE Database : Destructeur

Database::~Database() {

  int i, j;

  for (i=0; i<nbclasses; i++) delete classtable[i];
  for (i=0; i<nbobj; i++) delete objtable[i];
  PurgePages();
  for (i=0; i<NO; i++) {
    for (j=0; j<2; j++) {
      delete mo[i][j];
    }
    for (j=0; j<4; j++)
      delete mc[i][j];
  }

}

// CLASSE Database : Generation de la base

void Database::Generation() {

  int i, j, k, l, count;
  short pb;
  ChList *pile;
  Class *cclass;
  int selected_oid;


  
#ifdef CHANGE_SEED
  long int prev_seed;

  prev_seed = lp_tt;   
  lp_tt = DB_SEED;
#endif

  for (int i =0; i < NC; i++) {
    classtable.push_back(NULL);
  }
  for (int i =0; i < NO; i++) {
    objtable.push_back(NULL);
  } 

//  printf("\nDatabase generation...\n");

  // Etape preliminaire :
  // initialisation des indicateurs de hierarchie cyclique
  // et des indicateurs de hierachie d'heritage

  NoCycle[0]=1;                         // RefType #0 => heritage
  Inheritance[0]=1;
  for (i=1; i<NREFT; i++) {
    NoCycle[i]=0;
//    NoCycle[i]=VALEUR A FIXER;
    Inheritance[i]=0;
//    Inheritance[i]=VALEUR A FIXER;
  }

  // Etape 1 : Generation du schema - classes

  nbclasses=0;
  for (i=0; i<NC; i++) {
    // Instanciation de la classe Class
    classtable[i]=new Class(i,MAXNREF,BASESIZE);
    nbclasses++;
    // Determination du type de reference pour chaque reference
    // Assign a type to each class reference
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      classtable[i]->SetTRef(j,CRand(DIST1,0,NREFT-1));
  }

  // Etape 2 : Liens inter-classes

  for (i=0; i<NC; i++)
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      // Affectation des references
      // Assign which class this class references
      classtable[i]->SetCRef(j,classtable[CRand(DIST2,INFCLASS-1,SUPCLASS-1)]);
//      classtable[i]->SetCRef(j,classtable[VALEUR A FIXER]);

  // Etape 3 : Coherence des graphes acycliques

  for (i=0; i<NC; i++)
    for (j=0; j<classtable[i]->MaxNRef(); j++)
      // If the reference type is NoCycle (inheritance) reference type.
      if ((NoCycle[classtable[i]->GetTRef(j)])
	  && (classtable[i]->GetCRef(j)!=NULL)) {
        // Parcours de la hierarchie de la classe CRef(j)
        pile=new ChList;
        pile->Insert(1,classtable[i]->GetCRef(j)->Id());
        count=1;                        // Compteur cycle
        pb=0;                           // Pas de probleme
	// If after looking down the graph of objects from the 
	// root object (classtable[i]) you find that you come back to
	// the root object then a cycle exists.
	
        while ((pile->NbElem()) && (!pb)) {
          cclass=classtable[pile->Scan(1)];
          pile->Delete(1);
          count++;
          if ((cclass==classtable[i]) || (count>=NC))
            // Classe dans hierarchie ou cycle detecte
            pb=1;
	  //          else
	  // On empile les references correspondantes non nulles
	  for (k=0; k<cclass->MaxNRef(); k++)
	    if ((cclass->GetTRef(k)==classtable[i]->GetTRef(j))
		&& (cclass->GetCRef(k)!=NULL))
	      pile->Insert(1,cclass->GetCRef(k)->Id());
	}
        delete pile;
	// if setting a reference from i to j causes a cycle then set
	// the reference from i to j to NULL.
        if (pb) classtable[i]->SetCRef(j,NULL);
        else if (Inheritance[classtable[i]->GetTRef(j)]) { // heritage
          // MaJ taille des instances dans le graphe d'heritage
          pile=new ChList;
          pile->Insert(1,classtable[i]->GetCRef(j)->Id());
          while (pile->NbElem()) {
            cclass=classtable[pile->Scan(1)];
            pile->Delete(1);
	    // This adds base class size to each instance in the inheritance
	    // hierarchy.  
            cclass->SetSize(cclass->GetSize()+classtable[i]->BaseSize());
            for (k=0; k<cclass->MaxNRef(); k++)
	      // if reference type is same type of reference ( inheritence )
	      // as the root reference ( classtable[i]->GetCRef(j) ).
              if ((cclass->GetTRef(k)==classtable[i]->GetTRef(j))
		  && (cclass->GetCRef(k)!=NULL))
                pile->Insert(1,cclass->GetCRef(k)->Id());
          }
          delete pile;
	}
      }
  
  // Etape 4 : Instances
  
  nbobj=0;
  for (i=0; i<NO; i++) {
    // Instanciation de la classe Object
    objtable[i]=new Object(i,classtable[CRand(DIST3,0,NC-1)]);
    nbobj++;
    // Ajout dans l'iterateur de la classe
    objtable[i]->ClassPtr()->Iterator()->Insert(1,i);
  }
  

  // remember which were the first valid object references chosen
  for (i=0; i<NO; i++) {
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++) {
      if (objtable[i]->ClassPtr()->GetCRef(j)!=NULL) {
	objtable[i]->ClassPtr()->SetInitialRef(j, true);
      }
      else {
	objtable[i]->ClassPtr()->SetInitialRef(j, false);
      }
    }
  }
  
  int rvalid_ref;
  Class * cl;

  // now assign every non valid reference a valid id (chosen randomly from
  // existing valid references for the class.
  // This makes it so that not all valid references are all allocated in this
  // initialisation function
  for (i=0; i<NO; i++) {
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++) {
      if (objtable[i]->ClassPtr()->GetCRef(j)==NULL) {	
	while (1) {
	  rvalid_ref=CRand(1,0,objtable[i]->ClassPtr()->MaxNRef());
	  if (objtable[i]->ClassPtr()->GetCRef(rvalid_ref)!=NULL) {
	    break;
	  }
	}
	cl = objtable[i]->ClassPtr()->GetCRef(rvalid_ref);
	classtable[objtable[i]->ClassPtr()->Id()]->SetCRef(j, cl);
      }
    }
  }
  

  // Etape 5 : Liens inter-instances
  
  for (i=0; i<NO; i++) {
    // used to insure that you do not have more than one reference to the 
    // same object more than once
    set<int> temp_set;
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++) {
      // Reference objet a NULL
      objtable[i]->SetORef(j,NULL);
      // Si classe referencee par classe mere non nulle
      if (objtable[i]->ClassPtr()->InitialRefExists(j) == true)
        // Si iterateur de cette classe non vide
        if (objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->NbElem()) {
	   // Affectation reference
          l=CRand(DIST4,1,
		  objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->NbElem());
	  selected_oid = 
	    objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l);
	  while (temp_set.find(selected_oid) != temp_set.end()) {	    
	    l=CRand(DIST4,1,
		    objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->NbElem());
	    selected_oid =
	      objtable[i]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l);
	  }
	  temp_set.insert(selected_oid);
          objtable[i]->SetORef(j,objtable[selected_oid]);
	  // cout << "STATS = " << STATS << endl;
	  //	  STATS->IncTension(i, selected_oid, 1);
	  // Affectation Backref pour l'objet reference
	  objtable[selected_oid]->BackRef()->Insert(1,objtable[i]->OID());
        }
    }
  }
  //#ifdef DEBUG
  cout << "Average obj size = " << CalcAvgObjSize() << endl;
  cout << "Min obj size = " << CalcMinObjSize()  << endl;
  cout << "Max obj size = " << CalcMaxObjSize() << endl;
  //  exit(1);
  //#endif

#ifdef CHANGE_SEED
  lp_tt = prev_seed;
#endif

}

void Database::MoveDepAssignObjects() {
  int refOid;

  cout << "About to do assignment" << endl;
  for (int i =0; i < NO ; i++) {
    if (objtable[i] != NULL) {
      if (MOVE_WINDTYPE == 1) {
	/*	cout << "assign = " << i << endl;
	cout << "oid = " << objtable[i]->OID() << ", ";
	cout << "class id = " <<  objtable[i]->ClassPtr()->Id() << endl;  */
	HMWD->InsertObjectRandom(objtable[i]->ClassPtr()->Id(),
				 objtable[i]->OID());
	HMWD->InsertSameClassObject(objtable[i]->ClassPtr()->Id(),
				    objtable[i]->OID());
	for (int j = 0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	  if (objtable[i]->GetORef(j) != NULL) {
	    refOid = objtable[i]->GetORef(j)->OID();
	    HMWD->InsertSRef(objtable[i]->OID(), refOid);
	  }
	}
	
	set<int> tempDRefSet;
	
	while ((int)tempDRefSet.size() <  DEP_NUMDREF) {
	  
	  refOid = DB->GetRandomOid(1);
	  if ((tempDRefSet.find(refOid) == tempDRefSet.end())
	      && (refOid != objtable[i]->OID())) {
	    
	    tempDRefSet.insert(refOid);
	    HMWD->InsertDRef(objtable[i]->OID(), refOid);
	    
	  }
	}
	
      }
      else if (MOVE_WINDTYPE == 2) {
	HGMWD->InsertObjectRandom(objtable[i]->ClassPtr()->Id(),
				  objtable[i]->OID());
	HGMWD->InsertSameClassObject(objtable[i]->ClassPtr()->Id(),
				     objtable[i]->OID());
	for (int j = 0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	  if (objtable[i]->GetORef(j) != NULL) {
	    refOid = objtable[i]->GetORef(j)->OID();
	    HGMWD->InsertSRef(objtable[i]->OID(), refOid);
	  }
	}

	set<int> tempDRefSet;
	
	while ((int) tempDRefSet.size() <  DEP_NUMDREF) {

	  refOid = DB->GetRandomOid(1);
	  if ((tempDRefSet.find(refOid) == tempDRefSet.end())
	      && (refOid != objtable[i]->OID())) {
	    
	    tempDRefSet.insert(refOid);
	    HGMWD->InsertDRef(objtable[i]->OID(), refOid);

	  }
	}

      }
      else if (MOVE_WINDTYPE == 3) {
	HD->InsertObjectRandom(objtable[i]->ClassPtr()->Id(),
			       objtable[i]->OID());
	HD->InsertSameClassObject(objtable[i]->ClassPtr()->Id(),
				  objtable[i]->OID());
	for (int j = 0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	  if (objtable[i]->GetORef(j) != NULL) {
	    refOid = objtable[i]->GetORef(j)->OID();
	    HD->InsertSRef(objtable[i]->OID(), refOid);
	  }
	}

	set<int> tempDRefSet;	
	while ((int)tempDRefSet.size() <  DEP_NUMDREF) {

	  refOid = DB->GetRandomOid(1);
	  if ((tempDRefSet.find(refOid) == tempDRefSet.end())
	      && (refOid != objtable[i]->OID())) {
	    
	    tempDRefSet.insert(refOid);
	    HD->InsertDRef(objtable[i]->OID(), refOid);

	  }
	}
      }
      else if (MOVE_WINDTYPE == 4) {
	HHGD->InsertObjectRandom(objtable[i]->ClassPtr()->Id(),
			       objtable[i]->OID());
	HHGD->InsertSameClassObject(objtable[i]->ClassPtr()->Id(),
				  objtable[i]->OID());
	for (int j = 0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	  if (objtable[i]->GetORef(j) != NULL) {
	    refOid = objtable[i]->GetORef(j)->OID();
	    HHGD->InsertSRef(objtable[i]->OID(), refOid);
	  }
	}

	set<int> tempDRefSet;	
	while ((int)tempDRefSet.size() <  DEP_NUMDREF) {

	  refOid = DB->GetRandomOid(1);
	  if ((tempDRefSet.find(refOid) == tempDRefSet.end())
	      && (refOid != objtable[i]->OID())) {
	    
	    tempDRefSet.insert(refOid);
	    HHGD->InsertDRef(objtable[i]->OID(), refOid);

	  }
	}
      }
      else if (MOVE_WINDTYPE == 5) {
	HCD->InsertObjectRandom(objtable[i]->ClassPtr()->Id(),
			       objtable[i]->OID());
	HCD->InsertSameClassObject(objtable[i]->ClassPtr()->Id(),
				  objtable[i]->OID());
	for (int j = 0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	  if (objtable[i]->GetORef(j) != NULL) {
	    refOid = objtable[i]->GetORef(j)->OID();
	    HCD->InsertSRef(objtable[i]->OID(), refOid);
	  }
	}

	set<int> tempDRefSet;	
	while ((int)tempDRefSet.size() <  DEP_NUMDREF) {

	  refOid = DB->GetRandomOid(1);
	  if ((tempDRefSet.find(refOid) == tempDRefSet.end())
	      && (refOid != objtable[i]->OID())) {
	    
	    tempDRefSet.insert(refOid);
	    HCD->InsertDRef(objtable[i]->OID(), refOid);

	  }
	}
      }
    }
  }

  cout << "just before final assignment" << endl;
  if (MOVE_WINDTYPE == 1) {
    HMWD->Assign();
  }
  else if (MOVE_WINDTYPE == 2) {
    HGMWD->Assign();
  }
  else if (MOVE_WINDTYPE == 3) {
    HD->Assign();
  }
  else if (MOVE_WINDTYPE == 4) {
    HHGD->Assign();
  }
  else if (MOVE_WINDTYPE == 5) {
    HCD->Assign();
  }

  cout << "Finished performing assignment" << endl;
}

float Database::CalcAvgObjSize() {
  int total_obj_size = 0;
  int num_obj = 0;

  for (int i =0; i < NO ; i++) {
    if (objtable[i] != NULL) {
      total_obj_size += objtable[i]->ClassPtr()->GetSize();
      num_obj++;
    }
  }
  cout << "total db size (adding objects) = " << total_obj_size << endl;
  return (float)total_obj_size / (float) num_obj;
}


int Database::CalcMinObjSize() {

  int min_obj_size = 10000;
  for (int i =0; i < NO ; i++) {
    if (objtable[i] != NULL) {
      if ( objtable[i]->ClassPtr()->GetSize() < min_obj_size) {
	min_obj_size =  objtable[i]->ClassPtr()->GetSize();
      }
    }
  }
  
  return min_obj_size;
}

int Database::CalcMaxObjSize() {
  int max_obj_size = 0;

  for (int i =0; i < NO ; i++) {
    if (objtable[i] != NULL) {
      if ( objtable[i]->ClassPtr()->GetSize() > max_obj_size) {
	max_obj_size =  objtable[i]->ClassPtr()->GetSize();
      }
    }
  }
  
  return max_obj_size;
}

// takes the random distribution as parameter
int Database::GetRandomOid(int dist, int * past_non_reg) {
  int temp_oid;

  temp_oid=CRand(dist,0,NO-1);
  while (ValidOid(temp_oid) == false) {
    temp_oid=CRand(dist,0,NO-1);
  }
  
  return temp_oid;
}

bool Database::LinkExists(int prev_oid, int curr_oid) {
  for (int j = 0; j < objtable[prev_oid]->ClassPtr()->MaxNRef(); j++) {
    if (objtable[prev_oid]->GetORef(j)!=NULL) {
      if (objtable[prev_oid]->GetORef(j)->OID() == curr_oid) {
	return true;
      }
    }
  }
  return false;
}

void Database::PlaceNewObj( int new_oid ) {
  switch(INITP) {
  case 1: _SeqPl( new_oid );break;  // Placement initial sequentiel
  case 2: _OSeqPl( new_oid );break; // Placement initial sequentiel optimise
  case 9: _OtherPl( new_oid );break;    // Placement initial autre
  }
} 


// This function tries to use the oid given in obj_oid for inserting new
// references but if it finds that there are no free space to insert the new
// reference then it randomly selects a new object.
void Database::InsertRef(int * obj_id, int dist) {
  bool found_free_ref = false;
  int free_ref_num = 1000000000;
  int selected_oid;
  
  while (found_free_ref == false) {
    for (int j=0; j<objtable[*obj_id]->ClassPtr()->MaxNRef(); j++) {
      if ((objtable[*obj_id]->GetORef(j)==NULL) &&
	  (objtable[*obj_id]->ClassPtr()->GetCRef(j)->Iterator()->NbElem())) {
	found_free_ref = true;
	free_ref_num = j;
	break;
      }
    }
    if (found_free_ref == false) {
      *obj_id = GetRandomOid(dist);
    }
  }
  
  // to insure the same source and target object pairs are not repeated,
  // we must first get all existing object references as insert it into a set
  // so we can later check for the existence of the references a newly created
  // reference in the set.
  set<int> temp_set;
  for (int j=0; j<objtable[*obj_id]->ClassPtr()->MaxNRef(); j++) {
    if ((objtable[*obj_id]->GetORef(j)!=NULL) && (j != free_ref_num)) {
      temp_set.insert(objtable[*obj_id]->GetORef(j)->OID());
    }
  }

  int l = CRand(DIST4,1,
		objtable[*obj_id]->ClassPtr()->GetCRef(free_ref_num)
		->Iterator()->NbElem());
  selected_oid = 
    objtable[*obj_id]->ClassPtr()->GetCRef(free_ref_num)->Iterator()->Scan(l);
  while (temp_set.find(selected_oid) != temp_set.end()) {	    
    l=CRand(DIST4,1,
	    objtable[*obj_id]->ClassPtr()->GetCRef(free_ref_num)
	    ->Iterator()->NbElem());
    selected_oid =
      objtable[*obj_id]->ClassPtr()->GetCRef(free_ref_num)->Iterator()
      ->Scan(l);
  }

  objtable[*obj_id]->SetORef(free_ref_num, objtable[selected_oid]);
  // Affectation Backref pour l'objet reference
  objtable[selected_oid]->BackRef()->Insert(1,objtable[*obj_id]->OID());
}

int Database::InsertNewObj() {
  int new_oid, l;
  Object * new_obj;
  ChList ** newChListPtrPtr;
  
  new_oid = NO;
  NO++;
  ACTUALNO++;
  // create room for stats
  newChListPtrPtr = new (ChList*) [2];
  mo.push_back(newChListPtrPtr);
  newChListPtrPtr = new (ChList*) [4];
  mc.push_back(newChListPtrPtr);
  vo.push_back(0);
  for (int j=0; j<2; j++)
    mo[new_oid][j]=new ChList;
  for (int j=0; j<4; j++)
    mc[new_oid][j]=new ChList;
  
  new_obj = new Object(new_oid,classtable[CRand(DIST101,0,NC-1)]);
  objtable.push_back(new_obj);
  nbobj++;
  // Ajout dans l'iterateur de la classe
  objtable[new_oid]->ClassPtr()->Iterator()->Insert(1,new_oid);
  for (int j=0; j<objtable[new_oid]->ClassPtr()->MaxNRef(); j++) {
    // Reference objet a NULL
    objtable[new_oid]->SetORef(j,NULL);
    // Si classe referencee par classe mere non nulle
    if (objtable[new_oid]->ClassPtr()->InitialRefExists(j)==true) {
      // Si iterateur de cette classe non vide
      if (objtable[new_oid]->ClassPtr()->GetCRef(j)->Iterator()->NbElem()) {
	// Affectation reference
	l=CRand(DIST4,1,objtable[new_oid]->ClassPtr()->GetCRef(j)->Iterator()->NbElem());
	objtable[new_oid]->SetORef(j,objtable[objtable[new_oid]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l)]);
	// Affectation Backref pour l'objet reference
	objtable[objtable[new_oid]->ClassPtr()->GetCRef(j)->Iterator()->Scan(l)]->BackRef()->Insert(1,objtable[new_oid]->OID());
      }
    }
  }
  return new_oid;
}

void Database::DeleteRef(int * obj_id, int dist) {
  bool found_ref = false;
  int selected_ref, to_oid, temp_oid;

  // make sure the chosen object has an existing references 
  while (found_ref == false) {
    for(int j = 0; j < objtable[*obj_id]->ClassPtr()->MaxNRef(); j++) {
      if (objtable[*obj_id]->GetORef(j)!=NULL){
	found_ref = true;
	break;
      }
    }
    if (found_ref == false) {
      *obj_id = GetRandomOid(dist);
    }
  }
  
  // randomly select a valid reference to delete
  while (1) {
    selected_ref = CRand( 1, 0, (objtable[*obj_id]->ClassPtr()->MaxNRef()- 1));
    if (objtable[*obj_id]->GetORef(selected_ref) != NULL) {
      to_oid = objtable[*obj_id]->GetORef(selected_ref)->OID();
      // only set the tension to zero, instead of deleting the reference
      // because this way the clustering algorithm can still use the reference
      // as a back reference
      if (CLUSTP == 1) {
	STATS->SetTensionZero(*obj_id, to_oid);
      }
      else if (CLUSTP == 9) {
	int place = -1;
	for (int i =1; i <= DB->mo[*obj_id][0]->NbElem(); i++) {
	  if (mo[*obj_id][0]->Scan(i) == to_oid) {
	    place = i;
	  }
	}
	
	if (place != -1) {
	  DB->mo[*obj_id][0]->Delete(place);
	  DB->mo[*obj_id][1]->Delete(place);
	}

	place = -1;
	for (int i =1; i <= DB->mc[*obj_id][0]->NbElem(); i++) {
	  if (mc[*obj_id][0]->Scan(i) == to_oid) {
	    place = i;
	  }
	}
	if (place != -1) {
	  for (int j=0; j<4; j++)
	    mc[*obj_id][j]->Delete(place);
	}
      }
      objtable[*obj_id]->SetORef(selected_ref, NULL);
      // delete the back reference to this object
       for (int i=1; i <= objtable[to_oid]->BackRef()->NbElem();
	   i++) {
	 temp_oid = objtable[to_oid]->BackRef()->Scan(i);
	 if (temp_oid == *obj_id) {
	   objtable[to_oid]->BackRef()->Delete(i);
	   break;
	 }
       }
       break;    
    }
  }
}

bool Database::ValidOid(int i_oid) {
  if (objtable[i_oid] == NULL) {
    return false;
  }
  else {
    return true;
  }
}

void Database::DeleteObj(int del_oid) {
  int temp_oid, temp_oid2;

  ACTUALNO--;
  ACTUALNO--;
  if (CLUSTP == 9) {
    vo[del_oid] = 0;
    for (int j=0; j<2; j++) {
      delete mo[del_oid][j];
      mo[del_oid][j] = new ChList;
    }
    for (int j=0; j<4; j++) {
      delete mc[del_oid][j];
      mc[del_oid][j] = new ChList;
    }
  }
  else if ((CLUSTP == 1) || (CLUSTP ==2)) {
    STATS->DeleteStats(del_oid);
  }
  int page_num;
  // Delete object from the page that it resides in.
  for (int i =1; i <=   DB->GetObj(del_oid)->PageDisk()->NbElem(); i++) {
    page_num = DB->GetObj(del_oid)->PageDisk()->Scan(i);
    if (DB->GetPage(page_num)->DeleteObj(del_oid) == false) {
      cerr << "Can not find obj to delete in basem.h" << endl;
      exit(1);
    }  
    // re-calculate the amount free space in the page
    DB->GetPage(page_num)->ReCalcFree();
  }
  

  // Delete the reference to the object from the class iterator
  for (int i =1; i <=objtable[del_oid]->ClassPtr()->Iterator()->NbElem(); 
       i++) {
    temp_oid = objtable[del_oid]->ClassPtr()->Iterator()->Scan(i);
    if (temp_oid == del_oid) {
      objtable[del_oid]->ClassPtr()->Iterator()->Delete(i);
      break;
    } 
  }
  // remove all back references that point to this object
  for (int j=0; j<objtable[del_oid]->ClassPtr()->MaxNRef(); j++) {
    if (objtable[del_oid]->GetORef(j)!=NULL) {
      temp_oid = objtable[del_oid]->GetORef(j)->OID();
      if (CLUSTP == 9) {
	int place = -1;
	for (int i =1; i <= DB->mo[temp_oid][0]->NbElem(); i++) {
	  if (mo[temp_oid][0]->Scan(i) == del_oid) {
	    place = i;
	  }
	}

	if (place != -1) {
	  DB->mo[temp_oid][0]->Delete(place);
	  DB->mo[temp_oid][1]->Delete(place);
	}

	place = -1;
	for (int i =1; i <= DB->mc[temp_oid][0]->NbElem(); i++) {
	  if (mc[temp_oid][0]->Scan(i) == del_oid) {
	    place = i;
	  }
	}
	if (place != -1) {
	  for (int j=0; j<4; j++)
	    mc[temp_oid][j]->Delete(place);
	}
      }
      int k=0;
      for (int i=1; i <= objtable[temp_oid]->BackRef()->NbElem();
	   i++) {
	temp_oid2 = objtable[temp_oid]->BackRef()->Scan(i);
	if (temp_oid2 == del_oid) {
	  objtable[temp_oid]->BackRef()->Delete(i);
	  //  break;
	  k++;
	}
      }
   
    }
  }

  // remove all forward references of objects that point to this object
  for (int j=1; j<=objtable[del_oid]->BackRef()->NbElem(); j++) {
    temp_oid = objtable[del_oid]->BackRef()->Scan(j);
    for (int i=0; i < objtable[temp_oid]->ClassPtr()->MaxNRef(); i++) {
      if (objtable[temp_oid]->GetORef(i)!=NULL) {
	temp_oid2 = objtable[temp_oid]->GetORef(i)->OID();
	if (temp_oid2 == del_oid) {
	  objtable[temp_oid]->SetORef(i, NULL);
	  break;
	}
      }
    }
  }
  delete objtable[del_oid];
  objtable[del_oid] = NULL;
}


// CLASSE Database : Affichage de la base

void Database::Display() {

  int i, j;
  float sizec, refc, ref[MNREFT], refn[MNREFT];

  printf("\n*** Database Schema ***\n");
  sizec=0;
  refc=0;
  for (i=0; i<NREFT; i++) {
    ref[i]=0;
    refn[i]=0;
  }
  for (i=0; i<NC; i++) {
    classtable[i]->Display();
    sizec+=classtable[i]->GetSize();
    for (j=0; j<classtable[i]->MaxNRef(); j++) {
      ref[classtable[i]->GetTRef(j)]++;
      if (classtable[i]->GetCRef(j)!=NULL) {
        refc++;
        refn[classtable[i]->GetTRef(j)]++;
      }
    }
  }
  printf("\n Instances mean size                 : %f\n",sizec/NC);
  printf(" Mean number of references per class : %f\n",refc/NC);
  for (i=0; i<NREFT; i++) {
  printf(" Mean number of type #%d references   : %f (%f are not NULL)\n",i,ref[i]/NC,refn[i]/NC);
  }

  printf("\n*** Database Instanciation ***\n");
  refc=0;
  for (i=0; i<NO; i++) {
    objtable[i]->Display();
    for (j=0; j<objtable[i]->ClassPtr()->MaxNRef(); j++)
      if (objtable[i]->GetORef(j)!=NULL) refc++;
  }
  printf("\n Mean number of references per object: %f\n",refc/NO);
}

// CLASSE Database : Placement initial des objets

void Database::Initp() {

//  printf("Objects initial placement [%d]... ",INITP);

  switch(INITP) {
    case 1: SeqPl();break;      // Placement initial sequentiel
    case 2: OSeqPl();break;     // Placement initial sequentiel optimise
    case 9: OtherPl();break;    // Placement initial autre
  }

//  printf("%d pages used\n",nbpages);
}

// CLASSE Database : Placement initial sequentiel
void Database::SeqPl() {

  // Creation nouvelle page et insertion dans la liste
  cur_page=NewPage();
  AddPage(cur_page);  


  // Placement de chaque objet  
  for (int i=0; i<NO; i++) {
    _SeqPl(i);
  }


  cout << "number of pages = " << nbpages << endl;
}

void Database::_SeqPl(int new_oid) {
  int rsize;
  
  if (objtable[new_oid]->ClassPtr()->GetSize()<=PGSIZE) {
    // Objet normal, de taille inferieure a celle d'une page
    if (objtable[new_oid]->ClassPtr()->GetSize()>cur_page->Free()) {
      // Taille > place libre => nouvelle page
      cur_page=NewPage();
      AddPage(cur_page);
    }
    // Objet insere dans le contenu de la page
    cur_page->Objects()->Insert(1,new_oid);
    // Page dans la liste des pages de l'objet
    objtable[new_oid]->PageDisk()->Insert(1,cur_page->PageNum());
    // MAJ place libre
    cur_page->ResetFree(cur_page->Free()-objtable[new_oid]->ClassPtr()->GetSize());
  } else { // Objet long
    // rsize : taille restant a placer
    rsize=objtable[new_oid]->ClassPtr()->GetSize()-cur_page->Free();
    // Page courante saturee
    cur_page->ResetFree(0);
    // Objet insere dans le contenu de la page
    cur_page->Objects()->Insert(1,new_oid);
    // Page dans la liste des pages de l'objet
    objtable[new_oid]->PageDisk()->Insert(1,cur_page->PageNum());
    // Tant que l'objet n'est pas totalement stocke
    while (rsize>0) {
      // Nouvelle page
      cur_page=NewPage();
      AddPage(cur_page);
      // Objet insere dans le contenu de la page
      cur_page->Objects()->Insert(1,new_oid);
      // Page dans la liste des pages de l'objet
      objtable[new_oid]->PageDisk()->Insert(1,cur_page->PageNum());
      if (cur_page->Free()<rsize) {   // L'objet ne tient pas
	rsize=rsize-cur_page->Free(); // Taille restante --
	cur_page->ResetFree(0);       // Page saturee
      } else {
	cur_page->ResetFree(cur_page->Free()-rsize); // Espace libre --
	rsize=0;
      }
    }
  }
}

void Database::_OSeqPl(int new_oid) {
  Page *goodpage, *scan;
  int rsize;
  short space;

  if (objtable[new_oid]->ClassPtr()->GetSize()<=PGSIZE) {
    // Objet normal, de taille inferieure a celle d'une page
    if (objtable[new_oid]->ClassPtr()->GetSize()<=cur_page->Free())
      // taille <= place libre
      goodpage=cur_page;
    else { // Recherche d'une page d'accueil dans les pages existantes
      scan=pagelist;
      space=0;
      while ((!space) && (scan!=NULL))
	if (objtable[new_oid]->ClassPtr()->GetSize()<=scan->Free())
	  space=1;
	else
	  scan=scan->Next();
      if (space) goodpage=scan;
      else { // Nouvelle page
	cur_page=NewPage();
	AddPage(cur_page);
	goodpage=cur_page;
      }
    }
   
    // Objet insere dans le contenu de la page
    goodpage->Objects()->Insert(1,new_oid);
       
    // Page dans la liste des pages de l'objet
    objtable[new_oid]->PageDisk()->Insert(1,goodpage->PageNum());
    // MAJ place libre
    goodpage->ResetFree(goodpage->Free()-objtable[new_oid]->ClassPtr()->GetSize());
  } else { // Objet long
    // rsize : taille restant a placer
    rsize=objtable[new_oid]->ClassPtr()->GetSize()-cur_page->Free();
    // Page courante saturee
    cur_page->ResetFree(0);
    
    // Objet insere dans le contenu de la page
    cur_page->Objects()->Insert(1,new_oid);
    // Page dans la liste des pages de l'objet
    objtable[new_oid]->PageDisk()->Insert(1,cur_page->PageNum());
    // Tant que l'objet n'est pas totalement stocke
    while (rsize>0) {
      // Nouvelle page
      cur_page=NewPage();
      AddPage(cur_page);
      // Objet insere dans le contenu de la page
      cur_page->Objects()->Insert(1,new_oid);
      // Page dans la liste des pages de l'objet
      objtable[new_oid]->PageDisk()->Insert(1,cur_page->PageNum());
      if (cur_page->Free()<rsize) {   // L'objet ne tient pas
	rsize=rsize-cur_page->Free(); // Taille restante --
	cur_page->ResetFree(0);       // Page saturee
      } else {
	cur_page->ResetFree(cur_page->Free()-rsize); // Espace libre --
	rsize=0;
      }
    }
  }
}

bool Database::PageEmpty(int pg_num) {
  return GetPage(pg_num)->PageEmpty();
}

// CLASSE Database : Placement initial sequentiel optimise

void Database::OSeqPl() {

  // Creation nouvelle page et insertion dans la liste
  cur_page=NewPage();
  AddPage(cur_page);

  // Placement de chaque objet
  for (int i=0; i<NO; i++) {
    _OSeqPl(i);
  }
}

// CLASSE Database : Placement initial autre (defini par l'utilisateur)

void Database::OtherPl() {
  // CODE A INSERER ICI
  
}

void Database::_OtherPl(int new_oid) {
  
}

// CLASSE Database : Retourne la classe numero I

Class *Database::GetClass(int i) {

  if ((i>=0) && (i<nbclasses)) return classtable[i];
  else return NULL;
}

// CLASSE Database : Retourne l'objet numero I

Object *Database::GetObj(int i) {

  if ((i>=0) && (i<nbobj)) return objtable[i];
  else return NULL;
}

// CLASSE Database : Creation d'une nouvelle page

Page *Database::NewPage() {

  Page *newpage;
  newpage=new Page(cumulp++);
  return newpage;
}

// CLASSE Database : Ajout page a la liste (en tete)

void Database::AddPage(Page *page) {

  page->SetNext(pagelist);
  pagelist=page;
  nbpages++;
}

// CLASSE Database : Retourne un ptr sur la page num

// please no you must specify the page number of object number
Page *Database::GetPage(int num) {

  Page *cur;

  cur=pagelist;
  while ((cur!=NULL) && (cur->PageNum()!=num)) cur=cur->Next();
  if (cur!=NULL) return cur;
  else return NULL;
}

// CLASSE Database : Suppression d'une page

void Database::DelPage(int num) {

  Page *cur, *prev;

  prev=NULL;
  cur=pagelist;
  while ((cur!=NULL) && (cur->PageNum()!=num)) {
    prev=cur;
    cur=cur->Next();
  }
  if ((cur!=NULL) && (cur->PageNum()==num)) { // Suppression
    if (prev!=NULL) prev->SetNext(cur->Next());
    else pagelist=cur->Next();
    delete cur;
  }
  nbpages--;
}

// CLASSE Database : Suppression de toutes les pages

void Database::PurgePages() {

  Page *cur, *del;

  cur=pagelist;
  del=NULL;
  while (cur!=NULL) {
    del=cur;
    cur=cur->Next();
    delete del;
  }
  pagelist=NULL;
}

// CLASSE Database : Affichage des pages utilisees

void Database::DisplayPages() {

  Page *cur;
  int i;

  printf("\n*** Disk pages used ***\n");
  cur=pagelist;
  while (cur!=NULL) {
    printf("\nPage #%d\n",cur->PageNum());
    printf("\tFree space: %d\n",cur->Free());
    printf("\tObjects contained:");
    for (i=1; i<=cur->Objects()->NbElem(); i++)
      printf(" %d",cur->Objects()->Scan(i));
    printf("\n");
    cur=cur->Next();
  }
}

// CLASSE Database : Retourne le nombre de pages utilisees

int Database::NbPages() {
  return nbpages;
}


// CLASSE Database : Nouvelle periode d'observation

void Database::NewPeriod() {

  int i, j;

  // Periode suivante
  pi++;
  if (pi>=P) pi=0;
  // RAZ MO
  ni=0;
  for (i=0; i<NO; i++) {
    for (j=0; j<2; j++) {
      delete mo[i][j];
      mo[i][j]=new ChList;
    }
    vo[i]=0;
  }

  int del = 0;
  // delete very old consolidation entries
  for (i=0; i<NO; i++) {
    /*    bool found_delete = true;
    while (found_delete == true) {
    found_delete = false;*/
      for (j = 1; j <= DB->mc[i][2]->NbElem(); j++) {
	if ( ( pi - DB->mc[i][2]->Scan(j) ) > (NP % P)) {
	  //cout << "deleting mc pair (" << i << ", " << DB->mc[i][0]->Scan(j) << ")" << endl;
	  //cout << "DB->mc[i][2]->Scan(j) " << DB->mc[i][2]->Scan(j) << endl;
	  //cout << "current period = " << pi << endl;
	  /*DB->mc[i][0]->Delete(j);
	  DB->mc[i][1]->Delete(j);
	  DB->mc[i][2]->Delete(j);
	  DB->mc[i][3]->Delete(j);*/
	  
	  //found_delete = true;
	  //	  break;
	  //	}
	  del++;
	  DB->mc[i][1]->Update(j, 0);	  
      }
    }
  }
  cout << "number deleted  = " << del << endl;
}

// CLASSE Database : Affichage matrice MO

void Database::DisplayMO() {

  int i, j;

  printf("\n*** MO ***\n\n");
  for (i=0; i<NO; i++) {
    printf("%d)",i);
    if (mo[i][0]->NbElem()==mo[i][1]->NbElem())
      for (j=1; j<=mo[i][0]->NbElem(); j++)
        printf(" %d:%d",mo[i][0]->Scan(j),mo[i][1]->Scan(j));
    else                              // Probleme !!
      printf(" ERROR: Discrepancy in MO!");
    printf(" - %d\n",vo[i]);
  }
}

// CLASSE Database : Affichage matrice MC

void Database::DisplayMC() {

  int i, j;

  printf("\n*** MC ***\n\n");
  for (i=0; i<NO; i++) {
    printf("%d)",i);
    for (j=1; j<=mc[i][0]->NbElem(); j++)
      printf(" %d:%d/%d/%d",mc[i][0]->Scan(j),mc[i][1]->Scan(j),
      mc[i][2]->Scan(j),mc[i][3]->Scan(j));
    printf("\n");
  }
}

void Database::DisplayNumRef() {
  int num_ref = 0;
  
  for (int i=0; i<NO; i++) {
    if (objtable[i] != NULL) {
      for (int j=0; j < objtable[i]->ClassPtr()->MaxNRef(); j++) {
	if (objtable[i]->GetORef(j) != NULL) {
	  num_ref++;
	}
      }
    }
  }
  cout << "current total number of references = " << num_ref << endl;
}


vector<Page *> Database::GetPageList() {

  Page * cur;

  vector<Page *> pl;


  cur=pagelist;
  while (cur!=NULL) {
    pl.push_back(cur);
    cur=cur->Next();
  }

  return pl;

}

/////////////////////////////////////////////////////////////////////
// CLASSE Class
/////////////////////////////////////////////////////////////////////

// CLASSE Class : Constructeur

Class::Class(int i, short maxn, int bs) {

  id=i;
  maxnref=maxn;
  basesize=bs;
  iterator=new ChList;
  instsize=basesize;  // O2 : basesize + X (~700)
}

// CLASSE Class : Destructeur

Class::~Class() {

  delete iterator;
}

// CLASSE Class : Retourne maxnref

short Class::MaxNRef() {

  return maxnref;
}

// CLASSE Class : Retourne basesize

int Class::BaseSize() {

  return basesize;
}

// CLASSE Class : Initialise tref[i] a t

void Class::SetTRef(int i, short t) {

  if ((i>=0) && (i<maxnref)) tref[i]=t;
}

// CLASSE Class : Retourne tref[i]

short Class::GetTRef(int i) {

  if ((i>=0) && (i<maxnref)) return tref[i];
  else return -1;
}

// CLASSE Class : Initialise cref[i] a c

void Class::SetCRef(int i, Class *c) {

  if ((i>=0) && (i<maxnref)) cref[i]=c;
}

// CLASSE Class : Retourne cref[i]

Class *Class::GetCRef(int i) {

  if ((i>=0) && (i<maxnref)) return cref[i];
  else return NULL;
}

// CLASSE Class : Retourne l'adresse de l'iterateur

ChList *Class::Iterator() {

  return iterator;
}

// checks to see if a reference exists at this position when the schema was
// created, this is used to by insert object to know the initial level of
// connectivity of a new object.
bool Class::InitialRefExists(int ref) {
  return InitialRefEx[ref];
}


void Class::SetInitialRef(int ref, bool ire) {
  InitialRefEx[ref] = ire;
}

// CLASSE Class : Initialise instsize a s

void Class::SetSize(int s) {

  instsize=s;
}

// CLASSE Class : Retourne intsize

int Class::GetSize() {

  return instsize;
}

// CLASSE Class : Retourne id

int Class::Id() {

  return id;
}

// CLASSE Class : Affichage ecran

void Class::Display() {

  int i;

  printf("\n*** Class #%d\n",id);
  printf("\tmaxnref=%d\n",maxnref);
  printf("\tbasesize=%d\n",basesize);
  printf("\tcref(tref)=");
  for (i=0; i<maxnref; i++)
    if (cref[i]!=NULL) printf("%d(%d) ",cref[i]->Id(),tref[i]);
    else printf("NULL(%d) ",tref[i]);
  printf("\n\titerator=");
  for (i=1; i<=iterator->NbElem(); i++) printf("%d ",iterator->Scan(i));
  printf("\n\tinstsize=%d\n",instsize);
}

/////////////////////////////////////////////////////////////////////
// CLASSE Object
/////////////////////////////////////////////////////////////////////

// CLASSE Object : Constructeur

Object::Object(int id, Class *c) {

  oid=id;
  classptr=c;
  backref=new ChList;
  pagedisk=new ChList;
}

// CLASSE Object : Destructeur

Object::~Object() {

  delete backref;
  delete pagedisk;
}

// CLASSE Object : Retourne la classe de l'objet

Class *Object::ClassPtr() {

  return classptr;
}

// CLASSE Object : Initialise oref[i] a o

void Object::SetORef(int i, Object *o) {

  if ((i>=0) && (i<classptr->MaxNRef())) oref[i]=o;
}

// CLASSE Object : Retourne oref[i]

Object *Object::GetORef(int i) {

  if ((i>=0) && (i<classptr->MaxNRef())) return oref[i];
  else return NULL;
}

// CLASSE Object : Retourne l'adresse des refs inverses

ChList *Object::BackRef() {

  return backref;
}

// CLASSE Object : Retourne la liste des pages contenant l'objet

ChList *Object::PageDisk() {

  return pagedisk;
}


void Object::DeleteAllPageDisks() {
  delete pagedisk;
}

void Object::CreateNewPageDisk() {
  pagedisk = new ChList();
}

void Object::InsertPage(int pg_num) {
  pagedisk->Insert(1, pg_num);
}

// CLASSE Object : Retourne l'OID

int Object::OID() {

  return oid;
}

// CLASSE Object : Affichage ecran

void Object::Display() {

  int i;

  printf("\n*** Object #%d\n",oid);
  printf("\tclass=%d\n",classptr->Id());
  printf("\toref=");
  for (i=0; i<classptr->MaxNRef(); i++)
    if (oref[i]!=NULL) printf("%d ",oref[i]->OID());
    else printf("NULL ");
  printf("\n\tbackref=");
  for (i=1; i<=backref->NbElem(); i++) printf("%d ",backref->Scan(i));
  printf("\n\tpagedisk=");
  for (i=1; i<=pagedisk->NbElem(); i++) printf("%d ",pagedisk->Scan(i));
  printf("\n");
}

/////////////////////////////////////////////////////////////////////
// CLASSE Page
/////////////////////////////////////////////////////////////////////

// CLASSE Page : Constructeur

Page::Page(int num) {

  pagenum=num;
  free=PGSIZE;
  objects=new ChList;
}

int Page::CalcHeat() {
  int obj_id;
  int heat = 0;
  ObjStats * os;

  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    os = STATS->GetObjStats(obj_id);
    if (os != NULL) {
      heat += os->GetHeat();
    }
  }
  return heat;
}


float Page::CalcWaste() {
  int obj_id;
  float waste = 0.0;
  ObjStats * os;
  int size;
  //  int total_size = 0;

  // by defualt every object is assumed to have a heat of 1
  // this is so a zero heat object does not get division by zero when 
  // calculating waste
  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    os = STATS->GetObjStats(obj_id);
    size = DB->GetObj(obj_id)->ClassPtr()->GetSize();
    if (os != NULL) {
      waste += (float)size / (float)(os->GetHeat() +1);      
    }
    else {
      waste += (float)size;
    }
    //    total_size += size;
  }
  
  // consider the empty space in a page as one big object of heat 1.
  if (objects->NbElem() != 0) {
    waste += free;
  }

  return waste;
}


bool Page::PageEmpty() {
  bool page_empty;
   /*
  if (objects->NbElem() == 0) {
    page_empty = true;
  }
  else {
    page_empty = false;
  }
  */
  if (free == PGSIZE) {
    page_empty = true;
  }
  else {
    page_empty = false;
  }
  return page_empty;
}

int Page::CalcExtTension(BuffMgr * bmgr) {
  int obj_id;
  set<int> temp_set; // used to store objects that exist in this page so that
                     // calculating external tension the object being
                     // referenced can be check to see if it resides in this
                     // page or not, only increment tension if the referenced
                     // object does not reside on this page.
  int ExTension = 0;
  set<int> ref_set;

  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    temp_set.insert(obj_id);
  }  

  for(int i = 1; i <= objects->NbElem(); i++) {
    ObjStats * os;
    int ref_oid;

    obj_id = objects->Scan(i);
    os = STATS->GetObjStats(obj_id);
    if (os != NULL) {
      ref_oid = os->Start();
      while (ref_oid != NULL_OID) {
	if (temp_set.find(ref_oid) == temp_set.end()) {
	  ref_set.insert(ref_oid);
	} 
	ref_oid = os->NextOid();
      }
    }    
  }
#ifdef DEBUG  
  cout << "ref size b = " << ref_set.size() << endl;
#endif
  if (OPPORTUNISTIC_WRITE == true) {
    bmgr->RemoveNonUpdated(&ref_set);
  } 
  else if (OPPORTUNISTIC_READ == true) {
    bmgr->RemoveNonRes(&ref_set);
  }

#ifdef DEBUG
  cout << "ref size a = " << ref_set.size() << endl;
#endif
  for(int i = 1; i <= objects->NbElem(); i++) {
    ObjStats * os;
    int ref_oid;
    
    obj_id = objects->Scan(i);
    os = STATS->GetObjStats(obj_id);
    if (os != NULL) {
      ref_oid = os->Start();
      while (ref_oid != NULL_OID) {
	if (ref_set.find(ref_oid) != ref_set.end()) {	  
	  // External tension exists and must be calculated
	  ExTension += os->GetCurrTension();
#ifdef DEBUG
	  cout << "page rc = " << DB->GetObj(ref_oid)->PageDisk()->Scan(1) 
	       << endl;
	  for (int j =1 ; j <= DB->GetObj(ref_oid)->PageDisk()->NbElem(); 
	       j++) {
	    cout << "obj page = " << DB->GetObj(ref_oid)->PageDisk()->Scan(j)
		 << endl;
	  }
	  cout << "self page num = " << PageNum() << endl;
	  if (temp_set.find(ref_oid) != temp_set.end()) {
	    cout << "found it " << endl;
	  }
	  cout << "interested oid = " << ref_oid << endl;
	
#endif
	}
	ref_oid = os->NextOid();
      }
    }    
  }

  return ExTension;
}

// CLASSE Page : Destructeur

Page::~Page() {

  delete objects;
}

// CLASSE Page : Initialisation pointeur next

void Page::SetNext(Page *n) {

  next=n;
}

// CLASSE Page : Retourne le numero de page

int Page::PageNum() {

  return pagenum;
}

// CLASSE Page : Retourne la place libre

int Page::Free() {

  return free;
}
// CLASSE Page : Reinitialise la place libre

void Page::ResetFree(int f) {

  free=f;
}

void Page::ReCalcFree() {
  int total_size = 0;
  int obj_id, size;

  // by defualt every object is assumed to have a heat of 1
  // this is so a zero heat object does not get division by zero when 
  // calculating waste
  for(int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    size = DB->GetObj(obj_id)->ClassPtr()->GetSize();    
    if (size > PGSIZE){
      cerr << 
	"ERROR: This is not designed to handle large objects, basem.h, ";
      cerr << "Page::ReCalcFree()" << endl;
    }
    total_size += size;
  }
  free = PGSIZE - total_size;
}

// CLASSE Page : Retourne la liste des objets

ChList *Page::Objects() {

  return objects;
}

// Pre-condition: there must be enough room to fit the new object in the page
void Page::InsertObj(int i_oid) {
  objects->Insert(1, i_oid);
}

void Page::DeleteAllObjs() {
  delete objects;
}

bool Page::DeleteObj(int del_oid) {
  int obj_id;
  bool found_obj = false;

  for (int i = 1; i <= objects->NbElem(); i++) {
    obj_id = objects->Scan(i);
    if (obj_id == del_oid) {
      objects->Delete(i);
      found_obj = true;
      break;
    }    
  }
  return found_obj;
}

void Page::CreateNewObjList() {
  objects = new ChList();
}

// CLASSE Page : Retourne la page suivante

Page *Page::Next() {

  return next;
}








