
/////////////////////////////////////////////////////////////////////
// Package SIMULation (C++ discrete event simulation package)
// Version 1.1 g++, mars 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
// eventc.h : Definition des classes du gestionnaire d'evenements
// Variable selon les systemes simules
// Implantation VOODB2 (avril 1998)
/////////////////////////////////////////////////////////////////////

// Declaration des ressources actives

class User;
class TransMgr;
class ClustMgr;
class ObjMgr;
class BuffMgr;
class IOSubSys;


struct pair_t {
  int from_oid;
  int to_oid;
};

// partition information structure
struct part_info_t {
  set<int> * obj_set;
  int size;
};

/////////////////////////////////////////////////////////////////////
// CLASSE EventManager
/////////////////////////////////////////////////////////////////////
// Gestion des evenements de la simulation
/////////////////////////////////////////////////////////////////////

// Le gestionnaire d'evenements doit connaitre toutes les ressources
// (actives et passives).

class EventManager {

  public:

    // Methodes

    EventManager(Simulation *sim);               // Constructeur
    ~EventManager();                             // Destructeur
    void ExecuteEvent(int code, Client *client); // Execution d'un evenement
    void Init();                                 // Initialisation

    // Initialisation replication
    void InitRep();
    void Stats();                  // Calcul des stats (fin rep)
    void DisplayStats();  // Affichage des stats

  private:

    // Attributs

    Simulation *simul;                  // Ptr vers objet simulation

    // Ressources passives
    Resource *serverpr;                 // Processeur + memoire primaire serveur
    Resource *clientpr[MNUSERS];        // Processeur + memoire primaire clients
    Resource *disk;                     // Controleur + memoire secondaire serveur
    Resource *db;                       // Base de donnees (controle de concurrence)

    // Ressources actives
    User *user[MNUSERS];
    TransMgr *transmgr;
    ClustMgr *clustmgr;
    ObjMgr *objmgr;
    BuffMgr *buffmgr;
    IOSubSys *iosubsys;

    //used for calc. of long term expansion factor
    int _ltef_num_ideal_pages;
};

/////////////////////////////////////////////////////////////////////
// CLASSE Client
/////////////////////////////////////////////////////////////////////
// Entite customisable de simulation
/////////////////////////////////////////////////////////////////////

class Client {

  public :
           // Ajouter ici (en public) les attributs supplementaires
           int cliid;           // Identificateur client
           short type;          // Type de client
                                // 0 = Transaction
                                // 1 = Acces ensembliste
                                // 2 = Parcours simple
                                // 3 = Parcours hierarchique
                                // 4 = Parcours stochastique
	                        // 5 = Random attribute update transaction type
	                        // 6 = Object insert transaction type
	                        // 7 = Object delete transaction type
	                        // 8 = Sequential attribute update transaction 
                                //     type
	                        // 9 = Reference insertion transaction type
	                        // 10 = Reference deletion transaction type
                                // GROUPT = Demande de groupement
                                // DDWT = Ecriture directe sur disque 
                                          // direct disk write 
                                // VCT = Client fictif 
                                         //virtusl client	                    
	                      
           // Transaction
           int rootoid;         // Objet racine
           int trans_it;        // Numero d'iteration dans la transaction
           int prevoid;         // Objet precedemment accede
           int curdepth;        // Profondeur courante	   
           ChList *roots;       // Liste des "racines"
           ChList *toaccess;    // Liste des objets a acceder
	                        // list of objects to access
	   bool reached_end;    // the last object has been extracted
	   bool update_root_obj; // object that is currently the root object
	                         // is to be updated

           // Parcours hierarchique
           short reftype;       // Type de reference
           // Requete de page ou d'E/S
           ChList *pgnums;      // Numeros de page
           int pageindex;       // Index de page
           // Demande de groupement
           ChList *o1;          // Liste des couples d'objets (o1,o2) // list of object pairs to be 
           ChList *o2;          // a regrouper                        // clustered
           ChList *fc;          // avec le poids du lien (fc)         // weight of object pair

           // Attributs de base

           Client *next;
           Client *previous;	   

	   bool hotTrans;

           Client();

};

/////////////////////////////////////////////////////////////////////
// CLASSE User
/////////////////////////////////////////////////////////////////////
// RA1 : Utilisateur
/////////////////////////////////////////////////////////////////////

class User: public Resource {

  public:

    // Constructeur

    User(char n[STRS], int cap, Simulation *sim, short id, Resource *rsp);

    // Evenements

    void Event10(Client *client);       // Arrivee sur RA1
    void Event11(Client *client);       // Temps de latence
    void Event12(Client *client);       // Attente processeur
    void Event13(Client *client);       // Generation transaction

  private:

    // Identificateur

    short userid;

    // Pointeurs ressources

    Resource *rsproc;

    // Sous-systeme logique

    Client *GenTrans();                 // R1

    int prevRoot;

};


/////////////////////////////////////////////////////////////////////
// CLASSE TransMgr
/////////////////////////////////////////////////////////////////////
// RA2 : Gestionnaire de Transactions
/////////////////////////////////////////////////////////////////////

class TransMgr: public Resource {

  public:

    // Constructeur

    TransMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb, float t);

    // Evenements

    void Event20(Client *client);       // Arrivee sur RA2
    void Event21(Client *client);       // Attente processeur (1)
    void Event22(Client *client);       // Extraction objet / res. BD
    void Event23(Client *client);       // Attente processeur (2)
    void Event24(Client *client);       // Acquisition verrous
    void Event25(Client *client);       // Requete d'objet
    void Event26(Client *client);       // Attente processeur (3)
    void Event27(Client *client);       // Traitement transaction + rendu verrous
    void Event28(Client *client);       // Transfert reseau
    
    int GetNumIdealPages(); // returns the number of pages that the active
                            // portion of the DB maps to.

    void Event100(Client *client);

    void NextSeqOid(Client * client);

  private:

    // Temps de transfert reseau

    float tnet;

    // Pointeurs ressources

    Resource *rsproc[MNUSERS], *rsbd;

    // set of distinct objects accesed used for calculation of
    // long term expansion factor
    set<int> _ltef_distinct_objs;

    // Sous-systeme logique

    void ExtObj(Client *client);        // R2
    void NextSet(Client *client);       // R2.1 : Acces ensembliste
    void NextSimp(Client *client);      // R2.2 : Parcours simple
    void NextHier(Client *client);      // R2.3 : Parcours hierarchique
    void NextStoc(Client *client);      // R2.4 : Parcours stochastique
    void NextObjRAUpdate(Client * client); // R2.5 : random attribute update 
    void NextObjInsert(Client * client); // R2.6 : object insertion transaction
    void NextObjDelete(Client * client); // R2.7 : object deletion transaction
    void NextObjSAUpdate(Client * client); // R2.8 : sequential attribute
                                           //        update
    void NextRefInsertion(Client * client); // R2.9 : reference insertion
                                            //        transaction 
    void NextRefDeletion(Client * client);  // R2.10 : reference deletion
                                            //         transaction   

    void TrTrans();                     // R6

};


class DROClustObj {

 private:

  int _objId;
  DROClustObj * _placementNext;
  DROClustObj * _placementPrev;
  DROClustObj * _sortNext;

 public:  

  DROClustObj(int objId);
  DROClustObj * getPlacementNext();
  void setPlacementNext(DROClustObj * placementNext);
  DROClustObj * getPlacementPrev();  
  void setPlacementPrev(DROClustObj * placementPrev);
  int getOid();
  DROClustObj * getSortNext();
  void setSortNext(DROClustObj * sortNext);
};

/////////////////////////////////////////////////////////////////////
// CLASSE ClustMgr
/////////////////////////////////////////////////////////////////////
// RA3 : Gestionnaire de Groupement
/////////////////////////////////////////////////////////////////////

class ClustMgr: public Resource {

  public:

    // Constructeur

    ClustMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb);
    void SetBuffPtr(BuffMgr *ptr);      // Initialise le pointeur vers le BuffMgr
    void SetObjPtr(ObjMgr *ptr);        // Initialise le pointeur vers l'ObjMgr

    // Evenements

    void Event30(Client *client);       // Reservation RA3 + Selection Traitement ou Groupement

    void Event31(Client *client);       // Tt : Attente processeur
    void Event32(Client *client);       // Tt : R7
    void Event33(Client *client);       // Tt : Fin de traitement

    void Event34(Client *client);       // Gr : Blocage BD
    void Event35(Client *client);       // Gr : Destruction clients fictifs
    void Event36(Client *client);       // Gr : Attente processeur
    void Event37(Client *client);       // Gr : R8
    void Event38(Client *client);       // Gr : Fin de clustering

    
    void Initialise();

    
  private:
    int _clear_q_count;
    int _prp_rand_curr_page;

    // Pointeurs ressources

    Resource *rsproc[MNUSERS], *rsbd;
    ObjMgr *objmgr;
    BuffMgr *buffmgr;
    int freepage;
    int freepageleft;
    // Sous-systeme logique

    float DSTCCollect(Client *client);    // R7, retourne le temps necessaire
    
    float DSTCCluster(Client *client);         // Groupement DSTC, retourne le temps necessaire
    
    float Cluster(Client *client);      // R8, retourne le temps necessaire

    float ObjRequest(Client *client);   // Requete d'objet, retourne le temps necessaire
    float AccDisk(int pagenum);         // Acces disque direct, retourne le temps necessaire

    ChList *Schedule(ChList *ur[3]);    // DSTC : Ordonnancement des UR en clusters

    // graph partitioning clustering functions
    float GPCollect(Client * client);
    float GPCluster(Client * client);
    float GPCPage(Client * client, int selected_page, 
		  set<int> * reorg_scope_set);
    void GreedyGP(set<int> * page_set, 
		  multimap<const int, part_info_t, int_cmp> * obj_seq);    

    // PRP clustering functions
    float PRPCollect(Client *client);
    float PRPCluster(Client * client);
    float PRPPRIOAnalysis(Client * client);
    float PRPPRIOCPage(Client * client, int pg_num, set<int> * reorg_scope_set);

    float PRPRANDAnalysis(Client * client);   
    float PRPRANDCPage(Client * client, int page_num);

    void PRP(set<int> * reorg_scope_set,   
		 multimap<const int,  part_info_t, int_cmp> * obj_seq);    



    float GPAnalysis(Client * client);
    void AssignPartToPage(set<int> * obj_set, int page_num);
    float PlaceObj(set<int> * page_set, multimap<const int, part_info_t, 
		   int_cmp> * obj_seq, Client * client);

    // DRO functions
    float DROCollect(Client * client);
    float DROCluster(Client * client);
    float DROAnalysis(Client * client);
    float DROReorg(vector<DROPageStats *> & clustPages, Client * client);
    float DROGenClustList(PQueue & clustObjPQ, 
			  hash_map<int, int, hash<int>, eqoid> & clustFreqMap,
			  vector<int> & clustList);
    float DROCompResembRate(vector<int> & clustList, 
			    vector<DROPageStats *> & clustPages);
    float DROPhysicalReorg(vector<int> & clustList,
			   vector<DROPageStats *> & clustPages, 
			   vector<int> & movedPages, Client * client);    
    
    float DROStatsUpdate(vector<int> & movedPage);    
    
    DROClustObj * 
      DROResearchProcedureResult(DROClustObj * startingObject, int d,
				 hash_map<int, int, hash<int>, eqoid>
				 & clustFreqMap,
				 hash_map<int, DROClustObj *, hash<int>,
				 eqoid> & oidObjClustMap, DROClustObj * chainStartObject);


    // Attributs prives

    int lastpage;                       // Derniere page accedee en groupement

    ChList *Clust(ChList *ur[3]) {
      ChList *cluster=new ChList;
      int i,j,k,v;
      short trouve;
      for (k=0;k<2;k++)
        for (i=1;i<=ur[k]->NbElem();i++) {
          v=ur[k]->Scan(i);
          trouve=0;
          j=1;
          while ((!trouve)&&(j<=cluster->NbElem()))
            if (cluster->Scan(j)==v) trouve=1;
            else j++;
          if (!trouve) cluster->Insert(1,v);
        }
      return cluster;
    }
};

// Classe Node
// Noeud dans l'algorithme d'ordonnancement des UR de DSTC

class Node {
  public:
    int num;                      // Numero de noeud
    ChList *list;                 // Liste des objets du noeud
    Node *next;                   // Ptr noeud suivant
};

/////////////////////////////////////////////////////////////////////
// CLASSE ObjMgr
/////////////////////////////////////////////////////////////////////
// RA4 : Gestionnaire d'objets
/////////////////////////////////////////////////////////////////////

class ObjMgr: public Resource {

  public:

    // Constructeur

    ObjMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t);

    // Renseignements

    float Tnet();                       // Retourne tnet

    // Evenements

    void Event40(Client *client);       // Reservation RA4
    void Event41(Client *client);       // Demande processeur
    void Event42(Client *client);       // R3 + prechargement
    void Event43(Client *client);       // Objet en memoire

    // Sous-systeme logique

    void ExtPg(Client *client);         // R3
    void Prefetch(Client *client);      // Routine de prechargement

    void SetBuffPtr(BuffMgr * bm);

  private:

    // Pointeurs ressources

    Resource *rsproc[MNUSERS];

    // Temps reseau unitaire

    float tnet;

    // Shift the hot region of the benchmark if it is time to.
    // The hot region is moved every HOT_REGION_MOVE_RATE object accesses.
    void _MoveHotRegion();
    
    // Pointer to buffer manager
    BuffMgr * buffmgr;
    Simulation * _simul_upper;
};

/////////////////////////////////////////////////////////////////////
// CLASSE BuffMgr
/////////////////////////////////////////////////////////////////////
// RA5 : Gestionnaire de cache
/////////////////////////////////////////////////////////////////////

// Classe BuffSlot : objet "cellule" du cache

class BuffSlot {

public: int pgnum;              // Numero de page
        short updated;          // Flag de mise a jour
        int nbacc;              // Nombre d'acces (LFU, GCLOCK)
        float date[MKLRU];      // Date de derniere utilisation (LRU)
        short clockflag;        // Flag d'utilisation (CLOCK)
};

class Dates {
 public:
  float date[MKLRU];
};

class BuffMgr: public Resource {

 public:

  
  // Constructeur
  
  BuffMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t);
  
  // Renseignements
  
  float Tnet();                       // Retourne tnet
  short Updated(int num);             // Retourne 1 si la page num a ete "mise a jour", 0 sinon
  int Page(int num);                  // Retourne le numero de la page en position num
  
  // Modifications cache
  
  void InitBuffer();                  // Initialisation du cache (vide)
  void SetUpdated(int num);           // Marque la page numero num a "mise a jour"
    void SetBuffer(int num,int pgnum,float time); // Insere la page pgnum en place num au temps time
    
    // Evenements

    void Event50(Client *client);       // Reservation RA5
    void Event51(Client *client);       // Demande processeur (1)
    void Event52(Client *client);       // R4 et requete d'E/S si necessaire
    void Event53(Client *client);       // Demande processeur (2)
    void Event54(Client *client);       // Remplacement de page
    void Event55(Client *client);       // Transfert reseau

    // Partie publique du sous-systeme logique

    void AccPage(Client *client);       // R4
    int Replace();                      // Remplacement des pages dans le cache

    void DisplayExtTen();
    void DisplayHeat();
    int NextPgNum(int * iter);
    int NextNonEmptyPgNum(int * iter);
    int NextUpdatedPgNum(int * iter);
    int NextNonEmptyUpdatedPgNum(int * iter );
    void StartIter(int * iter);
    void RemoveNonRes(set<int> * oid_set);
    void RemoveNonUpdated(set<int> * oid_set);
    bool PageInMem(int pg_num);
    bool NonEmptyPageInMem(int pg_num);
    bool UpdatedPageInMem(int pg_num);
    bool NonEmptyUpdatedPageInMem(int pg_num);
    bool ObjInMem(int obj_id);
    bool ObjInUpdatedMem(int obj_id);

    int GetTotalDistinctPages();
    int GetNumWinsItered();
    int GetLTEFNumDPages();


  private:


    hash_map<int, Dates *, hash<int>, eqoid> LRU_K_Hist; 

    hash_map<int, float, hash<int>, eqoid> LRU_K_Last;

    // Pointeurs ressources

    Resource *rsproc;

    // Temps reseau unitaire

    float tnet;

    // Definition du cache memoire

    BuffSlot buffer[MBUFFSIZE];         // Cache
    int buffpos;                        // Position remplissage cache
    int clock;                          // Sauvegarde position / algos
                                        // LFU, LRU, CLOCK, GCLOCK

    // Sous-systeme logique

    int RANDOMrep();                    // Remplacement RANDOM
    int FIFOrep();                      // Remplacement FIFO
    int LFUrep();                       // Remplacement LFU
    int LRUrep();                       // Remplacement LRU-K
    int CLOCKrep();                     // Remplacement CLOCK
    int GCLOCKrep();                    // Remplacement GCLOCK

    int LRUHistRep();                   // LRU-K using history (the version on
                                        // the LRU-K [SIGMOD 93] paper.)

    int Otherep();                      // Remplacement autre (Utilisateur)


    // for calculation of Working Set Size (WSS)
    int _wss_num_pages;
    vector<int> _wss_window;
    int _num_wins_itered;
    int _total_distinct_pages;
    int _wss_distinct_pages;    
    
    // for calculation of long term expansion factor
    set <int> _ltef_distinct_pages;    
    
};

/////////////////////////////////////////////////////////////////////
// CLASSE IOSubSys
/////////////////////////////////////////////////////////////////////
// RA6 : Sous-systeme des Entrees/Sorties
/////////////////////////////////////////////////////////////////////

class IOSubSys: public Resource {

  public:

    // Constructeur

    IOSubSys(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsd);
    
    void SetBuffPtr(BuffMgr * bm);

    // Evenements

    void Event60(Client *client);       // Reservation RA6
    void Event61(Client *client);       // Demande processeur
    void Event62(Client *client);       // Demande disque
    void Event63(Client *client);       // Acces disque (R5)

  private:

    // Numero de la derniere page accedee

    int lastpage;

    // Pointeurs ressources

    Resource *rsproc, *rsdisk;

    // Sous-systeme logique

    void AccDisk(Client *client);       // R5

    BuffMgr * buffmgr;

};



