
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// pqueuem.h:  Custom made priority queue (methods)
///////////////////////////////////////////////////////////////////////////

void PQueue::InsertElem(float key, int page_num) {
  _queue.insert(pair<const float, int>(key, page_num));
  _r_queue[page_num] = key;
}
void PQueue::ClearQueue()  {
  _queue.clear();
  _r_queue.clear();
}

int PQueue::Size() {
  return _queue.size();
}

PQueue::PQueue() {
  EndInt = -1;
}

int PQueue::PopTopElem() {
  int r_value;

  multimap<const float, int, float_comp>::iterator it = _queue.begin();

  if (it != _queue.end()) {
    r_value = (*it).second;
    // _r_queue.erase(_pageExists.find((*it).second));
    _r_queue.erase((*it).second);
    _queue.erase(it);
    return r_value;
  }
  else {
    return NULL_PAGE_NUM;
  }
}

int PQueue::PopBottomElem() {
  int r_value;

  multimap<const float, int, float_comp>::reverse_iterator it 
    = _queue.rbegin();
  multimap<const float, int, float_comp>::iterator itf;

  if (it != _queue.rend()) {
    r_value = it->second;
    // _r_queue.erase(_pageExists.find((*it).second));
    _r_queue.erase(r_value);
    for (itf = _queue.lower_bound(it->first); itf != _queue.end(); itf++) {
      if (itf->second == r_value) {
	_queue.erase(itf);
	break;
      }      
    }
    return r_value;
  }
  else {
    return NULL_PAGE_NUM;
  }
}

bool PQueue::PageExists(int page_num) {
  if (_r_queue.find(page_num) != _r_queue.end()) {
    return true;
  }
  else {
    return false;
  }
}

// returns true if remove was successful
bool PQueue::RemovePage(int page_num) {
  float cb;
  
  hash_map<const int, float, hash<int>, eq_int>::iterator it;
  multimap<const float, int, float_comp>::iterator it2;

  if ((it = _r_queue.find(page_num)) != _r_queue.end()) {
    cb = (*it).second;
    _r_queue.erase(page_num);   
    if ((it2 = _queue.lower_bound(cb)) != _queue.end()) {
      while (it2 != _queue.end()) {
	if ((*it2).second == page_num) {	
	  _queue.erase(it2);
	  return true;
	}
	it2++;      
      }
      return false;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
  
}

int PQueue::StartIter() {
  _internal_iter = _queue.begin();
  if (_internal_iter == _queue.end()) {
    return EndInt;
  }
  else {
    return _internal_iter->second;
  }
}

int PQueue::Next() {
  _internal_iter++;
  if (_internal_iter == _queue.end()) {
    return EndInt;
  }
  else {
    return _internal_iter->second;
  }
}

void PQueue::RemoveAtIter() {
  _r_queue.erase(_internal_iter->second);   
  _queue.erase(_internal_iter);
}

int PQueue::End () {
  return EndInt;
}



