
/////////////////////////////////////////////////////////////////////
// Package SIMULation (C++ discrete event simulation package)
// Version 1.1 g++, fevrier 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// simulm.h : Definition des methodes du moteur de simulation
// Invariable
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// Inclusions
/////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>

/////////////////////////////////////////////////////////////////////
// CLASSE Simulation
/////////////////////////////////////////////////////////////////////


// CLASSE Simulation : Constructeur

Simulation::Simulation(float start, float max, long int seed) {
  Reset(start, max, seed);
  clientlist=NULL;
  scheduler=new Scheduler;
  eventmanager=new EventManager(this);
}

// CLASSE Simulation : Destructeur

Simulation::~Simulation() {

  delete scheduler;
  delete eventmanager;
}

// CLASSE Simulation : Execution de la simulation

void Simulation::Run(int nreplic) {

  int nextevent, charcount;
  Client *client;

  // Initialisation
  eventmanager->Init();

  printf("\nSimulation started... ");
  charcount=21;

  for (int i=1; i<=nreplic; i++) {

    charcount+=(digit(i)+3);
    if (charcount>79) {
      charcount=digit(i)+3;
      printf("\n");
    }
    printf("[%d]\n",i);
    
    // Initialisation de la replication
    tnow=tstart;
    eventmanager->InitRep();
    //    cout << "lp_tt = " << lp_tt << endl; 
    client=NewClient();      // NE PAS OUBLIER DE DETRUIRE LES CLIENTS !
    
    eventmanager->ExecuteEvent(0,client);

    // Moteur
    //    cout << "tnow = " << tnow << endl;
    // cout << "nb trans = " << NBTRANS[NTTRANS] << endl;
    //    DB->DisplayNumRef();
    int p =0;

    bool working_set_limit_reached = false;
    set<int> processed_obj_set;
    //    int curr_total_size = 0;
    
#ifdef TRANSLIMIT_ON
    while  ((((TRANSLIMIT < 0) && (working_set_limit_reached == false)) ||
	     ((TRANSLIMIT >= 0) && (NBTRANS[NTTRANS] < TRANSLIMIT))) && 
	    (!scheduler->IsEmpty())) {
#else
    while ((tnow<tmax) && (!scheduler->IsEmpty())) {
#endif

      if ((p % 10000) == 0) {
	//cout << "tnow = " << tnow << endl;
	//cout << "nb trans = " <<  NBTRANS[NTTRANS] << endl;
	//cout << "trans limit = " << TRANSLIMIT << endl;
	/*
	  if (TRANSLIMIT < 0) {
	  //   cout << "curr_total_size = " << curr_total_size << endl;
	  //cout << "total working set size = " << 
	  //  ((float) BUFFSIZE *  (float)PGSIZE) / (float) curr_total_size
	  //	 << endl;
	  }	
	  
	  DB->DisplayNumRef();
	*/
      }
      
      nextevent=scheduler->GetEventCode();
      tnow=scheduler->GetEventDate();
      client=scheduler->GetClient();
      scheduler->DestroyEvent();
      eventmanager->ExecuteEvent(nextevent,client);
      p++;
    }

    // Calculs statistiques
    eventmanager->Stats();

    // Destruction des clients encore presents dans le systeme
    PurgeClientList();
  }

  charcount+=17;
  if (charcount>79) printf("\n");
  printf("End of simulation\n");

  // Resultats
  

  eventmanager->DisplayStats();

}

// CLASSE Simulation : Retourne l'adresse du Scheduler

Scheduler *Simulation::Sched() {

  return scheduler;
}

// CLASSE Simulation : Retourne la valeur de tnow

float Simulation::Tnow() {

  return tnow;
}

// CLASSE Simulation : Retourne la valeur de tmax

float Simulation::Tmax() {

  return tmax;
}

// CLASSE Simulation : Reinitialisation

void Simulation::Reset(float start, float max, long int seed) {

  tstart=start;
  tmax=max;
  //  cout << "tmax in reset = " << tmax << endl; 
  if (seed>0) rseed=seed;
  else rseed=DEFAULT_SEED;

  // Randomize
  lp_tt=rseed;
}

// Classe Simulation : Creation d'un nouveau client dans clientlist

Client *Simulation::NewClient() {

  Client *newclient;

  newclient=new Client;
  // Insertion en tete
  newclient->next=clientlist;
  newclient->previous=NULL;
  if (clientlist!=NULL) clientlist->previous=newclient;
  clientlist=newclient;
  return newclient;
}

// Classe Simulation : Suppression d'un client dans clientlist

void Simulation::KillClient(Client *client) {

  Client *p, *n;

  p=client->previous;
  n=client->next;

  if (p!=NULL) p->next=n;
  else clientlist=n; // en tete
  if (n!=NULL) n->previous=p;

  delete client;
}

// Classe Simulation : Suppression de tous les clients dans clientlist

void Simulation::PurgeClientList() {

  Client *cour, *prec;

  cour=clientlist;
  prec=NULL;
  while (cour!=NULL) {
    prec=cour;
    cour=cour->next;
    delete prec;
  }
  clientlist=NULL;
}

/////////////////////////////////////////////////////////////////////
// CLASSE Scheduler
/////////////////////////////////////////////////////////////////////

// CLASSE Scheduler : Constructeur

Scheduler::Scheduler() {

  top=NULL;
  bottom=NULL;
}

// CLASSE Scheduler : Destructeur

Scheduler::~Scheduler() {

  Purge();
}

// CLASSE Scheduler : Retourne l'etat de l'echeancier
// (1 si vide, 0 si non vide)

int Scheduler::IsEmpty() {

  if (top==NULL) return 1;
  else return 0;
}

// CLASSE Scheduler : Insertion dans l'echeancier

void Scheduler::Schedule(int eventcode, float eventdate, Client *client) {

  SchedulerCell *prec, *cour, *nouv;

  nouv=new SchedulerCell(eventcode,eventdate,client);

  prec=NULL;
  cour=bottom;

  while ((cour!=NULL) && (eventdate<cour->Date())) {
    prec=cour;
    cour=cour->Previous();
  }

  nouv->SetPrevious(cour);
  nouv->SetNext(prec);

  if (prec!=NULL) prec->SetPrevious(nouv);
  else bottom=nouv;
  if (cour!=NULL) cour->SetNext(nouv);
  else top=nouv;
}

// CLASSE Scheduler : Retourne le code du premier evenement

int Scheduler::GetEventCode() {

  if (top!=NULL) return top->Code();
  else return -1;
}

// CLASSE Scheduler : Retourne la date du premier evenement

float Scheduler::GetEventDate() {

  if (top!=NULL) return top->Date();
  else return -1;
}

// CLASSE Scheduler : Retourne le pointeur vers le client

Client *Scheduler::GetClient() {

  if (top!=NULL) return top->Cli();
  else return NULL;
}

// CLASSE Scheduler : Detruit le premier evenement

void Scheduler::DestroyEvent() {

  SchedulerCell *sauve;

  sauve=top;
  if (top->Next()!=NULL) {
    top=top->Next();
    top->SetPrevious(NULL);
  } else {
    top=NULL;
    bottom=NULL;
  }
  delete sauve;
}

// CLASSE Scheduler : Vide l'echeancier

void Scheduler::Purge() {

  SchedulerCell *cour, *save;

  cour=bottom;
  while (cour!=NULL) {
    save=cour;
    cour=cour->Next();
    delete save;
  }

  top=NULL;
  bottom=NULL;
}

/////////////////////////////////////////////////////////////////////
// CLASSE SchedulerCell
/////////////////////////////////////////////////////////////////////

// CLASSE SchedulerCell : Constructeur

SchedulerCell::SchedulerCell(int code, float date, Client *cli) {

  eventcode=code;
  eventdate=date;
  client=cli;
  next=NULL;
  previous=NULL;
}

// CLASSE SchedulerCell : Retourne le code de l'evenement

int SchedulerCell::Code() {

  return eventcode;
}

// CLASSE SchedulerCell : Retourne la date de l'evenement

float SchedulerCell::Date() {

  return eventdate;
}

// CLASSE SchedulerCell : Retourne le pointeur vers le client

Client *SchedulerCell::Cli() {

  return client;
}

// CLASSE SchedulerCell : Retourne la cellule suivante

SchedulerCell *SchedulerCell::Next() {

  return next;
}

// CLASSE SchedulerCell : Retourne la cellule precedente

SchedulerCell *SchedulerCell::Previous() {

  return previous;
}

// CLASSE SchedulerCell : Reinitialise le pointeur vers la cellule suivante

void SchedulerCell::SetNext(SchedulerCell *newnext) {

  next=newnext;
}

// CLASSE SchedulerCell : Reinitialise le pointeur vers la cellule precendente

void SchedulerCell::SetPrevious(SchedulerCell *newprev) {

  previous=newprev;
}

/////////////////////////////////////////////////////////////////////
// CLASSE Resource
/////////////////////////////////////////////////////////////////////

// CLASSE Resource : Constructeur

Resource::Resource(char n[STRS], int cap, Simulation *sim) {

  strcpy(name,n);
  capacity=cap;
  simul=sim;
  top=NULL;
  bottom=NULL;
}

// CLASSE Resource : Destructeur

Resource::~Resource() {

  PurgeQueue();
}

// CLASSE Resource : Vide la file d'attente

void Resource::PurgeQueue() {

  QueueCell *cour, *save;

  cour=bottom;
  while (cour!=NULL) {
    save=cour;
    cour=cour->Next();
    delete save;
  }

  top=NULL;
  bottom=NULL;
}

// CLASSE Resource : Reservation de ressource (P)

void Resource::P(int event, Client *client, int prior) {

  ccapacity--;
  if (ccapacity>=0) {                    // Action immediate
    response-=simul->Tnow();
    simul->Sched()->Schedule(event,simul->Tnow(),client);
  } else {                               // Evenement mis en attente
    wait-=simul->Tnow();
    EnQueue(event,client,prior);
  }
}

// CLASSE Resource : Liberation de ressource (V)

void Resource::V() {

  Client *nextclient;

  ccapacity++;
  if (ccapacity>capacity) {
    printf("Error: capacity overflow for resource %s at time %f\n",name,simul->Tnow());
    ccapacity=capacity;
  }
  response+=simul->Tnow();
  nbserv++;
  if (!QueueEmpty()) {                  // File d'attente non vide
    nextclient=GetClient();
    wait+=simul->Tnow();
    response-=simul->Tnow();
    simul->Sched()->Schedule(GetEventCode(),simul->Tnow(),nextclient);
    DestroyTop();
  }
}

// CLASSE Resource : Retourne l'adresse de l'objet simulation

Simulation *Resource::Sim() {

  return simul;
}

// CLASSE Resource : Initialisation des compteurs statistiques

void Resource::ResetCounters() {

  ccapacity=capacity;
  response=0;
  wait=0;
  nbserv=0;
}

// CLASSE Resource : Initialisation des statistiques globales

void Resource::ResetStats() {

  int i;

  for (i=0; i<5; i++) {
    stats[i]=0;
    stats2[i]=0;
  }
  n=0;

}

// CLASSE Resource : Calcul des statistiques

void Resource::Stats() {

  int nbwait, nbbs, i;
  float s[5];

  if (ccapacity<0) {
    nbwait=-ccapacity;
    nbbs=capacity;
  } else {
    nbwait=0;
    nbbs=capacity-ccapacity;
  }

  // Temps de reponse replication
  if (nbserv!=0) s[0]=(response+nbbs*Sim()->Tnow())/nbserv;
  else s[0]=0;
  // Temps d'attente replication
  if ((nbserv+nbbs)!=0) s[1]=(wait+nbwait*Sim()->Tnow())/(nbserv+nbbs);
  else s[1]=0;
  // Servis replication
  s[2]=nbserv;
  // En cours replication
  s[3]=nbbs;
  // En attente replication
  s[4]=nbwait;

  // Cumul
  for (i=0; i<5; i++) {
    stats[i]+=s[i];
    stats2[i]+=s[i]*s[i];
  }
  n++;
}

// CLASSE Resource : Affichage des statistiques

void Resource::DisplayStats() {

  int i;

  // Calculs

  for (i=0; i<5; i++) {
    if (n!=0) mean[i]=stats[i]/n;
    else mean[i]=0;
    if (n!=0) dev[i]=(n*stats2[i]-stats[i]*stats[i])/(n*n);
    else dev[i]=0;
    if (dev[i]>0) dev[i]=sqrt(dev[i]);
    else dev[i]=0;
    if (n>1) cint[i]=t(n-1)*dev[i]/sqrt(n);
    else cint[i]=0;
  }

  // Affichage

  printf("\nStatistics for resource: %s (0.95 confidence interval)\n\n",name);
  printf("\t* Mean response time              : %10.2f\t+/- %10.2f\n",mean[0],cint[0]);
  printf("\t* Mean waiting time               : %10.2f\t+/- %10.2f\n",mean[1],cint[1]);
  printf("\t* Mean # of clients served        : %10.2f\t+/- %10.2f\n",mean[2],cint[2]);
  printf("\t* Mean # of clients being served  : %10.2f\t+/- %10.2f\n",mean[3],cint[3]);
  printf("\t* Mean # of clients still waiting : %10.2f\t+/- %10.2f\n",mean[4],cint[4]);
}

// CLASSE Resource : Retourne les stats (moyenne)

float Resource::Mean(short i) {

  if ((i>=0) && (i<=4)) return mean[i];
  else return -1;
}

// CLASSE Resource : Retourne les stats (ecart-type)

float Resource::Dev(short i) {

  if ((i>=0) && (i<=4)) return dev[i];
  else return -1;
}

// CLASSE Resource : Insertion dans la queue

void Resource::EnQueue(int eventcode, Client *client, int priority) {

  QueueCell *prec, *cour, *nouv;

  nouv=new QueueCell(eventcode,client,priority);

  prec=NULL;
  cour=bottom;

  while ((cour!=NULL) && (priority>cour->Priority())) {
    prec=cour;
    cour=cour->Previous();
  }

  nouv->SetPrevious(cour);
  nouv->SetNext(prec);

  if (prec!=NULL) prec->SetPrevious(nouv);
  else bottom=nouv;
  if (cour!=NULL) cour->SetNext(nouv);
  else top=nouv;
}

// CLASSE Resource : Retourne le code du premier evenement en file

int Resource::GetEventCode() {

  if (top!=NULL) return top->Code();
  else return -1;
}

// CLASSE Resource : Retourne le pointeur vers le premier client de la file

Client *Resource::GetClient() {

  if (top!=NULL) return top->Cli();
  else return NULL;
}

// CLASSE Resource : Elmination de la tete de file d'attente

void Resource::DestroyTop() {

  QueueCell *save;

  if (top!=NULL) {
    save=top;
    top=top->Next();
    if  (top!=NULL) top->SetPrevious(NULL);
    else bottom=NULL;
    delete save;
  }
}

// CLASSE Resource : Retourne l'etat de la file d'attente
// (1 : vide, 0 : non vide)

int Resource::QueueEmpty() {

  if (top==NULL) return 1;
  else return 0;
}

/////////////////////////////////////////////////////////////////////
// CLASSE QueueCell
/////////////////////////////////////////////////////////////////////

// CLASSE QueueCell : Constructeur

QueueCell::QueueCell(int code, Client *cli, int prior) {

  eventcode=code;
  client=cli;
  priority=prior;
  next=NULL;
  previous=NULL;
}

// CLASSE QueueCell : Retourne le code de l'evenement

int QueueCell::Code() {

  return eventcode;
}

// CLASSE QueueCell : Retourne le pointeur vers le client

Client *QueueCell::Cli() {

  return client;
}

// CLASSE QueueCell : Retourne la priorite de l'evenement

int QueueCell::Priority() {

  return priority;
}

// CLASSE QueueCell : Retourne la cellule suivante

QueueCell *QueueCell::Next() {

  return next;
}

// CLASSE QueueCell : Retourne la cellule precedente

QueueCell *QueueCell::Previous() {

  return previous;
}

// CLASSE QueueCell : Reinitialise le pointeur vers la cellule suivante

void QueueCell::SetNext(QueueCell *newnext) {

  next=newnext;
}

// CLASSE QueueCell : Reinitialise le pointeur vers la cellule precendente

void QueueCell::SetPrevious(QueueCell *newprev) {

  previous=newprev;
}

