
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// statsc.h: Statistics used by OPCF clustering algorithms (classes)
///////////////////////////////////////////////////////////////////////////

#include <hash_map.h>
#include <list.h>

typedef int tension_t;

// heat stat
// transition tension
// link tension

enum stat_t {
  heat, trans, link, direction
};


class TensionEntry {
 public:
  int obj_id;
  tension_t tension;
};



typedef list<TensionEntry *> TensionList_t;

typedef list<TensionEntry * >::iterator TensionLIter_t;

class CStats;

class ObjStats {
 private:
  int _heat;
  TensionList_t _tensionList;  // bi directional tension
  TensionList_t _dirTensionList; // uni directional tension
  TensionLIter_t _tli;
  int _obj_id;
  CStats * _stats;
  int _totalOutTension;
 public:
  
  ObjStats(int i_oid, bool inc_heat, CStats * stats);
  ~ObjStats();
  void IncHeat();
  void UpdateTension(int curr_oid, int incAmount);
  void UpdateDirTension(int curr_oid, int incAmount);
  void Display();
  int Start();
  int StartDir();
  int NextOid();
  int NextDirOid();
  int GetCurrTension();
  float GetCurrProbTension();
  int GetHeat();
  void DeleteTension(int i_oid);
  void SetTensionZero(int i_oid);
  int GetRootOid();
  int GetStatsSize();
};


typedef hash_map<int, ObjStats *, hash<int>, eqoid> StatsMap_t;
typedef hash_map<int, ObjStats *, hash<int>, eqoid>::iterator StatsMapIter_t;

class CStats {
 private:
  // stats used by graph partitioning algorithm
  StatsMap_t _stats;
  
  int _maxPageHeat;      // maximum heat
  int _maxPageExTension; // maximum page external tension
  float _maxPageWaste;   // maximum amount of waste in the page with 
                         // the maxium amount of waste
  StatsMapIter_t _iter;

 public:
  CStats() {
    _maxPageHeat = 0;
    _maxPageExTension = 0;
    _maxPageWaste = 0.0;
  };

  void IncStat(int prev_oid, int curr_oid, stat_t stat_type, 
	       int incTensionAmount);
  void Display();
  ObjStats * GetObjStats(int i_oid);
  void SetMaxPageHeat(int heat);
  void SetMaxPageExTension(int tension);
  void SetMaxPageWaste(float waste);

  int GetMaxPageHeat();
  int GetMaxPageExTension();
  float GetMaxPageWaste();

  bool Empty();

  void DeleteStats(int i_oid);
  void DeleteTension(int from_oid, int to_oid);
  void SetTensionZero(int from_oid, int to_oid);
  int GPBufferIter;
  void IncTension(int prev_oid, int curr_oid, stat_t stat_type, int incAmount);
  ~CStats();

  ObjStats * StartIter();
  ObjStats * NextIter();
  int GetTotalSize();
};








