
///////////////////////////////////////////////////////////////////////////
// DOCB Code
// Version 1.0, Feb. 2002
// Zhen He
// Australian National University
// statsm.h : Statistics used by OPCF clustering algorithms (methods)
///////////////////////////////////////////////////////////////////////////
ObjStats::ObjStats(int i_oid, bool inc_heat, CStats * stats) {
  if (inc_heat == true) {
    _heat = 1;
  }
  else {
    _heat = 0;
  }
  _obj_id = i_oid;
  _stats = stats;
  _totalOutTension = 0;
}

int ObjStats::GetRootOid() {
  return _obj_id;
}

void ObjStats::IncHeat() {
  _heat++;
}


int ObjStats::GetHeat() {
  return _heat;
}


int  ObjStats::Start() {
  _tli = _tensionList.begin();
  if (_tli != _tensionList.end() ) {
    return (*_tli)->obj_id;
  }
  else {
    return NULL_OID;
  }
}


int  ObjStats::StartDir() {
  _tli = _dirTensionList.begin();
  if (_tli != _dirTensionList.end() ) {
    return (*_tli)->obj_id;
  }
  else {
    return NULL_OID;
  }
}


float ObjStats::GetCurrProbTension() {
  return ((float)(*_tli)->tension / (float)_totalOutTension);
}


int ObjStats::GetCurrTension() {
  return (*_tli)->tension;
}



int ObjStats::NextOid() {
  int r_oid;
  
  _tli++;
  if (_tli == _tensionList.end()) {
    r_oid = NULL_OID;
  }
  else {
    r_oid = (*_tli)->obj_id;
  }
  return r_oid;
}


int ObjStats::NextDirOid() {
  int r_oid;
  
  _tli++;
  if (_tli == _dirTensionList.end()) {
    r_oid = NULL_OID;
  }
  else {
    r_oid = (*_tli)->obj_id;
  }
  return r_oid;
}

void ObjStats::UpdateTension(int curr_oid, int incAmount) {
   bool found = false;
   for (_tli = _tensionList.begin(); _tli != _tensionList.end(); _tli++) {
     if ((*_tli)->obj_id == curr_oid) {
       (*_tli)->tension += incAmount;
       found = true;
       break;
     }
   }
   if (found == false) {
     TensionEntry * te = new TensionEntry;
     te->obj_id = curr_oid;
     te->tension = incAmount;
     _tensionList.push_front(te);
   }
}


void ObjStats::UpdateDirTension(int curr_oid, int incAmount) {
   bool found = false;
   for (_tli = _dirTensionList.begin();
	_tli != _dirTensionList.end(); _tli++) {
     if ((*_tli)->obj_id == curr_oid) {
       (*_tli)->tension += incAmount;
       found = true;
       break;
     }
   }
   if (found == false) {
     TensionEntry * te = new TensionEntry;
     te->obj_id = curr_oid;
     te->tension = incAmount;
     _dirTensionList.push_front(te);
   }
   _totalOutTension += incAmount;
}

int ObjStats::GetStatsSize() {
  int size;
  size = (_dirTensionList.size() + _tensionList.size()) * sizeof(int);
  return size;
}


ObjStats::~ObjStats() {
  int id;

  for (list<TensionEntry * >::iterator it = _tensionList.begin();
       it != _tensionList.end(); it++) {
    // delete back references
      id = (*it)->obj_id;
      _stats->DeleteTension(id, _obj_id);
      delete *it;
  }
  
  for (list<TensionEntry * >::iterator it = _dirTensionList.begin();
       it != _dirTensionList.end(); it++) {
    delete *it;
  }

  // _tensionList will automatically be distroied since it is not a pointer
  // field
}

void ObjStats::DeleteTension(int i_oid) {
  TensionEntry * te;

  bool found = false;

  // cout << "deleting tension" << endl;
  for (list<TensionEntry * >::iterator it = _tensionList.begin();
       it != _tensionList.end(); it++) {
    if ((*it)->obj_id == i_oid) {
      te = *it;
      _totalOutTension -= te->tension;
      _tensionList.remove(*it);
      delete te;
      found = true;
      break;
    }
  }
 
  if (found == false) {
    cerr << "ERROR: statsm.h, tension to be deleted not found" 
	 << endl;
    exit(1);
  }
  
}

void ObjStats::SetTensionZero(int i_oid) {
  
  // cout << "deleting tension" << endl;
  for (_tli = _tensionList.begin(); _tli != _tensionList.end(); _tli++) {
    if ((*_tli)->obj_id == i_oid) {
      (*_tli)->tension = 0;     
      cout << "found tension to be set to zero" << endl;
      break;
    }
  }

  for (_tli = _dirTensionList.begin(); _tli != _dirTensionList.end(); _tli++) {
    if ((*_tli)->obj_id == i_oid) {
      (*_tli)->tension = 0;     
      cout << "found tension to be set to zero" << endl;
      break;
    }
  }

  _totalOutTension = 0;

}

void CStats::SetTensionZero(int from_oid, int to_oid) {
  if (_stats[from_oid] != NULL) {
    _stats[from_oid]->SetTensionZero(to_oid);
  }
}

void CStats::Display() {
   hash_map<int, ObjStats *, hash<int>, eqoid>::iterator it;

  for (it = _stats.begin(); it != _stats.end() ; it++) {
    ((*it).second)->Display();
  }
}

void CStats::SetMaxPageHeat(int heat) {
  if (heat > _maxPageHeat) {
    _maxPageHeat = heat;
  }
}


void CStats::SetMaxPageWaste(float waste) {
  if (waste > _maxPageWaste) {
    _maxPageWaste = waste;
  }
}

void CStats::SetMaxPageExTension(int tension) {
  if (tension > _maxPageExTension) {
    _maxPageExTension = tension;
  }
}

int CStats::GetMaxPageHeat() {
  return _maxPageHeat;
}

int CStats::GetMaxPageExTension() {
  return _maxPageExTension;
}

float CStats::GetMaxPageWaste() {
  return _maxPageWaste;
}


void ObjStats::Display() {
  cout << _obj_id << " : " << _heat << endl;
  cout << _obj_id << " -> ";
  TensionLIter_t tli;
  for (tli = _tensionList.begin(); tli != _tensionList.end(); tli++) {
    cout << (*tli)->obj_id << "(";
    cout << (*tli)->tension << ") ";
  }
  cout << endl;
}

ObjStats * CStats::GetObjStats(int i_oid) {
  return _stats[i_oid];
}

void CStats::DeleteStats(int i_oid) {
  if (_stats[i_oid] != NULL) {
    delete _stats[i_oid];
    _stats.erase(i_oid);
  }
}


void CStats::DeleteTension(int from_oid, int to_oid) {
  if (_stats[from_oid] != NULL)
    _stats[from_oid]->DeleteTension(to_oid);
}



// only increment tension but not heat
void CStats::IncTension(int prev_oid, int curr_oid, stat_t stat_type, 
			int incAmount) {

    if (_stats[curr_oid] ==NULL) {
      _stats[curr_oid] = new ObjStats(curr_oid, false, this);
    }
    
    if (_stats[prev_oid] ==NULL) {
      _stats[prev_oid] = new ObjStats(prev_oid, false, this);
    }
    
    if (prev_oid != curr_oid) {
      
      if (stat_type == trans) {
	_stats[prev_oid]->UpdateTension(curr_oid, incAmount);
	// also update backward tension for calculation of external tension
	// of a page
	_stats[curr_oid]->UpdateTension(prev_oid, incAmount);
	_stats[prev_oid]->UpdateDirTension(curr_oid, incAmount);
      }
      else if (stat_type == direction) {
	_stats[prev_oid]->UpdateDirTension(curr_oid, incAmount);
      }
    }
}

void CStats::IncStat(int prev_oid, int curr_oid, stat_t stat_type,
		     int incTensionAmount) {
  // insert a new gp stat entry for the current object if it does not 
  // already exist


    if (_stats[curr_oid] ==NULL) {
      _stats[curr_oid] = new ObjStats(curr_oid, true, this);
    } else {
      _stats[curr_oid]->IncHeat();
    }
    
    if (prev_oid != NULL_OID) {
      // either we are using transition or we are using link tension and a link
      // exists
      if ((stat_type == trans) || 
	  ((stat_type == link) 
	   &&(DB->LinkExists(prev_oid, curr_oid) == true))) {
	if (prev_oid != curr_oid) {	  
	  //      cout <<"updating tension" << endl;
	  // update the tension stat for the object just accessed and the 
	  // previously accessed object.
	  _stats[prev_oid]->UpdateTension(curr_oid, incTensionAmount);
	  // also update backward tension for calculation of external tension
	  // of a page
	  _stats[curr_oid]->UpdateTension(prev_oid, incTensionAmount);

	  _stats[prev_oid]->UpdateDirTension(curr_oid, incTensionAmount);
	}
      }
      else if (stat_type == direction) {	
	if (prev_oid != curr_oid) {	  
	  _stats[prev_oid]->UpdateDirTension(curr_oid, incTensionAmount);  
	}
      }
    }
}

bool CStats::Empty() {
  if (_stats.size() == 0) {
    return true;
  }
  else {
    return false;
  }
}

CStats::~CStats() {
  hash_map<int, ObjStats *, hash<int>, eqoid>::iterator it;

  for (it = _stats.begin(); it != _stats.end() ; it++) {
    delete it->second;
  }
}

ObjStats * CStats::StartIter() {
  _iter = _stats.begin();
  if (_iter == _stats.end()) {
    return NULL;
  }
  else {
    return _iter->second;
  }
}

ObjStats * CStats::NextIter() {
  _iter++;
  if (_iter != _stats.end()) {
    return _iter->second;
  }
  else {
    return NULL;
  }
}

int CStats::GetTotalSize() {
  hash_map<int, ObjStats *, hash<int>, eqoid>::iterator it;
  int totalSize = 0;

  for (it = _stats.begin(); it != _stats.end() ; it++) {
    totalSize += (it->second)->GetStatsSize();
  }
  return totalSize;
}








