/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database)
// Modele d'evaluation de performances generique pour SGBDOO
// Version 2.0 C++, avril 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// voodb2.cc : Programme principal
/////////////////////////////////////////////////////////////////////
// Updated
// -------
// Name: Zhen He
// Institution: Australian National University
// Date: Feb. 2002
/////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// Inclusions
/////////////////////////////////////////////////////////////////////

#include <stdio.h>

int HOTSELECTED;

float HOTPERCENTFROMTRACE;

#include "simutil.h"
#include "voodb2uc.h"
#include "voodb2.h"

// lots of misc types and classes
#include "misc.h"

#include "pqueuec.h"
#include "pqueuem.h"

#include "DOCBChangec.h"
#include "DOCBChangem.h"

#include "DOCBDepc.h"
#include "DOCBDepm.h"

#include "DOCBChDepc.h"
#include "DOCBChDepm.h"

HybridMovingWindDep * HMWD;
HybridGMovingWindDep * HGMWD;
HybridDep * HD;
HybridHeatGrowthDep * HHGD;
HybridCycleDep * HCD;

PQueue * PQ;
#include "basec.h"
Database *DB;                 // Objet base de donnees (var. globale)


#include "statsc.h"
CStats * STATS;
#include "statsm.h"

#include "DROStatc.h"
#include "DROStatm.h"

#include "simulc.h"
#include "eventc.h"
#include "voodb2um.h"
#include "basem.h"
#include "simulm.h"

#include "eventm.h"



/////////////////////////////////////////////////////////////////////
// Prototypes
/////////////////////////////////////////////////////////////////////

void GetParam();                // Saisie des parametres au clavier
void DisplayParam();            // Affichage des parametres
void ParamTest();               // Test des parametres

/////////////////////////////////////////////////////////////////////
// Programme principal
/////////////////////////////////////////////////////////////////////



int main(int argc, char ** argv) {

   char c;
  int i,j;
  Simulation *sim;
  
  // Intro.
  printf("\n*** VOODB2 (C++) ***\n");  

  // Saisie des parametres
  printf("\n(D)efault parameters / (C)ustom parameters : ");
  c=getchar();
  c = 'D';
  if ((c!='D') && (c!='d')) GetParam();

  DisplayParam();
  ParamTest();

  // Calcul des parametres induits

  PCT1  = PSET;
  PCT2  = PCT1  + PSIMPLE;
  PCT3  = PCT2  + PHIER;
  PCC   = PCT3  + PSTOCH;
  PORAU = PCC   + POBJRAU;
  POI   = PORAU + POBJI;
  POD   = POI   + POBJD;
  POSAU = POD   + POBJSAU;
  PRI   = POSAU + PREFI;
  PRD   = PRI   + PREFD;
  PID   = PRD   + POID;

  objAssMeth_t assignMethod;

  if (MOVE_ASSIGNMETHOD == 1) {
    assignMethod = class_method;
  }
  else if (MOVE_ASSIGNMETHOD == 2) {
    assignMethod = random_method;
  }
  else {
    cerr << "MOVE_ASSIGNMETHOD specified is valid, in voodb2.cc" << endl;
    exit(1);
  }

  if ( MOVE_WINDTYPE == 1 ) {

    HMWD = new HybridMovingWindDep( DEP_RANDOMPROB, DEP_SREFPROB, 
				    DEP_DREFPROB, DEP_TRAVERSEDPROB, 
				    DEP_CLASSPROB, DEP_U, DEP_C, 
				    MOVE_REGIONSIZE,
				    MOVE_HIGHESTPROBW, MOVE_LOWESTPROBW,
				    MOVE_H, assignMethod, DEP_R,
				    DEP_RANDHOTREGSIZE, DEP_RANDHOTREGPROB);
  }
  else if ( MOVE_WINDTYPE == 2 ) {

    HGMWD = new HybridGMovingWindDep( DEP_RANDOMPROB, DEP_SREFPROB, 
				      DEP_DREFPROB, DEP_TRAVERSEDPROB,
				      DEP_CLASSPROB, DEP_U, DEP_C,
				      MOVE_REGIONSIZE,
				      MOVE_HIGHESTPROBW, MOVE_LOWESTPROBW,
				      MOVE_H, assignMethod, 
				      MOVE_PROBWINCRSIZE, DEP_R,
				      DEP_RANDHOTREGSIZE, DEP_RANDHOTREGPROB);
    
  }
  else if ( MOVE_WINDTYPE == 3 ) {

    HD = new HybridDep(DEP_RANDOMPROB, DEP_SREFPROB, DEP_DREFPROB, 
		       DEP_TRAVERSEDPROB, DEP_CLASSPROB, DEP_U, DEP_C,
		       DEP_RANDHOTREGSIZE, DEP_RANDHOTREGPROB,
		       assignMethod, DEP_R);

  }
  else if ( MOVE_WINDTYPE == 4 ) {

    HHGD = new HybridHeatGrowthDep(DEP_RANDOMPROB, DEP_SREFPROB, 
				   DEP_DREFPROB, DEP_TRAVERSEDPROB,
				   DEP_CLASSPROB, DEP_U, DEP_C,
				   MOVE_SMALLREGIONSIZE, MOVE_LARGEREGIONSIZE,
				   MOVE_HIGHESTPROBW, MOVE_LOWESTPROBW,
				   MOVE_NUMPERIODS, assignMethod, 
				   TRANSLIMIT, DEP_R, DEP_RANDHOTREGSIZE,
				   DEP_RANDHOTREGPROB);    

  }
  else if ( MOVE_WINDTYPE == 5 ) {

    HCD = new HybridCycleDep(DEP_RANDOMPROB, DEP_SREFPROB, DEP_DREFPROB,
			     DEP_TRAVERSEDPROB, DEP_CLASSPROB, DEP_U, DEP_C,
			     MOVE_REGIONSIZE, MOVE_HIGHESTPROBW,
			     MOVE_LOWESTPROBW, MOVE_H, assignMethod, DEP_R,
			     DEP_RANDHOTREGSIZE, DEP_RANDHOTREGPROB);
  }
  
  j=2;
  STOCHTAB[0]=1.0/j;
  for (i=1; i<MAXNREF; i++) {
    j=j*2;
    STOCHTAB[i]=STOCHTAB[i-1]+1.0/j;
  }

  // Instanciation de l'objet simulation
#ifdef TRANSLIMIT_ON
  sim=new Simulation(0,TSIM, SIMSEED);
#else
  sim=new Simulation(0,TRANSLIMIT, SIMSEED);  
#endif
  // Execution de la simulation

  sim->Run(NREPLIC);

  // Nettoyage

  delete sim;

  cout << "-- hot percent = " << HOTPERCENTFROMTRACE  << endl;
}

/////////////////////////////////////////////////////////////////////
// Saisie des parametres au clavier
/////////////////////////////////////////////////////////////////////

void GetParam() {

  cout << endl;
  cout << "TSIM="; cin >> TSIM;
  cout << "TRANSLIMIT="; cin >> TRANSLIMIT;
  cout << "NREPLIC="; cin >> NREPLIC;

  cout << endl;
  cout << "SYSCLASS="; cin >> SYSCLASS;
  cout << "MULTILVL="; cin >> MULTILVL;
  cout << "NUSERS=";   cin >> NUSERS;
  cout << "PGSIZE=";   cin >> PGSIZE;
  cout << "BUFFSIZE="; cin >> BUFFSIZE;
  cout << "GETLOCK=";  cin >> GETLOCK;
  cout << "RELLOCK=";  cin >> RELLOCK;
  cout << "DISKSEA=";  cin >> DISKSEA;
  cout << "DISKLAT=";  cin >> DISKLAT;
  cout << "DISKTRA=";  cin >> DISKTRA;
  cout << "NETTHRU=";  cin >> NETTHRU;
  
  cout << endl;
  cout << "CLUSTP=";   cin >> CLUSTP;
  cout << "INITP=";    cin >> INITP;
  cout << "PGREP=";    cin >> PGREP;
  cout << "KLRU=";     cin >> KLRU;
  cout << "PREFETCH="; cin >> PREFETCH;

  cout << endl;
  cout << "NC=";       cin >> NC;
  cout << "MAXNREF=";  cin >> MAXNREF;
  cout << "BASESIZE="; cin >> BASESIZE;
  cout << "NO=";       cin >> NO;
  cout << "NREFT=";    cin >> NREFT;
  cout << "DIST1=";    cin >> DIST1;
  cout << "DIST2=";    cin >> DIST2;
  cout << "DIST3=";    cin >> DIST3;
  cout << "DIST4=";    cin >> DIST4;
  cout << "INFCLASS="; cin >> INFCLASS;
  cout << "SUPCLASS="; cin >> SUPCLASS;

  cout << endl;
  cout << "TRANS_THINK="; cin >> TRANS_THINK;
  cout << "PSET=";     cin >> PSET;
  cout << "PSIMPLE=";  cin >> PSIMPLE;
  cout << "PHIER=";    cin >> PHIER;
  cout << "PSTOCH=";   cin >> PSTOCH;
  cout << "NUPDT =";   cin >> NUPDT;
  cout << "POBJRAU=";  cin >> POBJRAU;
  cout << "PREFI=";    cin >> PREFI;
  cout << "PREFD=";    cin >> PREFD;
  cout << "POID=";    cin >> POID;
  cout << "POBJI=";    cin >> POBJI;
  cout << "POBJSAU=";  cin >> POBJSAU;
  cout << "PCLUSTD=";  cin >> PCLUSTD;  
  cout << "SETDEPTH="; cin >> SETDEPTH;
  cout << "SIMDEPTH="; cin >> SIMDEPTH;
  cout << "HIEDEPTH="; cin >> HIEDEPTH;
  cout << "STODEPTH="; cin >> STODEPTH;
  cout << "SEQOIDDEPTH="; cin >> SEQOIDDEPTH;
  cout << "DIST5=";    cin >> DIST5;
  cout << "DIST100=";    cin >> DIST100;
  cout << "DIST101=";    cin >> DIST101;
  cout << "DIST102=";    cin >> DIST102;
  cout << "DIST103=";    cin >> DIST103;

  cout << endl;
  cout << "HOTRANGE="; cin >> HOTRANGE;
  cout << "HOTACCESSPROB="; cin >> HOTACCESSPROB;
  cout << "HOT_REGION_MOVE_HOT_FRACT="; cin >> HOT_REGION_MOVE_HOT_FRACT;
  cout << "HOT_REGION_MOVE_RATE="; cin >> HOT_REGION_MOVE_RATE;
  cout << endl;

  cout << endl;
  cout << "DSTC Clustering" << endl;
  cout << "N=";        cin >> N;
  cout << "NP=";       cin >> NP;
  cout << "P=";        cin >> P;
  cout << "TFA=";      cin >> TFA;
  cout << "TFE=";      cin >> TFE;
  cout << "TFC=";      cin >> TFC;
  cout << "W=";        cin >> W;
  //  cout << "S=";        cin >> S;
  cout << endl;

  cout << "GP Clustering" << endl;
  cout << "GPNUMACCESS="; cin >> GPNUMACCESS;
  cout << "GPAGRAIN="; cin >> GPAGRAIN;
  cout << "GP_CB_THRESHOLD="; cin >> GP_CB_THRESHOLD;
  cout << "GPRGRAIN="; cin >> GPRGRAIN;
  cout << endl;


  cout << "PRP Clustering" << endl;
  cout << "PRPAGRAIN="; cin >> PRPAGRAIN;
  cout << "PRPNUMACCESS="; cin >> PRPNUMACCESS;
  cout << "PRPPRIORGRAIN="; cin >> PRPPRIORGRAIN;
  cout << "PRP_CB_THRESHOLD="; cin >> PRP_CB_THRESHOLD;
  cout << endl;

  cout << "DRO Clustering" << endl;
  cout << "DRONUMACCESS="; cin >> DRONUMACCESS;
  cout << "DROMINUR="; cin >> DROMINUR;
  cout << "DROMINLT="; cin >> DROMINLT;
  cout << "DROPCRate="; cin >> DROPCRate;
  cout << "DROMaxD="; cin >> DROMaxD;
  cout << "DROMaxDR="; cin >> DROMaxDR;
  cout << "DROMaxRR="; cin >> DROMaxRR;
  cout << endl;

  cout << "DOCB Parameters" << endl;
  
  cout << "MOVE_WINDTYPE="; cin >>  MOVE_WINDTYPE;
  cout << "DEP_RANDHOTREGSIZE="; cin >> DEP_RANDHOTREGSIZE;
  cout << "DEP_RANDHOTREGPROB="; cin >> DEP_RANDHOTREGPROB;
  cout << "DEP_C="; cin >> DEP_C;
  cout << "DEP_U="; cin >> DEP_U;
  cout << "DEP_NUMDREF="; cin >> DEP_NUMDREF;
  cout << "MOVE_REGIONSIZE="; cin >> MOVE_REGIONSIZE;
  cout << "MOVE_HIGHESTPROBW="; cin >> MOVE_HIGHESTPROBW;
  cout << "MOVE_LOWESTPROBW="; cin >> MOVE_LOWESTPROBW;
  cout << "MOVE_H="; cin >> MOVE_H;
  cout << "MOVE_ASSIGNMETHOD="; cin >> MOVE_ASSIGNMETHOD;
  cout << "MOVE_PROBWINCRSIZE="; cin >> MOVE_PROBWINCRSIZE;
  cout << "DEP_R="; cin >> DEP_R;
  cout << "MOVE_SMALLREGIONSIZE="; cin >> MOVE_SMALLREGIONSIZE;
  cout << "MOVE_LARGEREGIONSIZE="; cin >> MOVE_LARGEREGIONSIZE;
  cout << "MOVE_NUMPERIODS="; cin >> MOVE_NUMPERIODS;
  cout << "DEP_RANDOMPROB="; cin >> DEP_RANDOMPROB;
  cout << "DEP_SREFPROB=;"; cin >> DEP_SREFPROB;
  cout << "DEP_DREFPROB="; cin >> DEP_DREFPROB;
  cout << "DEP_TRAVERSEDPROB="; cin >> DEP_TRAVERSEDPROB;
  cout << "DEP_CLASSPROB="; cin >> DEP_CLASSPROB;

}

/////////////////////////////////////////////////////////////////////
// Affichage des parametres
/////////////////////////////////////////////////////////////////////

void DisplayParam() {

  printf("\n*** Simulation parameters ***\n\n");

  printf("     TSIM=%f\n",TSIM);
  printf("  NREPLIC=%d\n\n",NREPLIC);

  printf(" SYSCLASS=%d\n",SYSCLASS);
  printf(" MULTILVL=%d\n",MULTILVL);
  printf("   NUSERS=%d\n",NUSERS);
  printf("   PGSIZE=%d\n",PGSIZE);
  printf(" BUFFSIZE=%d\n",BUFFSIZE);
  printf("  GETLOCK=%f\n",GETLOCK);
  printf("  RELLOCK=%f\n",RELLOCK);
  printf("  DISKSEA=%f\n",DISKSEA);
  printf("  DISKLAT=%f\n",DISKLAT);
  printf("  DISKTRA=%f\n",DISKTRA);
  printf("  NETTHRU=%f\n\n",NETTHRU);

  printf("   CLUSTP=%d\n",CLUSTP);
  printf("    INITP=%d\n",INITP);
  printf("    PGREP=%d\n",PGREP);
  printf("     KLRU=%d\n",KLRU);
  printf(" PREFETCH=%d\n\n",PREFETCH);

  printf("       NC=%d\n",NC);
  printf("  MAXNREF=%d\n",MAXNREF);
  printf(" BASESIZE=%d\n",BASESIZE);
  printf("       NO=%d\n",NO);
  printf("    NREFT=%d\n",NREFT);
  printf("    DIST1=%d\n",DIST1);
  printf("    DIST2=%d\n",DIST2);
  printf("    DIST3=%d\n",DIST3);
  printf("    DIST4=%d\n",DIST4);
  printf(" INFCLASS=%d\n",INFCLASS);
  printf(" SUPCLASS=%d\n\n",SUPCLASS);

  printf("    TRANS_THINK=%f\n",TRANS_THINK);
  printf("     PSET=%f\n",PSET);
  printf("  PSIMPLE=%f\n",PSIMPLE);
  printf("    PHIER=%f\n",PHIER);
  printf("   PSTOCH=%f\n",PSTOCH);
  printf("  PCLUSTD=%f\n",PCLUSTD);
  printf(" SETDEPTH=%d\n",SETDEPTH);
  printf(" SIMDEPTH=%d\n",SIMDEPTH);
  printf(" HIEDEPTH=%d\n",HIEDEPTH);
  printf(" STODEPTH=%d\n",STODEPTH);
  printf("    DIST5=%d\n\n",DIST5);

  printf("DSTC    N=%d\n",N);
  printf("       NP=%d\n",NP);
  printf("        P=%d\n",P);
  printf("      TFA=%f\n",TFA);
  printf("      TFE=%f\n",TFE);
  printf("      TFC=%f\n",TFC);
  printf("        W=%f\n",W);
  //  printf("        S=%d\n",S);
}

/////////////////////////////////////////////////////////////////////
// Test des parametres
/////////////////////////////////////////////////////////////////////

void ParamTest() {

  short halt=0;

  if (MAXNREF>MMAXNREF) {
    printf("\nMAXNREF value (%d) is too large! It should be <= %d.\n",
      MAXNREF,MMAXNREF);
    halt=1;
  }
  if (NUSERS>MNUSERS) {
    printf("\nNUSERS value (%d) is too large! It should be <= %d.\n",
      NUSERS,MNUSERS);
    halt=1;
  }
  if (NUSERS>INFINITE) {
    printf("\nNUSERS value (%d) is too large! It should be <= %d (INFINITE).\n",
      NUSERS,INFINITE);
    halt=1;
  }
  if (NC>MNC) {
    printf("\nNC value (%d) is too large! It should be <= %d.\n",
      NC,MNC);
    halt=1;
  }
  if (NO>MNO) {
    printf("\nNO value (%d) is too large! It should be <= %d.\n",
      NO,MNO);
    halt=1;
  }
  if (NREFT>MNREFT) {
    printf("\nNREFT value (%d) is too large! It should be <= %d.\n",
      NREFT,MNREFT);
    halt=1;
  }
  if (KLRU>MKLRU) {
    printf("\nKLRU value (%d) is too large! It should be <= %d.\n",
      KLRU,MKLRU);
    halt=1;
  }
  if (BUFFSIZE>MBUFFSIZE) {
    printf("\nBUFFSIZE value (%d) is too large! It should be <= %d.\n",
      BUFFSIZE,MBUFFSIZE);
    halt=1;
  }



  float dep_sum = DEP_RANDOMPROB + DEP_SREFPROB + DEP_DREFPROB 
    + DEP_TRAVERSEDPROB + DEP_CLASSPROB;
  
  if ((dep_sum < 1.000000) || (dep_sum > 1.000001)) {
    cout << "Error: probabilities of the different dependency types" << endl;
    cout << "entered do not add up to 1.0 " << endl;
    cout << "It adds up to: "<< dep_sum << " instead" << endl;
    halt = 1;
  }
  
  float trans_sum =  POID + PREFD + PREFI + POBJSAU + POBJD + POBJI + 
    POBJRAU + PSTOCH + PHIER + PSIMPLE + PSET;
  cout << "POID = " << POID << endl;
  if ((trans_sum < 1.000000) || (trans_sum > 1.000001)) {
    cout << "Error: probabilities of the different types of transactions" 
	 << endl;
    cout << "entered do not add up to 1.0 " << endl;
    cout << "It adds up to: "<< trans_sum << " instead" << endl;
    halt = 1;
  }
  
  
  if (halt) {
    puts("\n*** STOP ***\n");
    exit(-1);
  }
}
