
/////////////////////////////////////////////////////////////////////
// VOODB2 (Virtual Object Oriented Database)
// Modele d'evaluation de performances generique pour SGBDOO
// Version 2.0 C++, aout 1998
// Jerome Darmont
// LIMOS, Universite Blaise Pascal (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// voodb2um.h : Definition des methodes utilitaires
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// CLASSE ChList
/////////////////////////////////////////////////////////////////////

// CLASSE ChList : Constructeur

ChList::ChList() {

  head=NULL;
  queue=NULL;
  nbelem=0;
  position=0;
  cur=NULL;
  prev=NULL;
  next=NULL;
}

// CLASSE ChList : Destructeur

ChList::~ChList() {

  IntCell *del;

  cur=head;
  while (cur!=NULL) {
    del=cur;
    cur=cur->Next();
    delete del;
  }
}

void * ChList::operator new(size_t size) {
  void *t;
  t = malloc(32);
  if (t == NULL) {
    cout << "malloc failed" << endl;
  }
  return t;
}

void ChList::operator delete(void * deadObject, size_t size) {
  free(deadObject);
}

// CLASSE ChList : Affichage des pointeurs

void ChList::Print() {

  printf("\n\t     nbelem = %d\n",nbelem);
  printf("\t head value = ");
  if (head!=NULL) printf("%d\n",head->Value());
  else printf("NULL\n");
  printf("\tqueue value = ");
  if (queue!=NULL) printf("%d\n",queue->Value());
  else printf("NULL\n");
  printf("\t   position = %d\n",position);
  printf("\t  cur value = ");
  if (cur!=NULL) printf("%d\n",cur->Value());
  else printf("NULL\n");
  printf("\t prev value = ");
  if (prev!=NULL) printf("%d\n",prev->Value());
  else printf("NULL\n");
  printf("\t next value = ");
  if (next!=NULL) printf("%d\n",next->Value());
  else printf("NULL\n");
}

// CLASSE ChList : Positionnement en place

// Attributs modifies en retour : position, cur, prev, next

void ChList::Pos(int place) {

  int dist1, dist2, dist3;

  // Calcul des distances a resp. head, pos et queue;

  dist1=abs(place-1);
  dist2=abs(position-place);
  dist3=abs(nbelem-place);

  // Positionnement dans la liste

  if ((dist1<dist2)&&(dist1<dist3)) { // En tete
    position=1;
    cur=head;
  } else if ((dist3<dist2)&&(dist3<dist1)) { // En queue
    position=nbelem;
    cur=queue;
  } // Sinon on reste a position

  // Choix du sens de parcours et parcours

  if ((cur==head)||(place>=position)) { // En "avant"
    if (cur!=NULL) prev=cur->Prev();
    else prev=NULL;
    while ((cur!=NULL) && (position<place)) {
      prev=cur;
      cur=cur->Next();
      position++;
    }
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  } else { // En "arriere"
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
    while ((cur!=NULL) && (position>place)) {
      next=cur;
      cur=cur->Prev();
      position--;
    }
    if (cur!=NULL) prev=cur->Prev();
    else prev=NULL;
  }
}

// CLASSE ChList :Insertion en place

void ChList::Insert(int place, int value) {

  IntCell *newcell;

  // Positionnement

  Pos(place);

  // Insertion nouvelle cellule

  newcell=new IntCell(value);
  newcell->SetNext(cur);
  newcell->SetPrev(prev);
  if (cur!=NULL) {
    cur->SetPrev(newcell);
    if (prev!=NULL) prev->SetNext(newcell);
    else head=newcell;          // Insertion en tete
  } else {
    if (position==1) {          // Premier element
      head=newcell;
      queue=newcell;

    } else {                    // Insertion en queue
      if (prev != NULL) {
	prev->SetNext(newcell);
      }
      queue=newcell;
    }
    next=NULL;
  }
  nbelem++;
  cur=newcell;
  prev=cur->Prev();
  next=cur->Next();
}

// CLASSE ChList : Modification en place

void ChList::Update(int place, int value) {

  Pos(place);

  // Modification valeur cellule courante
  if (cur!=NULL) cur->SetVal(value);
  else { // Repositionnement en tete
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  }
}

// CLASSE ChList : Suppression en place

void ChList::Delete(int place) {

  // Positionnement

  Pos(place);

  // Suppression cellule courante

  if (cur!=NULL) {
    if (prev!=NULL) prev->SetNext(next);
    else  head=next;             // Suppression en tete
    if (next!=NULL) next->SetPrev(prev);
    else {             // Suppression en queue
      queue=prev;
      position--;
    }
    delete cur;
    nbelem--;
    if (position>1) cur=prev;
    else cur=head;
    if (cur!=NULL) {
      prev=cur->Prev();
      next=cur->Next();
    } else {
      next=NULL;
      prev=NULL;
      position=0;
    }
  } else { // Repositionnement en tete
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
  }
}

// CLASSE ChList : Inversion de liste chainee

void ChList::Reverse() {

  IntCell *newhead, *newqueue, *lcur, *newcell;

  // Copie

  newhead=NULL;
  newqueue=NULL;
  lcur=head;
  while (lcur!=NULL) {
    // Nouvelle cellule inseree en tete de nouvelle liste
    newcell=new IntCell(lcur->Value());
    newcell->SetNext(newhead);
    newcell->SetPrev(NULL);
    // Bug fix that I put in
    if (newhead != NULL) newhead->SetPrev(newcell);
    if (newhead==NULL) newqueue=newcell;
    newhead=newcell;
    lcur=lcur->Next();
  }

  // Destruction ancienne liste et assignation pointeurs

  lcur=head;
  while (lcur!=NULL) {
    newcell=lcur;
    lcur=lcur->Next();
    delete newcell;
  }
  head=newhead;
  queue=newqueue;
  position=1;
  cur=head;
  prev=NULL;
  next=cur->Next();
}

// CLASSE ChList : Retour valeur place

int ChList::Scan(int place) {

  // Positionnement
  Pos(place);

  // Resultat
  if ((cur!=NULL) && (position==place)) return cur->Value();
  else { // Repositionnement en tete
    position=1;
    cur=head;
    prev=NULL;
    if (cur!=NULL) next=cur->Next();
    else next=NULL;
    return -1;
  }
}

// CLASSE ChList : Retour nombre d'elements

int ChList::NbElem() {

  return nbelem;
}

// CLASSE ChList : Affichage liste entiere

void ChList::Display() {

  IntCell *lcur;

  printf("LIST (%d elements):",nbelem);
  lcur=head;
  while (lcur!=NULL) {
    printf(" %d",lcur->Value());
    lcur=lcur->Next();
  }
  printf("\n");
}

/////////////////////////////////////////////////////////////////////
// CLASSE IntCell
/////////////////////////////////////////////////////////////////////

// CLASSE IntCell : Constructeur

IntCell::IntCell(int val) {

  value=val;
}

// CLASSE IntCell : Initialisation value

void IntCell::SetVal(int val) {

  value=val;
}

// CLASSE IntCell : Initialisation ptr suivant

void IntCell::SetNext(IntCell *n) {

  next=n;
}

// CLASSE IntCell : Initialisation ptr precedent

void IntCell::SetPrev(IntCell *n) {

  prev=n;
}

// CLASSE IntCell : Retourne la valeur

int IntCell::Value() {

  return value;
}

// CLASSE IntCell : Retourne le pointeur suivant

IntCell* IntCell::Next() {

  return next;
}

// CLASSE IntCell : Retourne le pointeur precedent

IntCell* IntCell::Prev() {

  return prev;
}




